Package: janusplot
Type: Package
Title: Asymmetric Smoothed-Association Matrices via GAM Fits
Version: 0.1.0
Authors@R: c(
    person("Max", "Moldovan", , "max.moldovan@adelaide.edu.au",
           role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0001-9680-8474"))
  )
Description: Render a pairwise, asymmetric smoothed-association matrix
    of continuous variables. Each cell shows the fitted spline from an
    'mgcv' generalised additive model, with the upper triangle
    displaying 'gam(x_j ~ s(x_i))' and the lower triangle
    'gam(x_i ~ s(x_j))'. Unlike Pearson's correlation matrix, the
    visualisation is intentionally asymmetric, revealing
    heteroscedasticity, leverage, and directional non-linearity that
    a single scalar correlation hides. An asymmetry index and a
    24-category shape taxonomy quantify the directional difference
    and qualitative form of each fitted smooth.
License: GPL (>= 3)
URL: https://github.com/max578/janusplot,
        https://max578.github.io/janusplot/
BugReports: https://github.com/max578/janusplot/issues
Encoding: UTF-8
Language: en-AU
Depends: R (>= 4.3.0)
Imports: mgcv (>= 1.9.0), ggplot2 (>= 3.5.0), patchwork (>= 1.1.0),
        grid, stats, cli (>= 3.6.0), lifecycle, rlang (>= 1.1.0)
Suggests: agridat, future, future.apply, knitr, MASS, palmerpenguins,
        rmarkdown, testthat (>= 3.0.0), vdiffr (>= 1.0.0), withr
VignetteBuilder: knitr
RoxygenNote: 7.3.3
Config/testthat/edition: 3
Config/testthat/parallel: true
Config/Needs/website: pkgdown
LazyData: true
NeedsCompilation: no
Packaged: 2026-04-23 14:06:24 UTC; a1222812
Author: Max Moldovan [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-9680-8474>)
Maintainer: Max Moldovan <max.moldovan@adelaide.edu.au>
Repository: CRAN
Date/Publication: 2026-04-28 18:30:08 UTC
Built: R 4.6.0; ; 2026-05-01 21:20:05 UTC; unix

# Generated by roxygen2: do not edit by hand

export(janusplot)
export(janusplot_data)
export(janusplot_shape_cutoffs)
export(janusplot_shape_hierarchy)
export(janusplot_shape_metrics)
export(janusplot_shape_sensitivity)
export(janusplot_shape_sensitivity_plot)
export(janusplot_shape_sensitivity_shapes)
export(janusplot_shape_sensitivity_summary)
importFrom(rlang,.data)
importFrom(rlang,arg_match)
importFrom(rlang,check_installed)
importFrom(rlang,is_installed)
importFrom(stats,as.dist)
importFrom(stats,as.formula)
importFrom(stats,complete.cases)
importFrom(stats,cor)
importFrom(stats,hclust)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,sd)

<!-- README.md is auto-generated from README.Rmd — edit README.Rmd, then
     run rmarkdown::render("README.Rmd") to refresh README.md. -->

# janusplot <img src="man/figures/logo.png" align="right" height="139" alt="janusplot logo" />

<!-- badges: start -->

[![R-CMD-check](https://github.com/max578/janusplot/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/max578/janusplot/actions/workflows/R-CMD-check.yaml)
[![test-coverage](https://github.com/max578/janusplot/actions/workflows/test-coverage.yaml/badge.svg)](https://github.com/max578/janusplot/actions/workflows/test-coverage.yaml)
[![pkgdown](https://github.com/max578/janusplot/actions/workflows/pkgdown.yaml/badge.svg)](https://max578.github.io/janusplot/)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
<!-- badges: end -->

Asymmetric, GAM-based smoothed-association matrices for continuous
variables. Each off-diagonal cell shows a directional
`mgcv::gam(y ~ s(x))` fit, so the upper and lower triangles tell
different stories whenever the relationship is genuinely asymmetric —
precisely where a scalar Pearson correlation loses information.

Two novelties, paired with the asymmetry story:

1.  **Asymmetry index**
    `A = |EDF_yx − EDF_xy| / (EDF_yx + EDF_xy) ∈ [0, 1]` — a
    single-number summary of the directional disparity per pair.
2.  **24-category shape taxonomy** — each fitted smooth is classified
    into one of 24 named shapes (`linear_up`, `skewed_peak`, `bimodal`,
    `bi_wave`, `rippled_monotone`, …) via a `(T, I)` dispatch on
    turning-point and inflection counts, with monotonicity/convexity
    indices for disambiguation. The taxonomy rolls up through three
    broader tiers: `archetype` (7), `monotonic` (3), `linear` (2) —
    grounded in shape-constrained regression (Pya & Wood 2015),
    dose-response pharmacology (Calabrese 2008), and Morse critical-
    point classification (Milnor 1963).

## Install

``` r
# development version from GitHub (fast; no vignettes)
pak::pak("max578/janusplot")

# or, with vignettes built locally (recommended — enables
# browseVignettes("janusplot") and vignette("janusplot")):
# install.packages("remotes")
remotes::install_github(
  "max578/janusplot",
  build_vignettes = TRUE,
  dependencies    = TRUE
)
```

Note: neither `pak::pak()` nor `devtools::install_github()` /
`remotes::install_github()` build vignettes by default when installing
from a source repository. The `build_vignettes = TRUE` flag above is
required for `browseVignettes()` to find them. The CRAN release ships
prebuilt vignettes and needs no extra flag.

## Quick start

``` r
library(janusplot)

# Palmer penguins — four continuous traits
d <- na.omit(palmerpenguins::penguins[,
       c("bill_length_mm", "bill_depth_mm",
         "flipper_length_mm", "body_mass_g")])
janusplot(d)
```

Default encoding:

- **Cell colour** — Pearson correlation on a diverging `RdBu` palette
  symmetric around zero (override via `colour_by = "spearman"` /
  `"kendall"` / `"edf"` / `"deviance_gap"` / `"none"`).
- **Bottom-left** — `A = ...` (asymmetry index) stacked over
  `EDF = ...`.
- **Top-right** — significance glyph for the smooth’s F-test
  (`· * ** ***`).
- **Below the matrix** — a standing reference legend illustrating all 24
  shape categories as canonical thumbnail splines, labelled
  `<name> (<code>)`.

Opt into a per-cell shape marker via `annotations`:

``` r
# Two-letter shape code, top-left (ASCII — safe on any font / PDF):
janusplot(d, annotations = c("edf", "A", "code"))

# Unicode shape glyph, bottom-right:
janusplot(d, annotations = c("edf", "A", "shape"),
          glyph_style = "unicode")
```

## Shape taxonomy — what gets classified

| Archetype (7) | Categories | Example |
|----|----|----|
| `monotone_linear` | `linear_up` `linear_down` | `y = x` |
| `monotone_curved` | `convex_up` `concave_up` `convex_down` `concave_down` `s_shape` `rippled_monotone` | `tanh`, `sqrt`, `exp(−x)` |
| `unimodal` | `u_shape` `inverted_u` `skewed_peak` `broad_peak` `rippled_peak` | `(x−.5)²`, `x·exp(−3x)`, plateau |
| `wave` | `wave` `warped_wave` `rippled_wave` `complex_wave` | `sin(2πx)` family |
| `multimodal` | `bimodal` `bimodal_ripple` `bi_wave` `bi_wave_ripple` | two-peak mix, `sin(4πx)` |
| `chaotic` | `complex` | ≥ 5 extrema / inflections |
| `degenerate` | `flat` `indeterminate` | constant / fit failure |

Full table (with 2-letter codes + monotonic / linear rollups + glyph +
gloss) via `janusplot_shape_hierarchy()`.

## Sensitivity study — built-in

The classifier’s recovery behaviour is characterised across a full
factorial of sample sizes × noise levels × ground-truth shapes:

``` r
# Precomputed 2160-fit demo sweep — zero wait
data("shape_sensitivity_demo")
janusplot_shape_sensitivity_plot(shape_sensitivity_demo, "recovery_curves")
janusplot_shape_sensitivity_plot(shape_sensitivity_demo, "confusion_archetype")

# Run your own — full grid in parallel
future::plan(future::multisession, workers = 4L)
res <- janusplot_shape_sensitivity(parallel = TRUE)
```

Four diagnostic plots (`"confusion_fine"` / `"confusion_archetype"` /
`"accuracy_grid"` / `"recovery_curves"`) + summary aggregations at fine
and archetype levels. Design, pre-registered hypotheses, and full
walk-through in the
[`shape-recognition-sensitivity`](https://max578.github.io/janusplot/articles/shape-recognition-sensitivity.html)
vignette.

## Key features

- **Real GAM fits** via `mgcv` — EDF, F-test p-values, confidence
  envelopes, random effects via `s(g, bs = "re")`.
- **Asymmetric matrix** — upper / lower triangles carry the two
  directional regressions.
- **Three correlation flavours** computed per pair — Pearson (default),
  Spearman, Kendall — surfaced in `janusplot_data()` output.
- **`adjust =` formula** — propagate covariates and random effects into
  every cell’s smooth.
- **`janusplot_data()`** — per-pair raw fits, correlations, EDFs,
  p-values, shape metrics + hierarchy columns, without the plot.
- **`janusplot(..., with_data = TRUE)`** — plot + tidy data frame in one
  call.
- **Parallel dispatch** via `future.apply` for large matrices and
  sensitivity sweeps.
- **Deterministic** — seed `2026L` pins the shipped demo + vignette
  figures; user sweeps accept `seed =`.

## Why asymmetric?

A Pearson correlation discards both the **shape** of an association and
the **direction** information that a non-linear data-generating process
leaves in its residuals. Under the additive-noise causal discovery
setting (Hoyer et al. 2009; Peters et al. 2014) the forward regression
`y ~ s(x)` and its inverse `x ~ s(y)` are generically asymmetric when
the underlying DGP is non-linear, and that asymmetry identifies the
causal direction under mild conditions. `janusplot` surfaces this
asymmetry as a **visual pre-discovery diagnostic** — not a causal
inference procedure. See the vignette and the accompanying paper for
scope and explicit non-claims.

## Documentation

- [Reference + articles](https://max578.github.io/janusplot/) (pkgdown
  site)
- `vignette("janusplot")` — quickstart + feature tour
- `vignette("shape-recognition-sensitivity")` — design, hypotheses,
  every diagnostic plot
- Paper: *Beyond Pearson: Visualising Asymmetric Non-linear Associations
  with Generalised Additive Models* (Moldovan, in preparation; target
  venue *R Journal*).

## Status

`R CMD check --as-cran` clean (0 errors, 0 warnings, 3 cosmetic NOTEs —
new submission / local env); 190 test expectations; 88.5 % coverage.

## Citation

``` r
citation("janusplot")
```

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse   = TRUE,
  comment    = "#>",
  fig.alt    = paste(
    "Asymmetric smoothed-association matrix produced by",
    "janusplot(); diagonal cells hold variable labels,",
    "off-diagonal cells show the fitted mgcv::gam spline",
    "with a 95% confidence envelope, raw-data scatter,",
    "and per-cell annotations for n, EDF, and smooth",
    "significance glyph."
  ),
  fig.width  = 7,
  fig.height = 7,
  dpi        = 110,
  eval       = TRUE
)
library(ggplot2)
set.seed(2026L)

## ----quickstart---------------------------------------------------------------
library(janusplot)

# Four numeric columns from mtcars (32 rows: small but illustrative)
janusplot(mtcars[, c("mpg", "hp", "wt", "qsec")])

## ----nonlinear----------------------------------------------------------------
n <- 300
x1 <- runif(n, -3, 3)
x2 <- x1^2 + rnorm(n, sd = 0.6)   # quadratic on x1
x3 <- sin(x1) + rnorm(n, sd = 0.4) # sinusoidal on x1
x4 <- rnorm(n)                     # independent
d  <- data.frame(x1 = x1, x2 = x2, x3 = x3, x4 = x4)

janusplot(d)

## ----asym---------------------------------------------------------------------
n <- 400
x <- runif(n, 0, 5)
y <- 0.5 * x + rnorm(n, sd = 0.3 + 0.4 * x)   # variance grows with x
d <- data.frame(x = x, y = y, z = rnorm(n))

janusplot(d)

## ----adjust, eval = requireNamespace("palmerpenguins", quietly = TRUE)--------
library(palmerpenguins)
pp <- na.omit(penguins)

# Without covariate
janusplot(pp[, c("bill_length_mm", "bill_depth_mm",
                 "flipper_length_mm", "body_mass_g")])

# With species as a fixed effect — resolves Simpson's-paradox geometry
janusplot(pp, vars = c("bill_length_mm", "bill_depth_mm",
                       "flipper_length_mm", "body_mass_g"),
         adjust = ~ species)

## ----palette-viridis, fig.height = 5, fig.width = 6---------------------------
d <- data.frame(
  x1 = runif(200, -3, 3),
  x2 = rnorm(200),
  x3 = rnorm(200)
)
d$x2 <- d$x1^2 + rnorm(200, sd = 0.8)  # non-linear

janusplot(d, palette = "viridis")  # default, colourblind-safe

## ----palette-brewer, fig.height = 5, fig.width = 6----------------------------
janusplot(d, palette = "RdYlBu")   # diverging, colourblind-safe

## ----palette-turbo, fig.height = 5, fig.width = 6-----------------------------
janusplot(d, palette = "turbo")    # high-contrast, NOT colourblind-safe

## ----missing------------------------------------------------------------------
# airquality has genuine NAs in Ozone and Solar.R
janusplot(airquality[, c("Ozone", "Solar.R", "Wind", "Temp")],
          na_action = "pairwise")

## ----hclust, eval = requireNamespace("MASS", quietly = TRUE)------------------
data(Boston, package = "MASS")
janusplot(Boston[, c("medv", "lstat", "rm", "age",
                     "indus", "nox", "dis")],
          order = "hclust")

## ----data---------------------------------------------------------------------
# Re-create the heteroscedastic example
n <- 400
het <- data.frame(
  x = runif(n, 0, 5),
  y = NA_real_
)
het$y <- 0.5 * het$x + rnorm(n, sd = 0.3 + 0.4 * het$x)

out <- janusplot_data(het, vars = c("x", "y"))
out$pairs[[1L]]$edf_yx
out$pairs[[1L]]$edf_xy
out$pairs[[1L]]$asymmetry_index

## ----derivs-fit, fig.height = 4.5---------------------------------------------
set.seed(2026L)
n  <- 300L
xs <- runif(n, -pi, pi)
df <- data.frame(
  x  = xs,
  y1 = xs + sin(3 * xs) + rnorm(n, sd = 0.15),
  y2 = 0.5 * xs^2       + rnorm(n, sd = 0.8)
)
janusplot(df, display = "fit", show_shape_legend = FALSE)

## ----derivs-d1, fig.height = 4.5----------------------------------------------
janusplot(df, display = "d1", show_shape_legend = FALSE)

## ----derivs-d2, fig.height = 4.5----------------------------------------------
janusplot(df, display = "d2", show_shape_legend = FALSE)

## ----derivs-sim, fig.height = 4.5---------------------------------------------
janusplot(df, display = "d1",
          derivative_ci = "simultaneous",
          derivative_ci_nsim = 2000L,
          show_shape_legend = FALSE)

## ----cite---------------------------------------------------------------------
citation("janusplot")

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.align = "center",
  fig.width = 8,
  fig.height = 6,
  dpi       = 110
)

## ----setup-pkg----------------------------------------------------------------
library(janusplot)
library(ggplot2)

janusplot_shape_sensitivity_shapes()

## ----demo-data----------------------------------------------------------------
data("shape_sensitivity_demo")
str(shape_sensitivity_demo, vec.len = 2)

## ----recovery-curves----------------------------------------------------------
janusplot_shape_sensitivity_plot(shape_sensitivity_demo,
                                 "recovery_curves")

## ----archetype-confusion, fig.width = 6, fig.height = 5-----------------------
janusplot_shape_sensitivity_plot(shape_sensitivity_demo,
                                 "confusion_archetype")

## ----accuracy-grid------------------------------------------------------------
janusplot_shape_sensitivity_plot(shape_sensitivity_demo,
                                 "accuracy_grid")

## ----summary------------------------------------------------------------------
head(janusplot_shape_sensitivity_summary(shape_sensitivity_demo,
                                         level = "archetype"), 10)

## ----full-sweep, eval = FALSE-------------------------------------------------
# # Configure parallel execution (optional) — you control the plan.
# future::plan(future::multisession, workers = 4L)
# 
# res <- janusplot_shape_sensitivity(parallel = TRUE)
# 
# # Save for your paper
# saveRDS(res, "shape_sensitivity_full.rds")
# janusplot_shape_sensitivity_plot(res, "recovery_curves")

## ----custom-subset, eval = FALSE----------------------------------------------
# strict <- janusplot_shape_cutoffs(mono_strong = 0.95, curv_low = 0.1)
# 
# res_strict <- janusplot_shape_sensitivity(
#   shapes     = c("wave", "bimodal", "bi_wave"),
#   n_grid     = c(200L, 500L),
#   sigma_grid = c(0.05, 0.10, 0.20),
#   n_rep      = 100L,
#   cutoffs    = strict
# )
# 
# janusplot_shape_sensitivity_summary(res_strict, level = "fine")

## ----session-info-------------------------------------------------------------
sessionInfo()

# Shared synthetic datasets for tests.
# Pinned seed + simple generators keep expected statistics stable.

make_linear_data <- function(n = 200L, seed = 1L) {
  withr::with_seed(seed, {
    x1 <- stats::rnorm(n)
    x2 <- 1 + 0.8 * x1 + stats::rnorm(n, sd = 0.5)
    x3 <- stats::rnorm(n)
    x4 <- 0.3 * x1 - 0.2 * x3 + stats::rnorm(n, sd = 0.4)
    data.frame(x1 = x1, x2 = x2, x3 = x3, x4 = x4)
  })
}

make_nonlinear_data <- function(n = 200L, seed = 2L) {
  withr::with_seed(seed, {
    x1 <- stats::runif(n, -3, 3)
    x2 <- x1^2 + stats::rnorm(n, sd = 0.6)
    x3 <- sin(x1) + stats::rnorm(n, sd = 0.4)
    x4 <- stats::rnorm(n)
    data.frame(x1 = x1, x2 = x2, x3 = x3, x4 = x4)
  })
}

make_heteroscedastic_data <- function(n = 300L, seed = 3L) {
  withr::with_seed(seed, {
    x1 <- stats::runif(n, 0, 5)
    x2 <- 0.5 * x1 + stats::rnorm(n, sd = 0.3 + 0.4 * x1)
    x3 <- stats::rnorm(n)
    data.frame(x1 = x1, x2 = x2, x3 = x3)
  })
}

test_that("janusplot returns a ggplot", {
  d <- make_linear_data(n = 100L)
  p <- janusplot(d, vars = c("x1", "x2", "x3"))
  expect_s3_class(p, "ggplot")
})

test_that("labels = 'border' (default) produces a ggplot", {
  d <- make_linear_data(n = 80L)
  p <- janusplot(d, vars = c("x1", "x2", "x3"))
  expect_s3_class(p, "ggplot")
})

test_that("labels = 'diagonal' preserves the legacy in-matrix layout", {
  d <- make_linear_data(n = 80L)
  p <- janusplot(d, vars = c("x1", "x2", "x3"), labels = "diagonal")
  expect_s3_class(p, "ggplot")
})

test_that("labels = 'none' suppresses variable names", {
  d <- make_linear_data(n = 80L)
  p <- janusplot(d, vars = c("x1", "x2", "x3"), labels = "none")
  expect_s3_class(p, "ggplot")
})

test_that("labels accepts label_srt = 90 and label_srt = 0", {
  d <- make_linear_data(n = 80L)
  expect_s3_class(
    janusplot(d, vars = c("x1", "x2"), label_srt = 90),
    "ggplot"
  )
  expect_s3_class(
    janusplot(d, vars = c("x1", "x2"), label_srt = 0),
    "ggplot"
  )
})

test_that("labels rejects unknown value and non-numeric label_srt", {
  d <- make_linear_data(n = 80L)
  expect_error(
    janusplot(d, vars = c("x1", "x2"), labels = "bogus"),
    class = "rlang_error"
  )
  expect_error(
    janusplot(d, vars = c("x1", "x2"), label_srt = "45"),
    regexp = "numeric"
  )
  expect_error(
    janusplot(d, vars = c("x1", "x2"), label_cex = -1),
    regexp = "positive|> 0"
  )
})

test_that("janusplot respects order = 'alphabetical'", {
  d <- make_linear_data(n = 100L)
  names(d) <- c("zulu", "bravo", "alpha", "mike")
  p <- janusplot(d, order = "alphabetical")
  expect_s3_class(p, "ggplot")
})

test_that("janusplot respects order = 'hclust' with >= 3 vars", {
  d <- make_linear_data(n = 100L)
  p <- janusplot(d, order = "hclust")
  expect_s3_class(p, "ggplot")
})

test_that("janusplot runs with adjust formula", {
  d <- make_linear_data(n = 100L)
  d$z <- stats::rnorm(100L)
  p <- janusplot(d, vars = c("x1", "x2"), adjust = ~ s(z))
  expect_s3_class(p, "ggplot")
})

test_that("janusplot runs with colour_by = 'none'", {
  d <- make_linear_data(n = 100L)
  p <- janusplot(d, vars = c("x1", "x2"), colour_by = "none")
  expect_s3_class(p, "ggplot")
})

test_that("fill_by is a deprecated alias of colour_by (warns, still works)", {
  d <- make_linear_data(n = 100L)
  expect_warning(
    p <- janusplot(d, vars = c("x1", "x2"), fill_by = "none"),
    regexp = "deprecated"
  )
  expect_s3_class(p, "ggplot")
})

test_that("janusplot runs with show_data = FALSE and show_ci = FALSE", {
  d <- make_linear_data(n = 100L)
  p <- janusplot(d, vars = c("x1", "x2"),
                 show_data = FALSE, show_ci = FALSE)
  expect_s3_class(p, "ggplot")
})

test_that("annotations vector toggles A/edf/shape/code", {
  d <- make_nonlinear_data(n = 150L)
  expect_s3_class(
    janusplot(d, vars = c("x1", "x2", "x3"),
              annotations = c("A", "edf", "shape")),
    "ggplot"
  )
  expect_s3_class(
    janusplot(d, vars = c("x1", "x2", "x3"),
              annotations = c("A", "edf", "code")),
    "ggplot"
  )
  expect_s3_class(
    janusplot(d, vars = c("x1", "x2", "x3"), annotations = character()),
    "ggplot"
  )
})

test_that("annotations rejects unknown values", {
  d <- make_linear_data(n = 80L)
  expect_error(
    janusplot(d, annotations = c("edf", "bogus")),
    class = "rlang_error"
  )
})

test_that("show_asymmetry is a deprecated alias, forwards to annotations", {
  d <- make_nonlinear_data(n = 100L)
  expect_warning(
    p <- janusplot(d, vars = c("x1", "x2"), show_asymmetry = TRUE),
    regexp = "deprecated"
  )
  expect_s3_class(p, "ggplot")
})

test_that("janusplot accepts every advertised palette", {
  d <- make_linear_data(n = 80L)
  palettes <- c("viridis", "magma", "inferno", "plasma", "cividis",
                "mako", "rocket", "turbo",
                "RdYlBu", "RdBu", "PuOr", "Spectral",
                "YlOrRd", "YlGnBu", "Blues", "Greens")
  for (pal in palettes) {
    p <- janusplot(d, vars = c("x1", "x2"), palette = pal)
    expect_s3_class(p, "ggplot")
  }
})

test_that("janusplot rejects unknown palette", {
  d <- make_linear_data(n = 80L)
  expect_error(janusplot(d, palette = "not_a_palette"),
               class = "rlang_error")
})

test_that("janusplot accepts split text_scale_diag / text_scale_off_diag", {
  d <- make_linear_data(n = 80L)
  expect_s3_class(
    janusplot(d, vars = c("x1", "x2"),
              text_scale_diag = 0.5, text_scale_off_diag = 1.2),
    "ggplot"
  )
  expect_s3_class(
    janusplot(d, vars = c("x1", "x2"),
              text_scale_diag = 1.4, text_scale_off_diag = 0.7),
    "ggplot"
  )
})

test_that("janusplot text-scale args reject non-positive or invalid values", {
  d <- make_linear_data(n = 80L)
  expect_error(janusplot(d, text_scale_diag = 0), class = "rlang_error")
  expect_error(janusplot(d, text_scale_diag = -1), class = "rlang_error")
  expect_error(janusplot(d, text_scale_off_diag = "big"),
               class = "rlang_error")
  expect_error(janusplot(d, text_scale_off_diag = c(1, 2)),
               class = "rlang_error")
})

test_that(".check_parallel_plan informs on sequential plan", {
  skip_if_not_installed("future.apply")
  skip_if_not_installed("future")
  # .throttle = FALSE bypasses rlang's per-session message dedup so
  # the test is deterministic regardless of run order.
  withr::with_options(
    list(future.plan = "sequential"),
    expect_message(
      janusplot:::.check_parallel_plan(parallel = TRUE, .throttle = FALSE),
      regexp = "sequential"
    )
  )
})

test_that(".check_parallel_plan is silent when parallel = FALSE", {
  expect_silent(janusplot:::.check_parallel_plan(parallel = FALSE))
})

test_that("janusplot show_glossary = FALSE returns plot without caption", {
  d <- make_linear_data(n = 80L)
  p <- janusplot(d, vars = c("x1", "x2"), show_glossary = FALSE)
  expect_s3_class(p, "ggplot")
})

test_that("janusplot show_glossary validation", {
  d <- make_linear_data(n = 80L)
  expect_error(janusplot(d, show_glossary = "yes"), class = "rlang_error")
  expect_error(janusplot(d, show_glossary = NA), class = "rlang_error")
  expect_error(janusplot(d, show_glossary = c(TRUE, FALSE)),
               class = "rlang_error")
})

test_that("janusplot glossary_scale accepts positive multipliers", {
  d <- make_linear_data(n = 80L)
  expect_s3_class(janusplot(d, vars = c("x1", "x2"),
                            glossary_scale = 0.7), "ggplot")
  expect_s3_class(janusplot(d, vars = c("x1", "x2"),
                            glossary_scale = 1.5), "ggplot")
})

test_that("janusplot glossary_scale validation", {
  d <- make_linear_data(n = 80L)
  expect_error(janusplot(d, glossary_scale = 0), class = "rlang_error")
  expect_error(janusplot(d, glossary_scale = -1), class = "rlang_error")
  expect_error(janusplot(d, glossary_scale = "big"), class = "rlang_error")
})

test_that(".cell_text_sizes applies diag and off_diag scales independently", {
  base <- janusplot:::.cell_text_sizes(4)
  off2 <- janusplot:::.cell_text_sizes(4, text_scale_off_diag = 2)
  dia2 <- janusplot:::.cell_text_sizes(4, text_scale_diag = 2)
  # off_diag scale doubles n_edf / glyph but leaves diag untouched
  expect_equal(off2$n_edf, base$n_edf * 2)
  expect_equal(off2$glyph, base$glyph * 2)
  expect_equal(off2$diag,  base$diag)
  # diag scale doubles diag but leaves n_edf untouched
  expect_equal(dia2$diag,  base$diag * 2)
  expect_equal(dia2$n_edf, base$n_edf)
})

test_that("janusplot with_data = TRUE returns list(plot, data)", {
  d <- make_linear_data(n = 100L)
  out <- janusplot(d, vars = c("x1", "x2", "x3"), with_data = TRUE)
  expect_named(out, c("plot", "data"))
  expect_s3_class(out$plot, "ggplot")
  expect_equal(nrow(out$data), 6L)
  expect_true(all(c("var_x", "var_y", "position", "n_used",
                    "edf", "pvalue", "signif", "dev_exp",
                    "asymmetry_index",
                    "cor_pearson", "cor_spearman", "cor_kendall",
                    "tie_ratio",
                    "monotonicity_index", "convexity_index",
                    "n_turning_points", "n_inflections",
                    "flat_range_ratio", "shape_category",
                    "shape_code", "shape_archetype",
                    "shape_monotonic", "shape_linear",
                    "colour_value") %in%
                    names(out$data)))
  expect_true(all(out$data$position %in% c("upper", "lower")))
})

skip_on_cran()  # heavy GAM-fit suite: full on CI, skipped on CRAN for time budget

# Tests for Feature 2 — focus filter (companion to compact tiers).
# Covers: quantile-string thresholds, numeric cutoff, NA -> no-op,
# mask correctness against a synthetic asymmetry distribution.

skip_if_no_mgcv <- function() {
  testthat::skip_if_not_installed("mgcv")
}

# ---------------------------------------------------------------
# Mask resolver
# ---------------------------------------------------------------

test_that("focus_by = NA returns all-TRUE mask", {
  fits <- list(a = list(edf = 1), b = list(edf = 3))
  mask <- janusplot:::.resolve_focus_mask(fits, focus_by = NA)
  expect_true(all(mask))
  expect_length(mask, 2L)
})

test_that("focus_by = 'edf' with q50 keeps the top half", {
  fits <- mapply(function(e, k) list(edf = e, key = k),
                 c(1, 2, 3, 4),
                 c("1_2", "1_3", "2_3", "1_4"), SIMPLIFY = FALSE)
  fits <- lapply(seq_along(fits), function(i) {
    structure(fits[[i]], key = c("1_2", "1_3", "2_3", "1_4")[i])
  })
  mask <- janusplot:::.resolve_focus_mask(
    fits, focus_by = "edf", focus_threshold = "q50"
  )
  expect_true(sum(mask) >= 2L)
  # The two highest-edf cells must be in focus.
  expect_true(mask[3L] || mask[4L])
})

test_that("focus_by = 'edf' with numeric cutoff filters correctly", {
  fits <- lapply(c(1, 2, 5, 8), function(e) list(edf = e))
  mask <- janusplot:::.resolve_focus_mask(
    fits, focus_by = "edf", focus_threshold = 3
  )
  expect_equal(unname(mask), c(FALSE, FALSE, TRUE, TRUE))
})

test_that("invalid focus_threshold string errors loudly", {
  fits <- lapply(c(1, 2), function(e) list(edf = e))
  expect_error(
    janusplot:::.resolve_focus_mask(
      fits, focus_by = "edf", focus_threshold = "qninety"
    ),
    "quantile string"
  )
})

# ---------------------------------------------------------------
# End-to-end public-API surface
# ---------------------------------------------------------------

test_that("janusplot(focus_by = 'asymmetry') renders a ggplot", {
  skip_if_no_mgcv()
  dat <- make_heteroscedastic_data(n = 200L, seed = 3L)
  p <- janusplot(dat, focus_by = "asymmetry", focus_threshold = "q50")
  expect_s3_class(p, "ggplot")
})

test_that("janusplot(focus_by = 'edf') renders a ggplot", {
  skip_if_no_mgcv()
  dat <- make_nonlinear_data(n = 200L, seed = 2L)
  p <- janusplot(dat, focus_by = "edf", focus_threshold = "q80")
  expect_s3_class(p, "ggplot")
})

test_that("janusplot(focus_by = NA) is the default no-op", {
  skip_if_no_mgcv()
  dat <- make_linear_data(n = 100L, seed = 1L)
  p_default <- janusplot(dat, vars = c("x1", "x2", "x3"), with_data = TRUE)
  p_na      <- janusplot(dat, vars = c("x1", "x2", "x3"),
                         focus_by = NA, with_data = TRUE)
  # Cell content metrics are identical because focus is rendering-only.
  cols <- c("edf", "asymmetry_index", "pvalue", "cor_spearman")
  expect_identical(p_default$data[, cols], p_na$data[, cols])
})

test_that("invalid focus_dim_alpha errors loudly", {
  dat <- make_linear_data(60L, seed = 1L)
  expect_error(
    janusplot(dat, vars = c("x1", "x2"),
              focus_by = "edf", focus_dim_alpha = -0.1),
    "in \\[0, 1\\]"
  )
  expect_error(
    janusplot(dat, vars = c("x1", "x2"),
              focus_by = "edf", focus_dim_alpha = 2),
    "in \\[0, 1\\]"
  )
})

test_that("invalid focus_by value errors loudly", {
  dat <- make_linear_data(60L, seed = 1L)
  expect_error(
    janusplot(dat, vars = c("x1", "x2"), focus_by = "nonsense"),
    "Unknown"
  )
})

test_that("linear pair recovers EDF close to 1", {
  d <- make_linear_data(n = 200L)
  f <- janusplot:::.fit_pair(
    x_name = "x1", y_name = "x2", data_full = d,
    adjust = NULL, method = "REML", k = -1L, bs = "tp",
    na_action = "pairwise"
  )
  expect_lt(f$edf, 2)
  expect_gt(f$edf, 0.8)
  expect_equal(f$n_used, 200L)
  expect_true(is.finite(f$pvalue))
})

test_that("quadratic pair recovers EDF well above 1", {
  d <- make_nonlinear_data(n = 200L)
  f <- janusplot:::.fit_pair(
    x_name = "x1", y_name = "x2", data_full = d,
    adjust = NULL, method = "REML", k = -1L, bs = "tp",
    na_action = "pairwise"
  )
  expect_gt(f$edf, 2)
})

test_that(".fit_pair respects adjust covariates", {
  d <- make_linear_data(n = 200L)
  d$site <- factor(sample(letters[1:3], 200L, replace = TRUE))
  f <- janusplot:::.fit_pair(
    x_name = "x1", y_name = "x2", data_full = d,
    adjust = ~ s(site, bs = "re"),
    method = "REML", k = -1L, bs = "tp",
    na_action = "pairwise"
  )
  expect_true(is.finite(f$edf))
  expect_s3_class(f$fit, "gam")
  # fit's formula should include the random effect
  expect_true(grepl("site", deparse(stats::formula(f$fit))))
})

test_that(".fit_pair handles pairwise-complete missingness", {
  d <- make_linear_data(n = 200L)
  d$x2[1:20] <- NA
  f <- janusplot:::.fit_pair(
    x_name = "x1", y_name = "x2", data_full = d,
    adjust = NULL, method = "REML", k = -1L, bs = "tp",
    na_action = "pairwise"
  )
  expect_equal(f$n_used, 180L)
})

test_that("janusplot_data returns expected structure", {
  d <- make_linear_data(n = 100L)
  out <- janusplot_data(d, vars = c("x1", "x2", "x3"))
  expect_named(out, c("vars", "pairs", "call"))
  expect_equal(length(out$pairs), 3L)
  first <- out$pairs[[1L]]
  expect_true(all(c(
    "var_i", "var_j", "edf_yx", "edf_xy", "pvalue_yx", "pvalue_xy",
    "dev_exp_yx", "dev_exp_xy", "asymmetry_index", "n_used",
    "cor_pearson", "cor_spearman", "cor_kendall", "tie_ratio",
    "monotonicity_index_yx", "convexity_index_yx",
    "monotonicity_index_xy", "convexity_index_xy",
    "n_turning_yx", "n_inflect_yx", "n_turning_xy", "n_inflect_xy",
    "shape_yx", "shape_xy",
    "shape_code_yx", "shape_code_xy",
    "shape_archetype_yx", "shape_archetype_xy",
    "shape_monotonic_yx", "shape_monotonic_xy",
    "shape_linear_yx", "shape_linear_xy"
  ) %in% names(first)))
  expect_null(first$fit_yx)
})

test_that("janusplot_data correlations lie in [-1, 1] and agree with stats::cor", {
  d <- make_linear_data(n = 200L)
  out <- janusplot_data(d, vars = c("x1", "x2"))
  p <- out$pairs[[1L]]
  expect_true(is.finite(p$cor_spearman))
  expect_true(all(abs(c(p$cor_pearson, p$cor_spearman, p$cor_kendall)) <= 1))
  # Sanity: recompute Pearson on the raw columns and compare.
  ref <- stats::cor(d$x1, d$x2, use = "pairwise.complete.obs")
  expect_equal(p$cor_pearson, ref, tolerance = 1e-10)
})

test_that("janusplot_data shape labels recover known canonical shapes", {
  # Synthetic pair: y = x (linear_up). Spearman ≈ 1.
  # mgcv REML smoothing on noisy linear data may yield a near-linear
  # spline with tiny wiggles, so allow any monotone-increasing family.
  withr::with_seed(42L, {
    n  <- 200L
    x1 <- stats::runif(n, 0, 10)
    x2 <- x1 + stats::rnorm(n, sd = 0.1)
    d  <- data.frame(x1 = x1, x2 = x2)
  })
  out <- janusplot_data(d, vars = c("x1", "x2"))
  p <- out$pairs[[1L]]
  expect_true(p$shape_yx %in%
                c("linear_up", "convex_up", "concave_up", "s_shape"))
  expect_gt(p$cor_spearman, 0.9)
})

test_that("janusplot_data keep_fits = TRUE retains gam objects", {
  d <- make_linear_data(n = 100L)
  out <- janusplot_data(d, vars = c("x1", "x2"), keep_fits = TRUE)
  expect_s3_class(out$pairs[[1L]]$fit_yx, "gam")
  expect_s3_class(out$pairs[[1L]]$fit_xy, "gam")
})

test_that("janusplot_data asymmetry_index is in [0, 1]", {
  d <- make_nonlinear_data(n = 200L)
  out <- janusplot_data(d)
  ax <- vapply(out$pairs, function(p) p$asymmetry_index, numeric(1L))
  ax <- ax[!is.na(ax)]
  expect_true(all(ax >= 0 & ax <= 1))
})

test_that("heteroscedastic data yields non-trivial asymmetry for x1-x2", {
  d <- make_heteroscedastic_data(n = 300L)
  out <- janusplot_data(d, vars = c("x1", "x2"))
  expect_true(is.finite(out$pairs[[1L]]$asymmetry_index))
})

skip_on_cran()  # heavy GAM-fit suite: full on CI, skipped on CRAN for time budget

# Tests for Feature 2 — scale-aware compact rendering.
# Covers: tier resolver behaviour at thresholds; compact = "always" /
# "never" overrides; backward-compat at k < 12; compact_levels
# overrides; public API validation.

# ---------------------------------------------------------------
# Tier resolver
# ---------------------------------------------------------------

test_that(".compact_tier returns 0 below default threshold", {
  expect_equal(janusplot:::.compact_tier(5L), 0L)
  expect_equal(janusplot:::.compact_tier(11L), 0L)
})

test_that(".compact_tier returns 1 at default tier-1 threshold", {
  expect_equal(janusplot:::.compact_tier(12L), 1L)
  expect_equal(janusplot:::.compact_tier(17L), 1L)
})

test_that(".compact_tier returns 2 at default tier-2 threshold", {
  expect_equal(janusplot:::.compact_tier(18L), 2L)
  expect_equal(janusplot:::.compact_tier(24L), 2L)
})

test_that(".compact_tier returns 3 at default tier-3 threshold", {
  expect_equal(janusplot:::.compact_tier(25L), 3L)
  expect_equal(janusplot:::.compact_tier(40L), 3L)
})

test_that("compact = 'never' forces tier 0 regardless of n_var", {
  expect_equal(janusplot:::.compact_tier(50L, compact = "never"), 0L)
})

test_that("compact = 'always' forces tier >= 1", {
  expect_equal(janusplot:::.compact_tier(3L, compact = "always"), 1L)
  expect_equal(janusplot:::.compact_tier(20L, compact = "always"), 2L)
  expect_equal(janusplot:::.compact_tier(30L, compact = "always"), 3L)
})

test_that("compact_threshold shifts the entire ladder", {
  expect_equal(
    janusplot:::.compact_tier(10L, compact_threshold = 15L),
    0L
  )
  expect_equal(
    janusplot:::.compact_tier(15L, compact_threshold = 15L),
    1L
  )
  expect_equal(
    janusplot:::.compact_tier(21L, compact_threshold = 15L),
    2L
  )
})

test_that("compact_levels override is honoured", {
  custom <- list(t1 = 8L, t2 = 16L, t3 = 24L)
  expect_equal(janusplot:::.compact_tier(7L,  compact_levels = custom), 0L)
  expect_equal(janusplot:::.compact_tier(8L,  compact_levels = custom), 1L)
  expect_equal(janusplot:::.compact_tier(16L, compact_levels = custom), 2L)
  expect_equal(janusplot:::.compact_tier(24L, compact_levels = custom), 3L)
})

test_that("invalid compact_levels errors loudly", {
  expect_error(
    janusplot:::.compact_tier(20L,
      compact_levels = list(t1 = 10L, t2 = 8L, t3 = 20L)),
    "t1 < t2 < t3"
  )
  expect_error(
    janusplot:::.compact_tier(20L,
      compact_levels = list(t1 = 1L, t2 = 5L, t3 = 10L)),
    "single integer >= 2"
  )
  expect_error(
    janusplot:::.compact_tier(20L,
      compact_levels = list(t1 = 5L, t2 = 10L)),
    "missing entries"
  )
})

# ---------------------------------------------------------------
# End-to-end: each tier renders without error
# ---------------------------------------------------------------

skip_if_no_mgcv <- function() {
  testthat::skip_if_not_installed("mgcv")
}

make_wide_data <- function(k_var, n = 80L, seed = 42L) {
  withr::with_seed(seed, {
    as.data.frame(matrix(stats::rnorm(n * k_var), ncol = k_var,
                         dimnames = list(NULL, sprintf("v%02d", seq_len(k_var)))))
  })
}

test_that("tier 0 (k = 5) renders a ggplot", {
  skip_if_no_mgcv()
  p <- janusplot(make_wide_data(5L))
  expect_s3_class(p, "ggplot")
})

test_that("tier 1 (k = 14) renders a ggplot", {
  skip_if_no_mgcv()
  p <- janusplot(make_wide_data(14L))
  expect_s3_class(p, "ggplot")
})

test_that("tier 2 (k = 19) renders a ggplot", {
  skip_if_no_mgcv()
  p <- janusplot(make_wide_data(19L))
  expect_s3_class(p, "ggplot")
})

test_that("tier 3 (k = 26) renders a ggplot", {
  skip_if_no_mgcv()
  p <- janusplot(make_wide_data(26L))
  expect_s3_class(p, "ggplot")
})

# ---------------------------------------------------------------
# Backward compatibility: at k < 12, output must be unchanged
# whether compact is "auto", "always" (forces tier 1) is OFF, or
# "never". Use the data summary as a proxy — fits + metrics
# computed are identical across compact settings.
# ---------------------------------------------------------------

test_that("compact setting does not affect fits / metrics (rendering-only knob)", {
  skip_if_no_mgcv()
  dat <- make_linear_data(n = 150L, seed = 1L)
  d_auto  <- janusplot(dat, vars = c("x1", "x2", "x3"),
                       compact = "auto", with_data = TRUE)$data
  d_always <- janusplot(dat, vars = c("x1", "x2", "x3"),
                        compact = "always", with_data = TRUE)$data
  d_never  <- janusplot(dat, vars = c("x1", "x2", "x3"),
                        compact = "never", with_data = TRUE)$data
  # Strip rendering-only columns; compare the fit-side metrics.
  cols <- c("edf", "asymmetry_index", "pvalue", "k_prime", "k_index",
            "cor_spearman", "shape_category", "monotonicity_index",
            "convexity_index")
  expect_identical(d_auto[, cols], d_always[, cols])
  expect_identical(d_auto[, cols], d_never[, cols])
})

# ---------------------------------------------------------------
# Public API validation
# ---------------------------------------------------------------

test_that("invalid compact value errors via arg_match", {
  expect_error(
    janusplot(make_linear_data(60L), compact = "sometimes"),
    "must be one of"
  )
})

test_that("invalid compact_threshold errors loudly", {
  expect_error(
    janusplot(make_linear_data(60L), compact_threshold = 0L),
    "integer >= 2"
  )
  expect_error(
    janusplot(make_linear_data(60L), compact_threshold = "twelve"),
    "integer >= 2"
  )
})

# Visual regression tests via vdiffr.
# Snapshots live under tests/testthat/_snaps/vdiffr/. After visual
# redesign (2026-04-21) old snapshots were retired. Regenerate with
# `testthat::snapshot_accept("vdiffr")` on a trusted machine.

skip_if_not_installed("vdiffr")

test_that("diagonal cell matches snapshot", {
  p <- janusplot:::.build_diagonal_cell("my_var")
  vdiffr::expect_doppelganger("diagonal-cell-label", p)
})

test_that("linear 3-var matrix (Spearman colour) matches snapshot", {
  d <- make_linear_data(n = 150L, seed = 1L)
  p <- janusplot(d, vars = c("x1", "x2", "x3"))
  vdiffr::expect_doppelganger("matrix-3var-linear-spearman", p)
})

test_that("non-linear 3-var matrix with shape glyphs matches snapshot", {
  d <- make_nonlinear_data(n = 200L, seed = 2L)
  p <- janusplot(d, vars = c("x1", "x2", "x3"),
                 annotations = c("edf", "A", "shape"))
  vdiffr::expect_doppelganger("matrix-3var-nonlinear-shape", p)
})

test_that("EDF-coloured matrix (legacy encoding) matches snapshot", {
  d <- make_nonlinear_data(n = 200L, seed = 2L)
  p <- janusplot(d, vars = c("x1", "x2", "x3"), colour_by = "edf")
  vdiffr::expect_doppelganger("matrix-3var-nonlinear-edf-colour", p)
})

test_that("colour_by = 'none' matrix matches snapshot", {
  d <- make_linear_data(n = 150L, seed = 1L)
  p <- janusplot(d, vars = c("x1", "x2"), colour_by = "none")
  vdiffr::expect_doppelganger("matrix-colour-none", p)
})

test_that("ASCII glyph fallback matches snapshot", {
  d <- make_nonlinear_data(n = 150L, seed = 2L)
  p <- janusplot(d, vars = c("x1", "x2", "x3"),
                 glyph_style = "ascii")
  vdiffr::expect_doppelganger("matrix-3var-nonlinear-ascii-glyphs", p)
})

skip_on_cran()  # heavy GAM-fit suite: full on CI, skipped on CRAN for time budget

# Tests for Feature 3 — axes rendering modes.
# Covers: byte-identity of fits across modes; label-suffix correctness;
# transformation correctness for each mode; tier-3 no-op; at-k
# rendering across all 4 modes for k in {2, 5, 10, 15, 20, 26}.

skip_if_no_mgcv <- function() {
  testthat::skip_if_not_installed("mgcv")
}

# ---------------------------------------------------------------
# .build_axis_transform unit tests — pure helper
# ---------------------------------------------------------------

test_that(".build_axis_transform 'original' is identity", {
  raw <- c(1.5, 2.5, 7.0, 11.0)
  tf <- janusplot:::.build_axis_transform(raw, "original")
  expect_equal(tf$fn(raw), raw)
  expect_equal(tf$fn(c(NA, 1.5, 99)), c(NA, 1.5, 99))
  expect_identical(tf$suffix, "")
})

test_that(".build_axis_transform 'standardised' centres + scales", {
  raw <- c(1, 2, 3, 4, 5)
  tf <- janusplot:::.build_axis_transform(raw, "standardised")
  out <- tf$fn(raw)
  expect_equal(mean(out), 0, tolerance = 1e-9)
  expect_equal(stats::sd(out), 1, tolerance = 1e-9)
  expect_identical(tf$suffix, " (z)")
})

test_that(".build_axis_transform 'centred' subtracts the mean", {
  raw <- c(10, 20, 30)
  tf <- janusplot:::.build_axis_transform(raw, "centred")
  expect_equal(tf$fn(raw), c(-10, 0, 10))
  expect_identical(tf$suffix, " (centred)")
})

test_that(".build_axis_transform 'rank' returns ecdf * n, monotone", {
  raw <- c(1, 2, 5, 11, 50)
  tf <- janusplot:::.build_axis_transform(raw, "rank")
  out <- tf$fn(raw)
  # ecdf * n yields strictly increasing values for strictly
  # increasing inputs without ties.
  expect_true(all(diff(out) > 0))
  expect_true(all(out >= 0 & out <= length(raw)))
  expect_identical(tf$suffix, " (rank)")
})

test_that(".build_axis_transform handles degenerate inputs without erroring", {
  # All-NA -> identity fallback
  tf_na <- janusplot:::.build_axis_transform(rep(NA_real_, 5), "standardised")
  expect_equal(tf_na$fn(c(1, 2, 3)), c(1, 2, 3))
  # Single value -> identity fallback for rank
  tf_one <- janusplot:::.build_axis_transform(c(7), "rank")
  expect_equal(tf_one$fn(c(7, 14)), c(7, 14))
  # Constant column -> sd defaults to 1 for standardised
  tf_const <- janusplot:::.build_axis_transform(c(5, 5, 5), "standardised")
  expect_equal(tf_const$fn(c(5, 5, 5)), c(0, 0, 0))
})

# ---------------------------------------------------------------
# Label suffixing
# ---------------------------------------------------------------

test_that(".label_with_suffix handles every mode correctly", {
  expect_identical(janusplot:::.label_with_suffix("mpg", "", "original"), "mpg")
  expect_identical(janusplot:::.label_with_suffix("mpg", " (z)", "standardised"),
                   "mpg (z)")
  expect_identical(janusplot:::.label_with_suffix("mpg", " (centred)", "centred"),
                   "mpg (centred)")
  expect_identical(janusplot:::.label_with_suffix("mpg", " (rank)", "rank"),
                   "rank(mpg)")
})

# ---------------------------------------------------------------
# Byte-identity of fits across modes — the load-bearing invariant
# ---------------------------------------------------------------

test_that("fits + metrics are byte-identical across axes modes", {
  skip_if_no_mgcv()
  dat <- make_nonlinear_data(n = 200L, seed = 2L)
  d_orig <- janusplot(dat, axes = "original",     with_data = TRUE)$data
  d_std  <- janusplot(dat, axes = "standardised", with_data = TRUE)$data
  d_ctr  <- janusplot(dat, axes = "centred",      with_data = TRUE)$data
  d_rnk  <- janusplot(dat, axes = "rank",         with_data = TRUE)$data
  # The fit-side columns must be identical; rendering-only knob.
  cols <- c("edf", "pvalue", "asymmetry_index", "dev_exp",
            "cor_pearson", "cor_spearman", "cor_kendall", "tie_ratio",
            "monotonicity_index", "convexity_index",
            "n_turning_points", "n_inflections", "shape_category",
            "k_prime", "k_index", "k_p", "k_flag")
  expect_identical(d_orig[, cols], d_std[, cols])
  expect_identical(d_orig[, cols], d_ctr[, cols])
  expect_identical(d_orig[, cols], d_rnk[, cols])
})

# ---------------------------------------------------------------
# At-k correctness: render at k = 2, 5, 10, 15, 20, 26 across all
# 4 axes modes (24 combinations). Each must build a ggplot
# without errors or warnings. Labels must reflect the chosen mode.
# ---------------------------------------------------------------

make_wide_data <- function(k_var, n = 80L, seed = 42L) {
  withr::with_seed(seed, {
    as.data.frame(matrix(stats::rnorm(n * k_var), ncol = k_var,
                         dimnames = list(NULL, sprintf("v%02d", seq_len(k_var)))))
  })
}

test_that("every (k, axes) combination renders without warnings", {
  skip_if_no_mgcv()
  k_values <- c(2L, 5L, 10L, 15L, 20L, 26L)
  modes    <- c("original", "standardised", "centred", "rank")
  for (k in k_values) {
    dat <- make_wide_data(k)
    for (m in modes) {
      p <- expect_silent(janusplot(dat, axes = m))
      expect_s3_class(p, "ggplot")
    }
  }
})

collect_label_strings <- function(g) {
  # Walk patchwork's plot list looking for ggplot layer labels.
  plots <- if (!is.null(g$patches$plots)) g$patches$plots else list(g)
  out <- character()
  for (p in plots) {
    if (inherits(p, "patchwork")) {
      out <- c(out, collect_label_strings(p))
      next
    }
    if (!inherits(p, "ggplot") || !length(p$layers)) next
    for (lyr in p$layers) {
      lbl <- lyr$aes_params$label %||% lyr$data$label
      if (!is.null(lbl)) out <- c(out, as.character(lbl))
    }
  }
  out
}

test_that("border labels carry the mode suffix at every k", {
  skip_if_no_mgcv()
  for (k in c(3L, 12L, 20L)) {
    dat <- make_wide_data(k)
    g_std  <- janusplot(dat, axes = "standardised", labels = "border")
    g_rnk  <- janusplot(dat, axes = "rank",         labels = "border")
    g_orig <- janusplot(dat, axes = "original",     labels = "border")
    std_labels  <- collect_label_strings(g_std)
    rnk_labels  <- collect_label_strings(g_rnk)
    orig_labels <- collect_label_strings(g_orig)
    expect_true(any(grepl(" \\(z\\)$", std_labels)),
                info = sprintf("k = %d standardised", k))
    expect_true(any(grepl("^rank\\(", rnk_labels)),
                info = sprintf("k = %d rank", k))
    # Original mode: no mode-suffix on label strings.
    expect_false(any(grepl(" \\(z\\)$|^rank\\(|\\(centred\\)$", orig_labels)),
                 info = sprintf("k = %d original", k))
  }
})

# ---------------------------------------------------------------
# Tier-3 no-op: at k >= 25 the cells don't render the curve, so
# `axes` is documented no-op. Calling it must NOT error.
# ---------------------------------------------------------------

test_that("tier 3 (k = 26) accepts every axes mode silently", {
  skip_if_no_mgcv()
  dat <- make_wide_data(26L)
  for (m in c("original", "standardised", "centred", "rank")) {
    p <- janusplot(dat, axes = m)
    expect_s3_class(p, "ggplot")
  }
})

# ---------------------------------------------------------------
# Public API validation
# ---------------------------------------------------------------

test_that("invalid axes value errors via arg_match", {
  skip_if_no_mgcv()
  dat <- make_linear_data(50L, seed = 1L)
  expect_error(
    janusplot(dat, vars = c("x1", "x2"), axes = "nonsense"),
    "must be one of"
  )
})

# ---------------------------------------------------------------
# save_as file output
# ---------------------------------------------------------------

test_that("save_as writes PNG / PDF with device from extension", {
  skip_if_no_mgcv()
  testthat::skip_if_not_installed("ggplot2")
  dat <- make_linear_data(60L, seed = 1L)
  tmpdir <- withr::local_tempdir()
  # PNG + PDF use base R grDevices — always available. SVG needs
  # svglite (ggplot2 dispatches to svglite::svglite for .svg), so
  # it gets its own conditional-skip test below.
  for (ext in c("png", "pdf")) {
    path <- file.path(tmpdir, paste0("matrix.", ext))
    p <- janusplot(dat, vars = c("x1", "x2", "x3"), save_as = path)
    expect_true(file.exists(path))
    expect_true(file.size(path) > 100)
    expect_s3_class(p, "ggplot")
  }
})

test_that("save_as writes SVG when svglite is available", {
  skip_if_no_mgcv()
  testthat::skip_if_not_installed("svglite")
  dat <- make_linear_data(60L, seed = 1L)
  tmpdir <- withr::local_tempdir()
  path <- file.path(tmpdir, "matrix.svg")
  p <- janusplot(dat, vars = c("x1", "x2", "x3"), save_as = path)
  expect_true(file.exists(path))
  expect_true(file.size(path) > 100)
  expect_s3_class(p, "ggplot")
})

test_that("save_as rejects unknown extensions loudly", {
  skip_if_no_mgcv()
  dat <- make_linear_data(60L, seed = 1L)
  expect_error(
    janusplot(dat, vars = c("x1", "x2"), save_as = "matrix.xyz"),
    "Cannot infer image device"
  )
  expect_error(
    janusplot(dat, vars = c("x1", "x2"), save_as = "matrix_no_extension"),
    "Cannot infer image device"
  )
})

test_that("save_as honours width / height / dpi overrides", {
  skip_if_no_mgcv()
  dat <- make_linear_data(60L, seed = 1L)
  tmpdir <- withr::local_tempdir()
  path <- file.path(tmpdir, "matrix.png")
  janusplot(dat, vars = c("x1", "x2", "x3"),
            save_as = path, save_width = 4, save_height = 4, save_dpi = 96)
  expect_true(file.exists(path))
  # 4x4 inches at 96 dpi -> 384 x 384 px; allow some slack since
  # patchwork may add small margins.
  expect_true(file.size(path) > 100)
})

skip_on_cran()  # heavy GAM-fit suite: full on CI, skipped on CRAN for time budget

# Tests for Feature 4 — fitting-engine dispatch (bam vs gam).
# Covers: both engines produce ggplot + with_data; both agree on
# linear DGP within EDF tolerance; engine + method columns surface
# in with_data; public API validation; arg_match guard.

skip_if_no_mgcv <- function() {
  testthat::skip_if_not_installed("mgcv")
}

# ---------------------------------------------------------------
# Both engines: end-to-end smoke
# ---------------------------------------------------------------

test_that("engine = 'bam' (default) produces ggplot + with_data", {
  skip_if_no_mgcv()
  dat <- make_linear_data(150L, seed = 1L)
  out <- janusplot(dat, vars = c("x1", "x2", "x3"), with_data = TRUE)
  expect_s3_class(out$plot, "ggplot")
  expect_true(all(out$data$engine == "bam"))
  expect_true(all(out$data$method == "fREML"))
})

test_that("engine = 'gam' produces ggplot + with_data with REML method", {
  skip_if_no_mgcv()
  dat <- make_linear_data(150L, seed = 1L)
  out <- janusplot(dat, vars = c("x1", "x2", "x3"),
                   engine = "gam", with_data = TRUE)
  expect_s3_class(out$plot, "ggplot")
  expect_true(all(out$data$engine == "gam"))
  expect_true(all(out$data$method == "REML"))
})

# ---------------------------------------------------------------
# Engines AGREE on a linear DGP: EDF should be ~1 for both.
# fREML and REML differ by ~1-3% in EDF on the same data; on a
# linear DGP this is well below the 0.5 tolerance.
# ---------------------------------------------------------------

test_that("both engines agree closely (<10% relative shift) on linear DGP", {
  skip_if_no_mgcv()
  dat <- make_linear_data(200L, seed = 1L)
  bam_data <- janusplot(dat, vars = c("x1", "x2"),
                        engine = "bam", with_data = TRUE)$data
  gam_data <- janusplot(dat, vars = c("x1", "x2"),
                        engine = "gam", with_data = TRUE)$data
  # fREML (bam) vs REML (gam) differ by a few percent on identical
  # data — this is the documented v0.1.0 -> v0.1.1 numerical shift.
  # Sanity check: relative shift <10%.
  rel_diff <- abs(bam_data$edf - gam_data$edf) /
    pmax(0.5, gam_data$edf)
  expect_true(all(rel_diff < 0.10),
              info = sprintf("max rel diff = %.3f", max(rel_diff)))
})

# ---------------------------------------------------------------
# Engines AGREE on a non-linear DGP: both detect non-linearity.
# Numerical EDF values may differ by 1-3%; sign of the test
# (EDF > 1) must agree.
# ---------------------------------------------------------------

test_that("both engines detect non-linearity on a quadratic DGP (forward direction)", {
  skip_if_no_mgcv()
  dat <- make_nonlinear_data(200L, seed = 2L)
  bam_d <- janusplot(dat, vars = c("x1", "x2"),
                     engine = "bam", with_data = TRUE)$data
  gam_d <- janusplot(dat, vars = c("x1", "x2"),
                     engine = "gam", with_data = TRUE)$data
  # Forward direction: x2 ~ s(x1) on x2 = x1^2 + noise. Both
  # engines should easily land EDF > 1.5 in this direction.
  forward_bam <- bam_d[bam_d$var_x == "x1" & bam_d$var_y == "x2", ]
  forward_gam <- gam_d[gam_d$var_x == "x1" & gam_d$var_y == "x2", ]
  expect_true(forward_bam$edf > 1.5)
  expect_true(forward_gam$edf > 1.5)
  # The EDF values from the two engines should agree closely
  # (documented ~1-3% shift; we accept <10% to be robust).
  rel_diff <- abs(forward_bam$edf - forward_gam$edf) / forward_gam$edf
  expect_lt(rel_diff, 0.10)
})

# ---------------------------------------------------------------
# discrete / nthreads only apply to bam.
# ---------------------------------------------------------------

test_that("discrete = TRUE on engine = bam still renders", {
  skip_if_no_mgcv()
  dat <- make_linear_data(200L, seed = 1L)
  p <- janusplot(dat, vars = c("x1", "x2", "x3"),
                 engine = "bam", discrete = TRUE)
  expect_s3_class(p, "ggplot")
})

test_that("discrete = TRUE on engine = gam is silently ignored (gam doesn't use it)", {
  skip_if_no_mgcv()
  dat <- make_linear_data(150L, seed = 1L)
  expect_silent(
    janusplot(dat, vars = c("x1", "x2", "x3"),
              engine = "gam", discrete = TRUE)
  )
})

# ---------------------------------------------------------------
# Backward-compat escape: engine = "gam" + method = "REML" must
# match v0.1.0 fitting backend exactly. We can't snapshot v0.1.0
# output from this session, but we CAN assert engine + method
# columns reflect the v0.1.0 defaults.
# ---------------------------------------------------------------

test_that("engine = 'gam' + default method = REML matches v0.1.0 stack", {
  skip_if_no_mgcv()
  dat <- make_linear_data(60L, seed = 1L)
  d <- janusplot(dat, vars = c("x1", "x2"),
                 engine = "gam", with_data = TRUE)$data
  expect_true(all(d$engine == "gam"))
  expect_true(all(d$method == "REML"))
})

# ---------------------------------------------------------------
# Public API validation
# ---------------------------------------------------------------

test_that("invalid engine errors via arg_match", {
  skip_if_no_mgcv()
  dat <- make_linear_data(50L, seed = 1L)
  expect_error(
    janusplot(dat, vars = c("x1", "x2"), engine = "torch"),
    "must be one of"
  )
})

test_that("invalid discrete / nthreads error loudly", {
  skip_if_no_mgcv()
  dat <- make_linear_data(50L, seed = 1L)
  expect_error(
    janusplot(dat, vars = c("x1", "x2"), discrete = NA),
    "TRUE or FALSE"
  )
  expect_error(
    janusplot(dat, vars = c("x1", "x2"), nthreads = 0),
    "positive integer"
  )
})

# ---------------------------------------------------------------
# User-supplied method override beats per-engine default.
# ---------------------------------------------------------------

test_that("explicit method overrides per-engine default", {
  skip_if_no_mgcv()
  dat <- make_linear_data(80L, seed = 1L)
  d <- janusplot(dat, vars = c("x1", "x2"),
                 engine = "gam", method = "GCV.Cp",
                 with_data = TRUE)$data
  expect_true(all(d$method == "GCV.Cp"))
})

# ---------------------------------------------------------------
# janusplot_data() carries engine + method on every pair.
# ---------------------------------------------------------------

test_that("janusplot_data carries engine + method per pair", {
  skip_if_no_mgcv()
  dat <- make_linear_data(80L, seed = 1L)
  res <- janusplot_data(dat, vars = c("x1", "x2", "x3"),
                        engine = "gam")
  for (p in res$pairs) {
    expect_identical(p$engine, "gam")
    expect_identical(p$method, "REML")
  }
})

# Internal helpers — NOT EXPORTED.
# Cell-layer logic: build one ggplot per matrix cell.

# ---------------------------------------------------------------
# Colour encodings.
#  * "spearman" / "pearson" / "kendall" — symmetric correlation
#     coefficient, diverging palette, limits c(-1, 1).
#  * "edf" / "deviance_gap" — non-linearity index, sequential palette.
#  * "none" — no fill.
# Supported values for colour_by:
.colour_choices <- function() {
  c("spearman", "pearson", "kendall", "edf", "deviance_gap", "none")
}

.is_correlation_colour <- function(colour_by) {
  colour_by %in% c("spearman", "pearson", "kendall")
}

.colour_value <- function(fit_obj, colour_by) {
  if (colour_by == "none") return(NA_real_)
  if (colour_by == "edf") return(fit_obj$edf)
  if (colour_by == "deviance_gap") return(fit_obj$dev_exp)
  corr <- fit_obj$corr
  if (is.null(corr)) return(NA_real_)
  switch(colour_by,
    spearman = corr$cor_spearman,
    pearson  = corr$cor_pearson,
    kendall  = corr$cor_kendall,
    NA_real_
  )
}

.colour_label <- function(colour_by) {
  switch(colour_by,
    edf           = "EDF",
    deviance_gap  = "Dev. expl.",
    spearman      = "corr",
    pearson       = "corr",
    kendall       = "corr",
    none          = NULL,
    NULL
  )
}

# Defaults that depend on the chosen colour_by.
.default_palette <- function(colour_by) {
  if (.is_correlation_colour(colour_by)) "RdBu" else "viridis"
}

# ---------------------------------------------------------------
# Supported palettes.
# Colour-blind safe by design:
#   - viridis family (viridis/magma/plasma/inferno/cividis/mako/rocket)
#   - ColorBrewer sequential CB-safe: YlOrRd, YlGnBu, Blues, Greens
#   - ColorBrewer diverging CB-safe: RdYlBu, RdBu, PuOr
# NOT colour-blind safe (kept for visual impact): turbo, Spectral
# ---------------------------------------------------------------

.palettes_viridis <- c(
  viridis = "D", magma = "A", inferno = "B", plasma = "C",
  cividis = "E", mako = "F", rocket = "G", turbo = "H"
)

.palettes_brewer <- c(
  "RdYlBu", "RdBu", "PuOr", "Spectral",
  "YlOrRd", "YlGnBu", "Blues", "Greens"
)

.palettes_diverging <- c("RdYlBu", "RdBu", "PuOr", "Spectral")

.palette_is_cb_safe <- function(palette) {
  cb_safe <- c(
    "viridis", "magma", "inferno", "plasma", "cividis", "mako", "rocket",
    "RdYlBu", "RdBu", "PuOr", "YlOrRd", "YlGnBu", "Blues", "Greens"
  )
  palette %in% cb_safe
}

.palette_choices <- function() {
  c(names(.palettes_viridis), .palettes_brewer)
}

# Endpoint colours for a diverging ColorBrewer palette, extracted via
# grDevices::colorRampPalette to avoid a hard dep on RColorBrewer.
.brewer_divergent_endpoints <- function(palette) {
  # Hand-picked 3-colour anchors for the 4 supported diverging palettes.
  # Values sourced from ColorBrewer 2.0 sequential levels; kept static
  # so we avoid runtime dependence on RColorBrewer.
  switch(palette,
    RdBu    = c(low = "#2166AC", mid = "#F7F7F7", high = "#B2182B"),
    RdYlBu  = c(low = "#313695", mid = "#FFFFBF", high = "#A50026"),
    PuOr    = c(low = "#542788", mid = "#F7F7F7", high = "#B35806"),
    Spectral = c(low = "#3288BD", mid = "#FFFFBF", high = "#D53E4F"),
    c(low = "#2166AC", mid = "#F7F7F7", high = "#B2182B")
  )
}

# Diverging symmetric scale for correlation colouring. RdBu (and peers)
# is reversed relative to brewer's default so negative = blue, positive
# = red, matching corrplot conventions.
.palette_scale_diverging <- function(palette, label, show_guide = FALSE) {
  g <- if (show_guide) {
    ggplot2::guide_colourbar(
      # barheight omitted; deferred to theme legend.key.height so the
      # bar flex-fills the available vertical space (corrplot-style).
      barwidth       = grid::unit(0.9, "lines"),
      title.position = "top",
      title.hjust    = 0.5,
      ticks.linewidth = 0.5,
      frame.linewidth = 0.3,
      frame.colour    = "grey50"
    )
  } else {
    "none"
  }
  eps <- .brewer_divergent_endpoints(palette)
  ggplot2::scale_fill_gradient2(
    low      = eps[["low"]],
    mid      = eps[["mid"]],
    high     = eps[["high"]],
    midpoint = 0,
    limits   = c(-1, 1),
    name     = label,
    na.value = "grey95",
    guide    = g
  )
}

.palette_scale <- function(palette, limits, label, colour_by,
                           show_guide = FALSE) {
  if (.is_correlation_colour(colour_by)) {
    # If user picked a sequential palette while colouring by correlation,
    # silently upgrade to the default diverging palette.
    if (!palette %in% .palettes_diverging) palette <- "RdBu"
    return(.palette_scale_diverging(palette, label, show_guide))
  }
  g <- if (show_guide) {
    ggplot2::guide_colourbar(
      # barheight omitted; deferred to theme legend.key.height so the
      # bar flex-fills the available vertical space (corrplot-style).
      barwidth       = grid::unit(0.9, "lines"),
      title.position = "top",
      title.hjust    = 0.5,
      ticks.linewidth = 0.5,
      frame.linewidth = 0.3,
      frame.colour    = "grey50"
    )
  } else {
    "none"
  }
  if (palette %in% names(.palettes_viridis)) {
    ggplot2::scale_fill_viridis_c(
      option   = .palettes_viridis[[palette]],
      limits   = limits,
      name     = label,
      na.value = "grey95",
      guide    = g
    )
  } else if (palette %in% .palettes_brewer) {
    ggplot2::scale_fill_distiller(
      palette   = palette,
      direction = 1,
      limits    = limits,
      name      = label,
      na.value  = "grey95",
      guide     = g
    )
  } else {
    cli::cli_abort(c(
      "Unknown {.arg palette}: {.val {palette}}.",
      i = "Choices: {.val {.palette_choices()}}"
    ))
  }
}

# ---------------------------------------------------------------
# Stand-alone colour-bar legend plot placed to the right of the matrix.
# ---------------------------------------------------------------

.build_legend_plot <- function(colour_by, limits, palette) {
  label <- .colour_label(colour_by)
  df <- data.frame(x = 1, v = limits)
  ggplot2::ggplot(df, ggplot2::aes(x = .data$x, y = .data$v)) +
    ggplot2::geom_point(
      ggplot2::aes(fill = .data$v),
      shape = 22, size = 0, stroke = 0, alpha = 0
    ) +
    .palette_scale(palette, limits, label, colour_by, show_guide = TRUE) +
    ggplot2::theme_void() +
    ggplot2::theme(
      legend.position        = "left",
      # Corrplot-style tracking: the bar flex-fills the legend box's
      # vertical extent via the grid null unit, so the colour-bar height
      # follows the matrix panel height regardless of figure size.
      legend.key.height      = grid::unit(1, "null"),
      legend.key.width       = grid::unit(0.9, "lines"),
      legend.title           = ggplot2::element_text(
        size = 10, face = "bold",
        margin = ggplot2::margin(b = 4)
      ),
      legend.text            = ggplot2::element_text(size = 9),
      legend.margin          = ggplot2::margin(0, 0, 0, 0),
      legend.box.margin      = ggplot2::margin(0, 0, 0, 0),
      legend.box.spacing     = grid::unit(0, "pt"),
      plot.margin            = ggplot2::margin(0, 0, 0, 4)
    )
}

# ---------------------------------------------------------------
# Off-diagonal cell: raw points + CI ribbon + fitted line.
# ---------------------------------------------------------------

.cell_text_sizes <- function(k,
                              text_scale_diag     = 1,
                              text_scale_off_diag = 1) {
  auto       <- max(0.35, (3 / max(k, 2L))^0.55)
  off_scale  <- auto * text_scale_off_diag
  diag_scale <- auto * text_scale_diag
  list(
    n_edf  = 4.0 * off_scale,
    glyph  = 4.6 * off_scale,
    asym   = 3.8 * off_scale,
    shape  = 6.0 * off_scale,
    empty  = 3.2 * off_scale,
    diag   = 5.2 * diag_scale
  )
}

.build_cell <- function(fit_obj, show_data, show_ci, colour_by, palette,
                        signif_glyph, annotations, shape_cutoffs,
                        glyph_style, asym_val,
                        colour_limits, is_upper,
                        text_sizes = .cell_text_sizes(3L),
                        display = "fit",
                        derivative_ci = "none") {
  # Single-quantity dispatch: every cell is one ggplot. "fit" calls
  # the historical renderer (unchanged behaviour), "d1" / "d2" call
  # the derivative renderer. No per-cell stacking; the matrix-level
  # title names the displayed quantity and the summary-table
  # `display` column tags programmatic output.
  display <- as.character(display)
  if (identical(display, "fit")) {
    return(.build_fit_panel(
      fit_obj = fit_obj, show_data = show_data, show_ci = show_ci,
      colour_by = colour_by, palette = palette,
      signif_glyph = signif_glyph, annotations = annotations,
      shape_cutoffs = shape_cutoffs, glyph_style = glyph_style,
      asym_val = asym_val, colour_limits = colour_limits,
      is_upper = is_upper, text_sizes = text_sizes
    ))
  }
  k <- switch(display, d1 = 1L, d2 = 2L,
              cli::cli_abort("Unknown display {.val {display}}."))
  .build_deriv_panel(
    fit_obj = fit_obj, order = k,
    derivative_ci = derivative_ci,
    text_sizes = text_sizes, glyph_style = glyph_style
  )
}

# Original cell body, factored out. Full fit + CI + scatter + all
# corner annotations + colour fill — identical output to the pre-
# derivative release when display = "fit".
.build_fit_panel <- function(fit_obj, show_data, show_ci, colour_by, palette,
                             signif_glyph, annotations, shape_cutoffs,
                             glyph_style, asym_val,
                             colour_limits, is_upper, text_sizes) {
  colour_val   <- .colour_value(fit_obj, colour_by)
  colour_label <- .colour_label(colour_by)

  p <- ggplot2::ggplot() +
    ggplot2::theme_void(base_size = 8) +
    ggplot2::theme(
      aspect.ratio     = 1,
      plot.margin      = ggplot2::margin(3, 3, 3, 3),
      plot.background  = ggplot2::element_rect(fill = NA, colour = NA),
      panel.background = ggplot2::element_rect(fill = NA, colour = NA),
      panel.border     = ggplot2::element_rect(fill = NA,
                                               colour = "grey55",
                                               linewidth = 0.35)
    )

  if (colour_by != "none" && !is.na(colour_val) &&
      all(is.finite(colour_limits))) {
    fill_df <- data.frame(fill_val = colour_val)
    p <- p +
      ggplot2::geom_rect(
        data    = fill_df,
        mapping = ggplot2::aes(xmin = -Inf, xmax = Inf,
                               ymin = -Inf, ymax = Inf,
                               fill = .data$fill_val),
        alpha   = 0.45,
        inherit.aes = FALSE
      ) +
      .palette_scale(palette, colour_limits, colour_label, colour_by)
  }

  if (nrow(fit_obj$pred) == 0L) {
    return(p + ggplot2::annotate(
      "text", x = 0.5, y = 0.5, label = "n < 5",
      size = text_sizes$empty, colour = "grey40"
    ) + ggplot2::xlim(0, 1) + ggplot2::ylim(0, 1))
  }

  if (show_data && nrow(fit_obj$raw) > 0L) {
    raw <- fit_obj$raw
    names(raw) <- c("x", "y")
    p <- p + ggplot2::geom_point(
      data = raw,
      ggplot2::aes(x = .data$x, y = .data$y),
      colour = "grey25", size = 0.4, alpha = 0.25,
      inherit.aes = FALSE
    )
  }
  if (show_ci) {
    p <- p + ggplot2::geom_ribbon(
      data = fit_obj$pred,
      ggplot2::aes(x = .data$x, ymin = .data$lo, ymax = .data$hi),
      fill = "#08306b", alpha = 0.20,
      inherit.aes = FALSE
    )
  }
  p <- p + ggplot2::geom_line(
    data = fit_obj$pred,
    ggplot2::aes(x = .data$x, y = .data$fit),
    colour = "#08306b", linewidth = 0.7,
    inherit.aes = FALSE
  )

  # Top-right: significance glyph
  if (signif_glyph) {
    glyph <- .pvalue_to_glyph(fit_obj$pvalue)
    if (nzchar(glyph)) {
      p <- p + ggplot2::annotate(
        "text", x = Inf, y = Inf, hjust = 1.25, vjust = 1.4,
        label = glyph, size = text_sizes$glyph, fontface = "bold"
      )
    }
  }

  # Bottom-left stack: A (top line) + EDF (bottom line), in whichever
  # combination the user asked for via annotations.
  bl_lines <- character()
  if ("A" %in% annotations && !is.na(asym_val)) {
    bl_lines <- c(bl_lines, sprintf("A = %.2f", asym_val))
  }
  if ("edf" %in% annotations && !is.na(fit_obj$edf)) {
    bl_lines <- c(bl_lines, sprintf("EDF = %.2f", fit_obj$edf))
  }
  if (length(bl_lines)) {
    p <- p + ggplot2::annotate(
      "text", x = -Inf, y = -Inf,
      hjust = -0.07, vjust = -0.15,
      label = paste(bl_lines, collapse = "\n"),
      size = text_sizes$n_edf, lineheight = 0.95,
      colour = "grey15", fontface = "plain"
    )
  }

  # Compute shape category once if either code (top-left) or shape
  # (bottom-right) is requested — they occupy distinct corners and
  # no longer compete.
  want_code  <- "code"  %in% annotations
  want_glyph <- "shape" %in% annotations
  shape_cat <- if (want_code || want_glyph) {
    .classify_shape(
      fit_obj$shape$monotonicity_index,
      fit_obj$shape$convexity_index,
      fit_obj$shape$n_turning_points, fit_obj$shape$n_inflections,
      fit_obj$shape$flat_range_ratio, shape_cutoffs
    )
  } else {
    NA_character_
  }

  # Top-left: 2-letter shape code. ASCII, font-safe, lightweight.
  if (want_code) {
    code_txt <- .shape_lookup(shape_cat, "code")
    if (!is.na(code_txt) && nzchar(code_txt)) {
      p <- p + ggplot2::annotate(
        "text", x = -Inf, y = Inf,
        hjust = -0.20, vjust = 1.4,
        label = code_txt,
        size = text_sizes$n_edf,
        colour = "grey20", fontface = "bold"
      )
    }
  }

  # Bottom-right: shape glyph. Unicode or ASCII per glyph_style.
  if (want_glyph) {
    g <- .shape_glyph(shape_cat, style = glyph_style)
    if (!is.na(g) && nzchar(g)) {
      p <- p + ggplot2::annotate(
        "text", x = Inf, y = -Inf,
        hjust = 1.15, vjust = -0.25,
        label = g,
        size = text_sizes$shape,
        colour = "grey25", fontface = "bold"
      )
    }
  }
  p
}

# ---------------------------------------------------------------
# Derivative sub-panel. Renders f^{(k)}(x) on the full x-grid with
# a 95% pointwise CI ribbon (LP-matrix SE) and a dashed zero-line
# reference. Deliberately minimal: no scatter, no cell fill, no
# corner annotations — all of those live on the fit panel. A small
# f' / f'' label in the top-left corner identifies the order.
# ---------------------------------------------------------------

.deriv_order_label <- function(order, style = c("ascii", "unicode")) {
  style <- rlang::arg_match(style)
  order <- as.integer(order)
  if (!is.finite(order) || order < 1L) return("")
  if (style == "unicode") {
    if (order == 1L) return("f\u2032")          # f'
    if (order == 2L) return("f\u2033")          # f''
    return(paste0("f", strrep("\u2032", order)))
  }
  # ASCII fallback: d1f/dx, d2f/dx2, ...
  if (order == 1L) return("df/dx")
  if (order == 2L) return("d2f/dx2")
  sprintf("d%df/dx%d", order, order)
}

.build_deriv_panel <- function(fit_obj, order, derivative_ci = "none",
                               text_sizes, glyph_style = "ascii") {
  p <- ggplot2::ggplot() +
    ggplot2::theme_void(base_size = 8) +
    ggplot2::theme(
      aspect.ratio     = 1,
      plot.margin      = ggplot2::margin(2, 3, 2, 3),
      plot.background  = ggplot2::element_rect(fill = NA, colour = NA),
      panel.background = ggplot2::element_rect(fill = NA, colour = NA),
      panel.border     = ggplot2::element_rect(
        fill = NA, colour = "grey55", linewidth = 0.35
      )
    )

  deriv_df <- fit_obj$deriv[[as.character(as.integer(order))]]
  if (is.null(deriv_df) || nrow(deriv_df) == 0L) {
    return(p + ggplot2::annotate(
      "text", x = 0.5, y = 0.5,
      label = sprintf("deriv %d unavailable", as.integer(order)),
      size = text_sizes$empty, colour = "grey40"
    ) + ggplot2::xlim(0, 1) + ggplot2::ylim(0, 1))
  }

  # Reference line at zero — sign flips of the derivative correspond
  # to turning points (k=1) or inflections (k=2), so the zero crossing
  # is the feature of interest the panel exists to surface.
  p <- p + ggplot2::geom_hline(
    yintercept = 0, linetype = "dashed", colour = "grey50",
    linewidth = 0.3
  )

  # Ribbon only when the caller has explicitly opted in. The
  # pointwise / simultaneous selection was resolved at fit time and
  # baked into deriv_df$lo / deriv_df$hi, so here we just honour the
  # mode indicator.
  if (!identical(derivative_ci, "none") &&
      all(c("lo", "hi") %in% names(deriv_df)) &&
      all(is.finite(deriv_df$lo)) && all(is.finite(deriv_df$hi))) {
    p <- p + ggplot2::geom_ribbon(
      data = deriv_df,
      ggplot2::aes(x = .data$x, ymin = .data$lo, ymax = .data$hi),
      fill = "#6a3d9a", alpha = 0.20, inherit.aes = FALSE
    )
  }
  p <- p + ggplot2::geom_line(
    data = deriv_df,
    ggplot2::aes(x = .data$x, y = .data$fit),
    colour = "#6a3d9a", linewidth = 0.6, inherit.aes = FALSE
  )

  # Top-left order label. ASCII by default; Unicode primes when the
  # caller has opted in via glyph_style = "unicode".
  lab <- .deriv_order_label(order, style = glyph_style)
  if (nzchar(lab)) {
    p <- p + ggplot2::annotate(
      "text", x = -Inf, y = Inf,
      hjust = -0.15, vjust = 1.4,
      label = lab,
      size = text_sizes$n_edf * 0.95,
      colour = "grey20", fontface = "italic"
    )
  }
  p
}

# ---------------------------------------------------------------
# Blank diagonal cell — used when labels live on the border (or are
# suppressed entirely). Same panel geometry as an off-diagonal cell
# (no fill, thin grey border) so the matrix grid reads uniformly.
# ---------------------------------------------------------------

.build_blank_diagonal_cell <- function() {
  ggplot2::ggplot() +
    ggplot2::xlim(0, 1) + ggplot2::ylim(0, 1) +
    ggplot2::theme_void(base_size = 8) +
    ggplot2::theme(
      aspect.ratio     = 1,
      plot.margin      = ggplot2::margin(3, 3, 3, 3),
      plot.background  = ggplot2::element_rect(fill = NA, colour = NA),
      panel.background = ggplot2::element_rect(fill = NA, colour = NA),
      panel.border     = ggplot2::element_rect(fill = NA,
                                               colour = "grey55",
                                               linewidth = 0.35)
    )
}

# ---------------------------------------------------------------
# Border-label cells — variable names in the top strip (rotated)
# and left strip (horizontal, right-aligned). Mirrors corrplot's
# tl.pos = "lt" convention. No panel border, no aspect ratio;
# widths / heights are set by the assembly layer.
# ---------------------------------------------------------------

.build_top_label_cell <- function(label, srt = 45, cex = 1) {
  hj <- if (identical(srt, 0)) 0.5 else 0
  vj <- if (identical(srt, 0)) 0 else 0.5
  ggplot2::ggplot() +
    ggplot2::annotate(
      "text", x = 0.5, y = 0,
      label = label, angle = srt,
      hjust = hj, vjust = vj,
      size = 3.5 * cex, colour = "grey10"
    ) +
    ggplot2::xlim(0, 1) + ggplot2::ylim(0, 1) +
    ggplot2::coord_cartesian(clip = "off") +
    ggplot2::theme_void(base_size = 8) +
    ggplot2::theme(
      plot.margin      = ggplot2::margin(3, 3, 3, 3),
      plot.background  = ggplot2::element_rect(fill = NA, colour = NA),
      panel.background = ggplot2::element_rect(fill = NA, colour = NA)
    )
}

.build_left_label_cell <- function(label, cex = 1) {
  ggplot2::ggplot() +
    ggplot2::annotate(
      "text", x = 1, y = 0.5,
      label = label,
      hjust = 1, vjust = 0.5,
      size = 3.5 * cex, colour = "grey10"
    ) +
    ggplot2::xlim(0, 1) + ggplot2::ylim(0, 1) +
    ggplot2::coord_cartesian(clip = "off") +
    ggplot2::theme_void(base_size = 8) +
    ggplot2::theme(
      plot.margin      = ggplot2::margin(3, 3, 3, 3),
      plot.background  = ggplot2::element_rect(fill = NA, colour = NA),
      panel.background = ggplot2::element_rect(fill = NA, colour = NA)
    )
}

.build_corner_cell <- function() {
  ggplot2::ggplot() +
    ggplot2::xlim(0, 1) + ggplot2::ylim(0, 1) +
    ggplot2::theme_void(base_size = 8) +
    ggplot2::theme(
      plot.margin      = ggplot2::margin(0, 0, 0, 0),
      plot.background  = ggplot2::element_rect(fill = NA, colour = NA),
      panel.background = ggplot2::element_rect(fill = NA, colour = NA)
    )
}

# ---------------------------------------------------------------
# Diagonal cell — kernel density of the variable with a rug of raw
# values along the bottom edge. Mirrors the GGally::ggpairs default
# diagonal so readers can see tail weight, bimodality, support
# clipping, and any concentration of mass that would otherwise be
# invisible from the off-diagonal smooths alone.
# ---------------------------------------------------------------

.build_density_rug_cell <- function(var_name, x_values,
                                    text_sizes = .cell_text_sizes(3L)) {
  x <- x_values[is.finite(x_values)]
  if (length(x) < 5L) return(.build_blank_diagonal_cell())
  dens <- tryCatch(stats::density(x, na.rm = TRUE),
                   error = function(e) NULL)
  if (is.null(dens) || !length(dens$x)) {
    return(.build_blank_diagonal_cell())
  }
  dens_df <- data.frame(x = dens$x, y = dens$y)
  rug_df  <- data.frame(x = x)
  ggplot2::ggplot() +
    ggplot2::geom_area(
      data    = dens_df,
      mapping = ggplot2::aes(x = .data$x, y = .data$y),
      fill    = "grey55", colour = NA, alpha = 0.35,
      inherit.aes = FALSE
    ) +
    ggplot2::geom_line(
      data    = dens_df,
      mapping = ggplot2::aes(x = .data$x, y = .data$y),
      colour  = "grey20", linewidth = 0.45,
      inherit.aes = FALSE
    ) +
    ggplot2::geom_rug(
      data    = rug_df,
      mapping = ggplot2::aes(x = .data$x),
      sides   = "b", colour = "grey25",
      alpha   = 0.35, length = grid::unit(0.07, "npc"),
      inherit.aes = FALSE
    ) +
    ggplot2::theme_void(base_size = 8) +
    ggplot2::theme(
      aspect.ratio     = 1,
      plot.margin      = ggplot2::margin(3, 3, 3, 3),
      plot.background  = ggplot2::element_rect(fill = NA, colour = NA),
      panel.background = ggplot2::element_rect(fill = NA, colour = NA),
      panel.border     = ggplot2::element_rect(
        fill = NA, colour = "grey55", linewidth = 0.35
      )
    )
}

# ---------------------------------------------------------------
# Diagonal cell — the variable name, bold, neutral grey background.
# Used only when labels = "diagonal" (legacy layout).
# ---------------------------------------------------------------

.build_diagonal_cell <- function(var_name,
                                 text_sizes = .cell_text_sizes(3L)) {
  char_scale <- min(1, 10 / max(1L, nchar(var_name)))
  label_size <- text_sizes$diag * char_scale
  ggplot2::ggplot() +
    ggplot2::annotate(
      "text", x = 0.5, y = 0.5, label = var_name,
      size = label_size, fontface = "bold", colour = "grey10"
    ) +
    ggplot2::xlim(0, 1) + ggplot2::ylim(0, 1) +
    ggplot2::theme_void(base_size = 8) +
    ggplot2::theme(
      aspect.ratio     = 1,
      plot.margin      = ggplot2::margin(3, 3, 3, 3),
      plot.background  = ggplot2::element_rect(fill = NA, colour = NA),
      panel.background = ggplot2::element_rect(fill = "grey92",
                                               colour = NA),
      panel.border     = ggplot2::element_rect(fill = NA,
                                               colour = "grey55",
                                               linewidth = 0.35)
    )
}

# Snapshot tests for stable text surfaces. If the glossary caption
# wording drifts (which it shouldn't without a NEWS entry), this
# test fails and prompts a regeneration via
# testthat::snapshot_accept("snapshot").

test_that("p-value glyph ladder matches snapshot", {
  expect_snapshot({
    lapply(
      c(0.0001, 0.005, 0.03, 0.08, 0.5, NA_real_),
      janusplot:::.pvalue_to_glyph
    )
  })
})

test_that("palette choice vector matches snapshot", {
  expect_snapshot(janusplot:::.palette_choices())
})

test_that(".cell_text_sizes output at k = 3 matches snapshot", {
  expect_snapshot(janusplot:::.cell_text_sizes(3L))
})

test_that(".cell_text_sizes output at k = 10 matches snapshot", {
  expect_snapshot(janusplot:::.cell_text_sizes(10L))
})

# Shape-recognition sensitivity study (package-native).
# Public API:
#   janusplot_shape_sensitivity_shapes()   - list canonical truths
#   janusplot_shape_sensitivity()          - run the sweep
#   janusplot_shape_sensitivity_summary()  - aggregate raw results
#   janusplot_shape_sensitivity_plot()     - confusion / accuracy /
#                                            recovery-curve plots
#
# Ground-truth generators live on x in [0, 1] with y normalised to
# [0, 1] so sigma parameterises noise consistently across shapes
# (sigma is a fraction of the y-range).

# ---------------------------------------------------------------
# Internal helpers
# ---------------------------------------------------------------

.norm01 <- function(y) {
  r <- range(y, na.rm = TRUE)
  if (diff(r) == 0) return(rep(0.5, length(y)))
  (y - r[1L]) / diff(r)
}

.shape_sensitivity_generators <- function() {
  list(
    linear_up    = function(x) x,
    linear_down  = function(x) 1 - x,
    convex_up    = function(x) x^2,
    concave_up   = function(x) sqrt(x),
    convex_down  = function(x) (1 - x)^2,
    concave_down = function(x) sqrt(1 - x),
    s_shape      = function(x) 1 / (1 + exp(-10 * (x - 0.5))),
    u_shape      = function(x) .norm01(4 * (x - 0.5)^2),
    inverted_u   = function(x) .norm01(1 - 4 * (x - 0.5)^2),
    skewed_peak  = function(x) .norm01(x * exp(-3 * x)),
    broad_peak   = function(x) .norm01(exp(-((x - 0.5) * 3)^4)),
    wave         = function(x) 0.5 + 0.5 * sin(2 * pi * (x - 0.25)),
    bimodal      = function(x) {
      .norm01(exp(-((x - 0.25) * 6)^2) + exp(-((x - 0.75) * 6)^2))
    },
    bi_wave      = function(x) 0.5 + 0.5 * sin(4 * pi * (x - 0.125))
  )
}

# Expected archetype per truth — pulled from the live taxonomy so the
# sensitivity study cannot drift from the classifier code.
.shape_sensitivity_archetypes <- function() {
  gens  <- .shape_sensitivity_generators()
  stats::setNames(.shape_lookup(names(gens), "archetype"), names(gens))
}

.sens_simulate_one <- function(truth, n, sigma, seed, cutoffs,
                               generators, archetype_truth) {
  set.seed(seed, kind = "default")
  x <- sort(stats::runif(n, 0, 1))
  gen   <- generators[[truth]]
  y_tru <- gen(x)
  y     <- y_tru + stats::rnorm(n, sd = sigma)
  d     <- data.frame(x = x, y = y)

  fit <- tryCatch(
    suppressWarnings(mgcv::gam(y ~ s(x), data = d, method = "REML")),
    error = function(e) NULL
  )
  if (is.null(fit)) {
    return(data.frame(
      truth              = truth,
      n                  = n,
      sigma              = sigma,
      seed               = seed,
      predicted          = NA_character_,
      correct            = NA,
      archetype_truth    = unname(archetype_truth[truth]),
      archetype_pred     = NA_character_,
      archetype_correct  = NA,
      monotonicity_index = NA_real_,
      convexity_index    = NA_real_,
      n_turn             = NA_integer_,
      n_inflect          = NA_integer_,
      error              = "gam_fit_failed",
      stringsAsFactors   = FALSE
    ))
  }
  m <- janusplot_shape_metrics(fit, x_name = "x", newdata = d,
                               cutoffs = cutoffs)
  arche_pred <- .shape_lookup(m$shape_category, "archetype")
  arche_true <- unname(archetype_truth[truth])
  data.frame(
    truth              = truth,
    n                  = n,
    sigma              = sigma,
    seed               = seed,
    predicted          = m$shape_category,
    correct            = isTRUE(m$shape_category == truth),
    archetype_truth    = arche_true,
    archetype_pred     = arche_pred,
    archetype_correct  = isTRUE(arche_pred == arche_true),
    monotonicity_index = m$monotonicity_index,
    convexity_index    = m$convexity_index,
    n_turn             = m$n_turning_points,
    n_inflect          = m$n_inflections,
    error              = NA_character_,
    stringsAsFactors   = FALSE
  )
}

# ---------------------------------------------------------------
# Public API
# ---------------------------------------------------------------

#' Canonical ground-truth shapes for the sensitivity study
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Return the names of every canonical ground-truth shape that
#' [janusplot_shape_sensitivity()] can simulate from. Fourteen shapes
#' spanning five archetypes (`monotone_linear`, `monotone_curved`,
#' `unimodal`, `wave`, `multimodal`). The `chaotic` and `degenerate`
#' archetypes are out of scope (no realistic deterministic generator).
#'
#' @returns Character vector of length 14 — the generator names.
#'
#' @seealso [janusplot_shape_sensitivity()], [janusplot_shape_hierarchy()].
#'
#' @examples
#' janusplot_shape_sensitivity_shapes()
#' @family shape-sensitivity
#' @export
janusplot_shape_sensitivity_shapes <- function() {
  names(.shape_sensitivity_generators())
}

#' Shape-recognition sensitivity study
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Run a full-factorial sensitivity sweep for the janusplot 24-category
#' shape classifier. For each combination of ground-truth shape, sample
#' size `n`, noise level `sigma`, and replicate, the sweep:
#' \enumerate{
#'   \item Generates `n` points from the noiseless canonical curve on
#'     `[0, 1]` + Gaussian noise with SD = `sigma` (fraction of the
#'     y-range, so signal-to-noise is comparable across shapes).
#'   \item Fits `mgcv::gam(y ~ s(x), method = "REML")`.
#'   \item Runs [janusplot_shape_metrics()] to classify the fitted smooth.
#'   \item Records correctness at both the fine (24-category) and
#'     archetype (7-family) levels.
#' }
#'
#' The function is the package-native implementation of
#' `simulation/scripts/scenario_4_shape_recognition.R`. A small
#' precomputed dataset is shipped as [shape_sensitivity_demo] for
#' downstream examples without requiring users to re-run the sweep.
#'
#' @param shapes Character vector of ground-truth names from
#'   [janusplot_shape_sensitivity_shapes()]. Default `NULL` → all 14.
#' @param n_grid Integer vector of sample sizes. Default
#'   `c(50L, 100L, 200L, 500L)`.
#' @param sigma_grid Numeric vector of noise levels (fraction of the
#'   y-range). Default `c(0.02, 0.05, 0.10, 0.20, 0.40)`.
#' @param n_rep Integer. Replicates per cell. Default `200L`.
#' @param cutoffs Named list of classification thresholds; see
#'   [janusplot_shape_cutoffs()].
#' @param parallel Logical. If `TRUE` and `future.apply` is installed,
#'   dispatch replicates in parallel. The caller is responsible for
#'   configuring `future::plan()` (e.g. `future::plan(future::multisession)`).
#' @param seed Integer. Base seed — each fit uses `seed + row_index`
#'   so results are reproducible and cell-permutation-invariant.
#' @param verbose Logical. Print progress messages to the console.
#'   Default is `interactive()`.
#'
#' @returns A data frame with one row per fit. Columns:
#' \describe{
#'   \item{`truth`}{Ground-truth shape name.}
#'   \item{`n`}{Sample size for this fit.}
#'   \item{`sigma`}{Noise level for this fit.}
#'   \item{`seed`}{RNG seed used.}
#'   \item{`predicted`}{Classifier output at the fine (24-category) level.}
#'   \item{`correct`}{Logical — does `predicted == truth`?}
#'   \item{`archetype_truth`}{Expected archetype for `truth`.}
#'   \item{`archetype_pred`}{Archetype of `predicted`.}
#'   \item{`archetype_correct`}{Logical — archetype-level correctness.}
#'   \item{`monotonicity_index`}{Monotonicity index `M` (see
#'     [janusplot_shape_metrics()]).}
#'   \item{`convexity_index`}{Convexity index `C` (see
#'     [janusplot_shape_metrics()]).}
#'   \item{`n_turn`, `n_inflect`}{Recovered turning-point and
#'     inflection counts.}
#'   \item{`error`}{`"gam_fit_failed"` when `mgcv::gam()` errored;
#'     `NA` otherwise.}
#' }
#'
#' @seealso [janusplot_shape_sensitivity_summary()],
#'   [janusplot_shape_sensitivity_plot()],
#'   [janusplot_shape_sensitivity_shapes()], [shape_sensitivity_demo].
#'
#' @examples
#' # Tiny-run smoke test (< 2 seconds): 3 shapes x 2 n x 2 sigma x 5 reps.
#' res <- janusplot_shape_sensitivity(
#'   shapes     = c("linear_up", "u_shape", "wave"),
#'   n_grid     = c(100L, 200L),
#'   sigma_grid = c(0.05, 0.20),
#'   n_rep      = 5L,
#'   verbose    = FALSE
#' )
#' head(res)
#' janusplot_shape_sensitivity_summary(res, level = "archetype")
#' @family shape-sensitivity
#' @export
janusplot_shape_sensitivity <- function(
    shapes     = NULL,
    n_grid     = c(50L, 100L, 200L, 500L),
    sigma_grid = c(0.02, 0.05, 0.10, 0.20, 0.40),
    n_rep      = 200L,
    cutoffs    = janusplot_shape_cutoffs(),
    parallel   = FALSE,
    seed       = 2026L,
    verbose    = interactive()
) {
  generators <- .shape_sensitivity_generators()
  all_shapes <- names(generators)
  if (is.null(shapes)) shapes <- all_shapes
  if (!is.character(shapes) || !length(shapes)) {
    cli::cli_abort("{.arg shapes} must be a non-empty character vector.")
  }
  bad <- setdiff(shapes, all_shapes)
  if (length(bad)) {
    cli::cli_abort(c(
      "Unknown {.arg shapes} value{?s}: {.val {bad}}.",
      i = "Allowed: {.val {all_shapes}}."
    ))
  }
  if (!is.numeric(n_grid)     || !length(n_grid)     || any(n_grid < 5L)) {
    cli::cli_abort("{.arg n_grid} must be a numeric vector with every value >= 5.")
  }
  if (!is.numeric(sigma_grid) || !length(sigma_grid) || any(sigma_grid < 0)) {
    cli::cli_abort("{.arg sigma_grid} must be a non-negative numeric vector.")
  }
  if (!is.numeric(n_rep) || length(n_rep) != 1L || n_rep < 1L) {
    cli::cli_abort("{.arg n_rep} must be a single positive integer.")
  }

  archetype_truth <- .shape_sensitivity_archetypes()
  grid <- expand.grid(
    truth = shapes, n = as.integer(n_grid), sigma = sigma_grid,
    rep = seq_len(as.integer(n_rep)),
    stringsAsFactors = FALSE, KEEP.OUT.ATTRS = FALSE
  )
  grid$seed <- as.integer(seed) + seq_len(nrow(grid))

  total <- nrow(grid)
  if (isTRUE(verbose)) {
    cli::cli_inform(c(
      i = paste0(
        "Running {total} fits: {length(shapes)} shape{?s} x ",
        "{length(n_grid)} n x {length(sigma_grid)} sigma x {n_rep} rep{?s}",
        if (parallel) " (parallel)" else " (serial)", "."
      )
    ))
  }

  use_parallel <- isTRUE(parallel) &&
    rlang::is_installed("future.apply")
  if (isTRUE(parallel) && !use_parallel) {
    cli::cli_warn(
      "{.arg parallel = TRUE} but {.pkg future.apply} is not installed; running serially."
    )
  }

  worker <- function(i) {
    .sens_simulate_one(
      truth           = grid$truth[i],
      n               = grid$n[i],
      sigma           = grid$sigma[i],
      seed            = grid$seed[i],
      cutoffs         = cutoffs,
      generators      = generators,
      archetype_truth = archetype_truth
    )
  }

  rows <- if (use_parallel) {
    future.apply::future_lapply(
      seq_len(total), worker,
      future.seed     = TRUE,
      future.packages = c("janusplot", "mgcv")
    )
  } else {
    lapply(seq_len(total), worker)
  }
  do.call(rbind, rows)
}

#' Summarise a shape-sensitivity sweep
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Aggregate the raw output of [janusplot_shape_sensitivity()] into a
#' per-cell mean-accuracy table at either the fine (24-category) or
#' archetype (7-family) level.
#'
#' @param results Data frame returned by [janusplot_shape_sensitivity()].
#' @param level One of `"fine"` (default) or `"archetype"`.
#'
#' @returns A data frame with columns `truth`, `n`, `sigma`, `accuracy`.
#'
#' @examples
#' data("shape_sensitivity_demo", package = "janusplot")
#' head(janusplot_shape_sensitivity_summary(shape_sensitivity_demo,
#'                                          level = "archetype"))
#' @family shape-sensitivity
#' @export
janusplot_shape_sensitivity_summary <- function(
    results, level = c("fine", "archetype")
) {
  level <- rlang::arg_match(level)
  col <- if (level == "fine") "correct" else "archetype_correct"
  if (!all(c("truth", "n", "sigma", col) %in% names(results))) {
    cli::cli_abort(
      "{.arg results} must come from {.fn janusplot_shape_sensitivity}."
    )
  }
  agg <- stats::aggregate(
    as.integer(results[[col]]),
    by  = list(truth = results$truth, n = results$n, sigma = results$sigma),
    FUN = mean, na.rm = TRUE
  )
  names(agg)[ncol(agg)] <- "accuracy"
  agg
}

# ---------------------------------------------------------------
# Plot helpers
# ---------------------------------------------------------------

.sens_plot_confusion_fine <- function(results) {
  tax <- .shape_taxonomy()
  lvls_truth <- intersect(tax$category, unique(results$truth))
  lvls_pred  <- tax$category
  tbl <- as.data.frame(table(
    truth     = factor(results$truth,     levels = lvls_truth),
    predicted = factor(results$predicted, levels = lvls_pred)
  ))
  tbl$proportion <- with(tbl, Freq / ave(Freq, truth, FUN = sum))
  ggplot2::ggplot(tbl, ggplot2::aes(x = .data$predicted, y = .data$truth,
                                    fill = .data$proportion)) +
    ggplot2::geom_tile(colour = "grey80") +
    ggplot2::geom_text(ggplot2::aes(
      label = ifelse(.data$Freq == 0, "",
                     sprintf("%.2f", .data$proportion))),
      size = 2.2, colour = "white") +
    ggplot2::scale_fill_viridis_c(name = "P(pred | truth)",
                                  limits = c(0, 1), na.value = "grey95") +
    ggplot2::scale_x_discrete(drop = FALSE) +
    ggplot2::scale_y_discrete(limits = rev) +
    ggplot2::labs(title = "Fine-grained confusion",
                  x = "Predicted category",
                  y = "Ground-truth category") +
    ggplot2::theme_minimal(base_size = 9) +
    ggplot2::theme(
      axis.text.x = ggplot2::element_text(angle = 45, hjust = 1),
      panel.grid  = ggplot2::element_blank(),
      plot.title  = ggplot2::element_text(face = "bold")
    )
}

.sens_plot_confusion_archetype <- function(results) {
  lvls <- c("monotone_linear", "monotone_curved", "unimodal",
            "wave", "multimodal", "chaotic", "degenerate")
  tbl <- as.data.frame(table(
    truth     = factor(results$archetype_truth, levels = lvls),
    predicted = factor(results$archetype_pred,  levels = lvls)
  ))
  tbl$proportion <- with(tbl, Freq / ave(Freq, truth, FUN = sum))
  ggplot2::ggplot(tbl, ggplot2::aes(x = .data$predicted, y = .data$truth,
                                    fill = .data$proportion)) +
    ggplot2::geom_tile(colour = "grey80") +
    ggplot2::geom_text(ggplot2::aes(
      label = ifelse(.data$Freq == 0, "",
                     sprintf("%.2f", .data$proportion))),
      size = 3, colour = "white") +
    ggplot2::scale_fill_viridis_c(name = "P(pred | truth)",
                                  limits = c(0, 1), na.value = "grey95") +
    ggplot2::scale_x_discrete(drop = FALSE) +
    ggplot2::scale_y_discrete(limits = rev) +
    ggplot2::labs(title = "Archetype confusion",
                  x = "Predicted archetype",
                  y = "Ground-truth archetype") +
    ggplot2::theme_minimal(base_size = 10) +
    ggplot2::theme(
      axis.text.x = ggplot2::element_text(angle = 30, hjust = 1),
      panel.grid  = ggplot2::element_blank(),
      plot.title  = ggplot2::element_text(face = "bold")
    )
}

.sens_plot_accuracy_grid <- function(results) {
  agg <- janusplot_shape_sensitivity_summary(results, "archetype")
  ggplot2::ggplot(agg, ggplot2::aes(x = factor(.data$sigma),
                                    y = factor(.data$n),
                                    fill = .data$accuracy)) +
    ggplot2::geom_tile(colour = "white") +
    ggplot2::geom_text(ggplot2::aes(label = sprintf("%.2f", .data$accuracy)),
                       colour = "white", size = 3) +
    ggplot2::facet_wrap(~ truth, ncol = 4) +
    ggplot2::scale_fill_viridis_c(name = "P(correct)", limits = c(0, 1)) +
    ggplot2::labs(x = expression(sigma), y = "n",
                  title = "Archetype-level recovery accuracy") +
    ggplot2::theme_minimal(base_size = 9) +
    ggplot2::theme(
      panel.grid = ggplot2::element_blank(),
      plot.title = ggplot2::element_text(face = "bold"),
      strip.text = ggplot2::element_text(face = "bold")
    )
}

.sens_plot_recovery_curves <- function(results) {
  agg <- janusplot_shape_sensitivity_summary(results, "archetype")
  ggplot2::ggplot(agg, ggplot2::aes(x = .data$sigma, y = .data$accuracy,
                                    colour = factor(.data$n))) +
    ggplot2::geom_line(linewidth = 0.6) +
    ggplot2::geom_point(size = 1.6) +
    ggplot2::facet_wrap(~ truth, ncol = 4) +
    ggplot2::scale_y_continuous(limits = c(0, 1)) +
    ggplot2::scale_colour_viridis_d(name = "n") +
    ggplot2::labs(
      x = expression("Noise " ~ sigma ~ " (fraction of y-range)"),
      y = "P(archetype correct)",
      title = "Archetype recovery curves"
    ) +
    ggplot2::theme_minimal(base_size = 9) +
    ggplot2::theme(
      strip.text = ggplot2::element_text(face = "bold"),
      plot.title = ggplot2::element_text(face = "bold")
    )
}

#' Visualise a shape-sensitivity sweep
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Produce one of four diagnostic plots from the raw data frame returned
#' by [janusplot_shape_sensitivity()]:
#' \describe{
#'   \item{`"confusion_fine"`}{24 x (|shapes|) confusion matrix at the
#'     fine category level — rows = ground truth, columns = predicted,
#'     cells coloured by `P(pred | truth)`.}
#'   \item{`"confusion_archetype"`}{7 x 7 confusion matrix at the
#'     archetype level.}
#'   \item{`"accuracy_grid"`}{per-shape heatmap of archetype-level
#'     accuracy across the `(n, sigma)` design.}
#'   \item{`"recovery_curves"`}{accuracy as a function of sigma,
#'     one line per sample size, faceted by shape.}
#' }
#'
#' @param results Data frame from [janusplot_shape_sensitivity()] or the
#'   precomputed [shape_sensitivity_demo].
#' @param type One of `"confusion_fine"`, `"confusion_archetype"`,
#'   `"accuracy_grid"`, or `"recovery_curves"`.
#'
#' @returns A [ggplot2::ggplot] object.
#'
#' @examples
#' data("shape_sensitivity_demo", package = "janusplot")
#' janusplot_shape_sensitivity_plot(shape_sensitivity_demo,
#'                                  "recovery_curves")
#' @family shape-sensitivity
#' @export
janusplot_shape_sensitivity_plot <- function(
    results,
    type = c("confusion_fine", "confusion_archetype",
             "accuracy_grid", "recovery_curves")
) {
  type <- rlang::arg_match(type)
  switch(type,
    confusion_fine      = .sens_plot_confusion_fine(results),
    confusion_archetype = .sens_plot_confusion_archetype(results),
    accuracy_grid       = .sens_plot_accuracy_grid(results),
    recovery_curves     = .sens_plot_recovery_curves(results)
  )
}

test_that("janusplot_shape_sensitivity_shapes returns 14 canonical names", {
  shapes <- janusplot_shape_sensitivity_shapes()
  expect_type(shapes, "character")
  expect_length(shapes, 14L)
  expect_true(all(shapes %in% janusplot_shape_hierarchy()$category))
})

test_that("janusplot_shape_sensitivity runs and returns a well-formed data frame", {
  res <- janusplot_shape_sensitivity(
    shapes     = c("linear_up", "u_shape", "wave"),
    n_grid     = c(100L, 200L),
    sigma_grid = c(0.05, 0.20),
    n_rep      = 3L,
    verbose    = FALSE
  )
  expect_s3_class(res, "data.frame")
  expect_equal(nrow(res), 3L * 2L * 2L * 3L)
  expect_true(all(c("truth", "n", "sigma", "seed", "predicted",
                    "correct", "archetype_truth", "archetype_pred",
                    "archetype_correct",
                    "monotonicity_index", "convexity_index",
                    "n_turn", "n_inflect", "error") %in% names(res)))
  # Every truth in, every truth accounted for
  expect_setequal(unique(res$truth), c("linear_up", "u_shape", "wave"))
})

test_that("rejects unknown shapes / malformed grids", {
  expect_error(
    janusplot_shape_sensitivity(shapes = "not_a_shape", n_rep = 2L,
                                verbose = FALSE),
    regexp = "Unknown"
  )
  expect_error(
    janusplot_shape_sensitivity(n_grid = c(3L), n_rep = 2L,
                                verbose = FALSE),
    regexp = "n_grid"
  )
  expect_error(
    janusplot_shape_sensitivity(sigma_grid = c(-0.1), n_rep = 2L,
                                verbose = FALSE),
    regexp = "sigma_grid"
  )
  expect_error(
    janusplot_shape_sensitivity(n_rep = 0L, verbose = FALSE),
    regexp = "n_rep"
  )
})

test_that("summary returns the expected accuracy table", {
  res <- janusplot_shape_sensitivity(
    shapes     = c("linear_up", "u_shape"),
    n_grid     = c(100L),
    sigma_grid = c(0.05, 0.20),
    n_rep      = 3L,
    verbose    = FALSE
  )
  fine <- janusplot_shape_sensitivity_summary(res, "fine")
  arch <- janusplot_shape_sensitivity_summary(res, "archetype")
  expect_true(all(c("truth", "n", "sigma", "accuracy") %in% names(fine)))
  expect_true(all(c("truth", "n", "sigma", "accuracy") %in% names(arch)))
  expect_equal(nrow(fine), 2L * 1L * 2L)
  expect_true(all(fine$accuracy    >= 0 & fine$accuracy    <= 1))
  expect_true(all(arch$accuracy    >= 0 & arch$accuracy    <= 1))
})

test_that("all four plot types render to ggplot objects", {
  data("shape_sensitivity_demo", package = "janusplot",
       envir = environment())
  for (tp in c("confusion_fine", "confusion_archetype",
               "accuracy_grid", "recovery_curves")) {
    p <- janusplot_shape_sensitivity_plot(shape_sensitivity_demo, tp)
    expect_s3_class(p, "ggplot")
  }
})

test_that("plot rejects unknown types", {
  data("shape_sensitivity_demo", package = "janusplot",
       envir = environment())
  expect_error(
    janusplot_shape_sensitivity_plot(shape_sensitivity_demo, "bogus"),
    class = "rlang_error"
  )
})

test_that("linear_up is recognised as monotone at moderate n, low noise", {
  # Sanity bound on H1-style behaviour for the easiest shape. The
  # REML spline may pick up tiny curvature from noisy linear data
  # (landing on convex_up / concave_up — still monotone increasing),
  # so we test the coarser `shape_monotonic` tier: the classifier
  # must at least recognise the truth as monotone.
  res <- janusplot_shape_sensitivity(
    shapes     = "linear_up",
    n_grid     = 500L,
    sigma_grid = 0.05,
    n_rep      = 20L,
    verbose    = FALSE
  )
  monot_pred <- janusplot:::.shape_lookup(res$predicted, "monotonic")
  expect_gte(mean(monot_pred == "monotone", na.rm = TRUE), 0.95)
})

# Internal helpers — NOT EXPORTED.
# Shape-types legend: always shows the full taxonomy as a grid of
# canonical 1-cm thumbnail splines. Font-independent, publication-
# ready. Attached below the matrix at full width.

# Build the complete shape-categories legend plot. Renders all rows of
# .shape_taxonomy() so the plate doubles as a stable reference key
# readers can navigate back to regardless of which categories appear
# in the matrix they are inspecting.
.build_shape_legend_plot <- function(ncol = 6L,
                                     line_colour = "#08306b") {
  tax <- .shape_taxonomy()
  # Build one (x, y, category, label) data frame, with y per-category
  # normalised to [0, 1] for consistent panel scales.
  rows <- lapply(tax$category, function(cat) {
    df <- .shape_thumbnail_df(cat)
    y <- df$y
    if (all(is.na(y))) {
      df$y_norm <- NA_real_
    } else {
      r <- range(y, na.rm = TRUE)
      df$y_norm <- if (diff(r) == 0) 0.5 else (y - r[1]) / diff(r)
    }
    df
  })
  thumbs <- do.call(rbind, rows)
  # Panel strip: `label (code)` — keeps the compact 2-letter code
  # visible next to the long name so readers can cross-reference
  # cells that render the code.
  panel_label <- paste0(
    tax$label, " (", tax$code, ")"
  )
  names(panel_label) <- tax$category
  thumbs$panel <- panel_label[thumbs$category]
  # Preserve taxonomy order in the facet layout.
  level_labels <- panel_label
  thumbs$panel <- factor(thumbs$panel, levels = level_labels)

  # Special-case render: `indeterminate` has no thumbnail curve. We
  # draw a centred "?" annotation on a placeholder panel so the grid
  # stays rectangular.
  indet_panel <- panel_label[["indeterminate"]]
  flat_panel  <- panel_label[["flat"]]

  p <- ggplot2::ggplot(thumbs) +
    ggplot2::geom_line(
      ggplot2::aes(x = .data$x, y = .data$y_norm),
      colour = line_colour, linewidth = 0.8, na.rm = TRUE
    ) +
    # Flat category: dashed horizontal marker so it reads as "flat"
    # rather than empty panel.
    ggplot2::geom_hline(
      data = data.frame(panel = factor(flat_panel, levels = level_labels)),
      ggplot2::aes(yintercept = 0.5),
      colour = line_colour, linewidth = 0.6, linetype = "dashed"
    ) +
    ggplot2::geom_text(
      data = data.frame(
        panel = factor(indet_panel, levels = level_labels),
        x = 0.5, y = 0.5, glyph = "?"
      ),
      ggplot2::aes(x = .data$x, y = .data$y, label = .data$glyph),
      colour = "grey40", size = 5, fontface = "bold"
    ) +
    ggplot2::facet_wrap(~ panel, ncol = as.integer(ncol)) +
    ggplot2::scale_x_continuous(limits = c(-0.02, 1.02), expand = c(0, 0)) +
    ggplot2::scale_y_continuous(limits = c(-0.1, 1.1), expand = c(0, 0)) +
    ggplot2::labs(title = "Shape categories",
                  x = NULL, y = NULL) +
    ggplot2::theme_void(base_size = 8) +
    ggplot2::theme(
      plot.title        = ggplot2::element_text(
        size = 10, face = "bold",
        margin = ggplot2::margin(t = 4, b = 6)
      ),
      strip.text        = ggplot2::element_text(
        size = 8.5, colour = "grey15",
        margin = ggplot2::margin(t = 3, b = 1)
      ),
      strip.background  = ggplot2::element_rect(fill = NA, colour = NA),
      panel.background  = ggplot2::element_rect(fill = "grey98",
                                                colour = NA),
      panel.border      = ggplot2::element_rect(fill = NA,
                                                colour = "grey75",
                                                linewidth = 0.35),
      panel.spacing.x   = grid::unit(4, "pt"),
      panel.spacing.y   = grid::unit(6, "pt"),
      plot.margin       = ggplot2::margin(4, 4, 4, 4)
    )
  p
}

# (Retained helper — still used by the cell renderer when the user
# explicitly opts into cell glyphs via annotations = "shape".)
.present_shape_categories <- function(fits, cutoffs) {
  vapply(fits, function(f) {
    .classify_shape(
      f$shape$monotonicity_index, f$shape$convexity_index,
      f$shape$n_turning_points, f$shape$n_inflections,
      f$shape$flat_range_ratio, cutoffs
    )
  }, character(1L))
}

# Internal helpers — NOT EXPORTED.
# Matrix-assembly layer: stitch cells + diagonals into a k x k patchwork
# (labels on diagonals or suppressed) or a (k+1) x (k+1) patchwork
# with a top row + left column of border labels (corrplot tl.pos = "lt").
#
# Composite layout (post-matrix):
#   top row    -> matrix | colour-bar (colour-bar optional)
#   bottom row -> shape-types legend (optional, spans full width)

.assemble_matrix <- function(cells_by_ij, diag_cells, vars, colour_by,
                             colour_limits, palette,
                             shape_legend_plot = NULL,
                             labels = "border",
                             label_srt = 45,
                             label_cex = 1) {
  k <- length(vars)
  has_border <- identical(labels, "border")

  if (!has_border) {
    grid <- vector("list", k * k)
    for (i in seq_len(k)) {
      for (j in seq_len(k)) {
        idx <- (i - 1L) * k + j
        if (i == j) {
          grid[[idx]] <- diag_cells[[i]]
        } else {
          key <- sprintf("%d_%d", i, j)
          grid[[idx]] <- cells_by_ij[[key]]
        }
      }
    }
    matrix_grid <- patchwork::wrap_plots(
      grid, nrow = k, ncol = k, byrow = TRUE
    )
  } else {
    # (k + 1) x (k + 1) grid:
    #   row 0     -> corner | top labels (columns)
    #   rows 1..k -> left labels (rows) | matrix cells
    grid <- vector("list", (k + 1L) * (k + 1L))
    for (r in seq_len(k + 1L)) {
      for (c_ in seq_len(k + 1L)) {
        idx <- (r - 1L) * (k + 1L) + c_
        if (r == 1L && c_ == 1L) {
          grid[[idx]] <- .build_corner_cell()
        } else if (r == 1L) {
          grid[[idx]] <- .build_top_label_cell(
            vars[c_ - 1L], srt = label_srt, cex = label_cex
          )
        } else if (c_ == 1L) {
          grid[[idx]] <- .build_left_label_cell(
            vars[r - 1L], cex = label_cex
          )
        } else {
          i <- r - 1L
          j <- c_ - 1L
          if (i == j) {
            grid[[idx]] <- diag_cells[[i]]
          } else {
            key <- sprintf("%d_%d", i, j)
            grid[[idx]] <- cells_by_ij[[key]]
          }
        }
      }
    }

    # Strip allocations. Horizontal: 2.0 cm absolute for left labels.
    # Vertical: taller at 45/90 so rotated text has room to breathe.
    strip_w <- grid::unit(2.0, "cm")
    strip_h <- if (identical(label_srt, 0)) {
      grid::unit(0.6, "cm")
    } else if (abs(label_srt) >= 80) {
      grid::unit(2.0, "cm")
    } else {
      grid::unit(1.6, "cm")
    }

    matrix_grid <- patchwork::wrap_plots(
      grid,
      nrow    = k + 1L,
      ncol    = k + 1L,
      byrow   = TRUE,
      widths  = grid::unit.c(strip_w, rep(grid::unit(1, "null"), k)),
      heights = grid::unit.c(strip_h, rep(grid::unit(1, "null"), k))
    )
  }

  has_colourbar <- !(colour_by == "none" || !all(is.finite(colour_limits)))
  has_shape_leg <- !is.null(shape_legend_plot)

  # --- Top row: matrix + optional colour-bar ---
  top_row <- if (has_colourbar) {
    patchwork::wrap_plots(
      list(matrix_grid,
           .build_legend_plot(colour_by, colour_limits, palette)),
      nrow   = 1L,
      widths = grid::unit.c(grid::unit(1, "null"),
                            grid::unit(1.8, "cm"))
    )
  } else {
    matrix_grid
  }

  if (!has_shape_leg) return(top_row)

  # --- Bottom row: shape legend spans full width ---
  patchwork::wrap_plots(
    list(top_row, shape_legend_plot),
    ncol    = 1L,
    heights = grid::unit.c(grid::unit(1, "null"),
                           grid::unit(0.42, "null"))
  )
}

# Shape metrics for janusplot cells.
# Two continuous indices (monotonicity M, convexity C), two discrete
# counts (turning points, inflections), one ratio (flat range), and a
# rule-based classifier into 12 shape categories. All weights are the
# empirical density of x on the prediction grid, so the metrics describe
# the smooth *where the data actually live*, not the extrapolated tails.
#
# Exposed publicly via janusplot_shape_metrics() and
# janusplot_shape_cutoffs(). Internals are dot-prefixed.

# ---------------------------------------------------------------
# Default cutoff list. Tunable via the cutoffs argument.
# ---------------------------------------------------------------

#' Default cutoff thresholds for `shape_category` classification
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Returns the named list of thresholds used to map the continuous
#' monotonicity (`M`) and convexity (`C`) indices (plus inflection
#' counts) into a discrete `shape_category`. Expose so callers can
#' override individual thresholds or pass a fully custom list to
#' [janusplot()] / [janusplot_shape_metrics()].
#'
#' @param ... Optional named overrides to merge into the defaults.
#'
#' @returns A named list with numeric thresholds:
#' \describe{
#'   \item{`mono_strong`}{`|M|` threshold for a strictly monotone smooth (default `0.9`).}
#'   \item{`mono_mod`}{`|M|` threshold for a curved-but-monotone smooth (default `0.5`).}
#'   \item{`mono_nonmono`}{`|M|` below this is considered non-monotone (default `0.3`).}
#'   \item{`mono_s`}{`|M|` threshold for labelling an S-shape (default `0.5`).}
#'   \item{`curv_low`}{`|C|` below this is considered near-linear curvature (default `0.2`).}
#'   \item{`curv_mod`}{`|C|` threshold for a clearly curved monotone (default `0.5`).}
#'   \item{`curv_strong`}{`|C|` threshold for a U-shape / inverted-U shape
#'     (default `0.5`).}
#'   \item{`flat`}{`range(fit) / sd(y)` below this is called `flat` (default `0.05`).}
#' }
#'
#' @examples
#' janusplot_shape_cutoffs()
#' janusplot_shape_cutoffs(curv_mod = 0.6, flat = 0.02)
#' @family shape-metrics
#' @export
janusplot_shape_cutoffs <- function(...) {
  defaults <- list(
    mono_strong  = 0.9,
    mono_mod     = 0.5,
    mono_nonmono = 0.3,
    mono_s       = 0.5,
    curv_low     = 0.2,
    curv_mod     = 0.5,
    curv_strong  = 0.5,
    flat         = 0.05
  )
  overrides <- list(...)
  if (length(overrides)) {
    unknown <- setdiff(names(overrides), names(defaults))
    if (length(unknown)) {
      cli::cli_abort(c(
        "Unknown cutoff name{?s}: {.val {unknown}}.",
        i = "Allowed: {.val {names(defaults)}}."
      ))
    }
    for (nm in names(overrides)) {
      v <- overrides[[nm]]
      if (!is.numeric(v) || length(v) != 1L || !is.finite(v)) {
        cli::cli_abort("Cutoff {.val {nm}} must be a single finite number.")
      }
      defaults[[nm]] <- v
    }
  }
  defaults
}

# ---------------------------------------------------------------
# Canonical shape taxonomy: single source of truth for categories,
# glyphs, ASCII fallbacks, and one-line glosses. Consumed by the
# classifier, the cell renderer, and the shape-types legend.
# ---------------------------------------------------------------

.shape_taxonomy <- function() {
  # Order within the data frame drives legend layout (top-left to
  # bottom-right, row-major). Group monotone shapes, then peaks,
  # then waves, then multi-peak, with flat / indeterminate last.
  #
  # Four hierarchy columns layered from finest to coarsest:
  #   category  — 24-way fine label (primary dispatch)
  #   code      — unique 2-letter ASCII shorthand (cell-safe)
  #   archetype — 7-family grouping (SCAM / dose-response inspired)
  #   monotonic — monotone / non_monotone / degenerate
  #   linear    — linear / non_linear / degenerate
  data.frame(
    category  = c(
      "linear_up", "linear_down",
      "convex_up", "concave_up",
      "convex_down", "concave_down",
      "s_shape", "rippled_monotone",
      "u_shape", "inverted_u",
      "skewed_peak", "broad_peak", "rippled_peak",
      "wave", "warped_wave", "rippled_wave", "complex_wave",
      "bimodal", "bimodal_ripple",
      "bi_wave", "bi_wave_ripple",
      "complex",
      "flat", "indeterminate"
    ),
    code      = c(
      "lu", "ld",
      "vu", "cu", "vd", "cd",
      "ss", "rm",
      "us", "iu",
      "sp", "bp", "rp",
      "wv", "ww", "rw", "cw",
      "bm", "mr",
      "bw", "wr",
      "cx",
      "ft", "id"
    ),
    archetype = c(
      "monotone_linear", "monotone_linear",
      "monotone_curved", "monotone_curved",
      "monotone_curved", "monotone_curved",
      "monotone_curved", "monotone_curved",
      "unimodal", "unimodal",
      "unimodal", "unimodal", "unimodal",
      "wave", "wave", "wave", "wave",
      "multimodal", "multimodal",
      "multimodal", "multimodal",
      "chaotic",
      "degenerate", "degenerate"
    ),
    monotonic = c(
      "monotone", "monotone",
      "monotone", "monotone", "monotone", "monotone",
      "monotone", "monotone",
      "non_monotone", "non_monotone",
      "non_monotone", "non_monotone", "non_monotone",
      "non_monotone", "non_monotone", "non_monotone", "non_monotone",
      "non_monotone", "non_monotone",
      "non_monotone", "non_monotone",
      "non_monotone",
      "degenerate", "degenerate"
    ),
    linear    = c(
      "linear", "linear",
      "non_linear", "non_linear", "non_linear", "non_linear",
      "non_linear", "non_linear",
      "non_linear", "non_linear",
      "non_linear", "non_linear", "non_linear",
      "non_linear", "non_linear", "non_linear", "non_linear",
      "non_linear", "non_linear",
      "non_linear", "non_linear",
      "non_linear",
      "degenerate", "degenerate"
    ),
    glyph     = c(
      "\u2197", "\u2198",
      "\u2310", "\u2312", "\u2323", "\u00ac",
      "\u222b", "\u2248",
      "\u222a", "\u2229",
      "\u22cf", "\u2293", "\u223f",
      "\u223f", "\u219d", "\u2307", "\u2307",
      "MM", "MM", "MN", "MN", "X",
      "\u2014", "?"
    ),
    ascii     = c(
      "/", "\\",
      "J", "r", "u", "L",
      "S", "~",
      "U", "^",
      "P", "#", "P",
      "~", "~", "~", "~",
      "M", "M", "N", "N", "X",
      "-", "?"
    ),
    label     = c(
      "linear up", "linear down",
      "convex up", "concave up",
      "convex down", "concave down",
      "S-shape", "rippled monotone",
      "U-shape", "inverted U",
      "skewed peak", "broad peak", "rippled peak",
      "wave", "warped wave", "rippled wave", "complex wave",
      "bimodal", "bimodal + ripple",
      "bi-wave", "bi-wave + ripple", "complex",
      "flat", "indeterminate"
    ),
    gloss     = c(
      "monotone increasing, near-linear",
      "monotone decreasing, near-linear",
      "accelerating growth",
      "saturating growth (diminishing returns)",
      "decelerating decay",
      "accelerating decay",
      "monotone with one inflection",
      "monotone with multiple curvature flips",
      "non-monotone bowl (single minimum)",
      "non-monotone dome (single maximum)",
      "asymmetric single extremum, one inflection",
      "single extremum with shoulders, two inflections",
      "single extremum with ripples",
      "one full oscillation (sine-like)",
      "wave with asymmetric curvature",
      "wave with extra inflection",
      "wave with multiple ripples",
      "two extrema pairs (double peak)",
      "double peak with extra ripples",
      "two full oscillations",
      "bi-wave with extra ripples",
      "five or more extrema / inflections",
      "no meaningful variation",
      "fit unavailable or undefined"
    ),
    stringsAsFactors = FALSE
  )
}

# Quick lookup helper — given a vector of category names, returns the
# corresponding column (code / archetype / monotonic / linear) from
# the single source-of-truth taxonomy table.
.shape_lookup <- function(category, column) {
  tax <- .shape_taxonomy()
  idx <- match(category, tax$category)
  tax[[column]][idx]
}

#' Shape-category taxonomy table
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Return the full janusplot shape taxonomy as a data frame with four
#' hierarchy columns plus presentation fields. The taxonomy is the
#' single source of truth consumed by the classifier, the cell
#' renderer, the legend plate, and the `janusplot_data()` output.
#'
#' Hierarchy columns (finest → coarsest):
#' \describe{
#'   \item{`category`}{24-way fine label (`linear_up`, `skewed_peak`,
#'     `bimodal`, …). Computed per cell by [janusplot()].}
#'   \item{`code`}{Unique two-letter ASCII shorthand (safe on any
#'     font or typesetting pipeline) — e.g. `lu` for `linear_up`.}
#'   \item{`archetype`}{Seven-family grouping: `monotone_linear`,
#'     `monotone_curved`, `unimodal`, `wave`, `multimodal`,
#'     `chaotic`, `degenerate`.}
#'   \item{`monotonic`}{Three-way coarse classification: `monotone`
#'     / `non_monotone` / `degenerate`.}
#'   \item{`linear`}{Binary: `linear` / `non_linear` /
#'     `degenerate`.}
#' }
#'
#' The broader tiers (linear/non-linear, monotone/non-monotone) are
#' textbook calculus; the archetype layer maps cleanly to
#' shape-constrained regression vocabulary (Pya & Wood 2015;
#' Meyer 2008) and to dose-response shape categories (Calabrese
#' 2008; Calabrese & Baldwin 2001). The `(T, I)` dispatch
#' underlying each fine category is a coarsened Morse-theoretic
#' critical-point classification (Milnor 1963).
#'
#' @returns A data frame with 24 rows and columns `category`,
#'   `code`, `archetype`, `monotonic`, `linear`, `glyph`, `ascii`,
#'   `label`, `gloss`.
#'
#' @references
#' Calabrese, E. J. (2008). Hormesis: why it is important to
#'   toxicology and toxicologists. *Environmental Toxicology and
#'   Chemistry*, **27**(7), 1451–1474.
#'
#' Meyer, M. C. (2008). Inference using shape-restricted regression
#'   splines. *Annals of Applied Statistics*, **2**(3), 1013–1033.
#'
#' Milnor, J. (1963). *Morse Theory*. Princeton University Press.
#'
#' Pya, N., & Wood, S. N. (2015). Shape constrained additive models.
#'   *Statistics and Computing*, **25**(3), 543–559.
#'
#' @examples
#' tax <- janusplot_shape_hierarchy()
#' head(tax[, c("category", "code", "archetype", "monotonic", "linear")])
#' # Count how many categories live in each archetype
#' table(tax$archetype)
#' @family shape-metrics
#' @export
janusplot_shape_hierarchy <- function() {
  .shape_taxonomy()
}

# Canonical (x, y) thumbnail for each shape category. The legend
# renders these as mini-panels, which is font-independent and more
# visually informative than a Unicode glyph. All thumbnails live on
# x in [0, 1] with y normalised to [0, 1] at draw time.
.shape_thumbnail_df <- function(category, n = 80L) {
  x <- seq(0, 1, length.out = as.integer(n))
  y <- switch(category,
    linear_up        = x,
    linear_down      = 1 - x,
    convex_up        = x^2,
    concave_up       = sqrt(x),
    convex_down      = (1 - x)^2,
    concave_down     = sqrt(1 - x),
    s_shape          = 1 / (1 + exp(-10 * (x - 0.5))),
    rippled_monotone = x + 0.08 * sin(8 * pi * x),
    u_shape          = 4 * (x - 0.5)^2,
    inverted_u       = 1 - 4 * (x - 0.5)^2,
    skewed_peak      = (x * exp(1 - 3 * x)) * (3 * exp(1)) / exp(2),
    broad_peak       = exp(-((x - 0.5) * 3)^4),
    rippled_peak     = exp(-((x - 0.5) * 3)^2) +
                         0.08 * sin(12 * pi * x),
    wave             = 0.5 + 0.5 * sin(2 * pi * x - pi / 2),
    warped_wave      = 0.5 + 0.5 * sin(2 * pi * (x^1.4) - pi / 2),
    rippled_wave     = 0.5 + 0.5 * sin(2 * pi * x - pi / 2) +
                         0.07 * sin(8 * pi * x),
    complex_wave     = 0.5 + 0.5 * sin(2 * pi * x - pi / 2) +
                         0.18 * sin(10 * pi * x),
    bimodal          = exp(-((x - 0.25) * 6)^2) +
                         exp(-((x - 0.75) * 6)^2),
    bimodal_ripple   = exp(-((x - 0.25) * 6)^2) +
                         exp(-((x - 0.75) * 6)^2) +
                         0.05 * sin(14 * pi * x),
    bi_wave          = 0.5 + 0.5 * sin(4 * pi * x - pi / 2),
    bi_wave_ripple   = 0.5 + 0.5 * sin(4 * pi * x - pi / 2) +
                         0.07 * sin(14 * pi * x),
    complex          = 0.5 + 0.35 * sin(6 * pi * x) *
                         cos(3 * pi * x),
    flat             = rep(0.5, length(x)),
    indeterminate    = rep(NA_real_, length(x)),
    rep(NA_real_, length(x))
  )
  data.frame(x = x, y = y, category = category,
             stringsAsFactors = FALSE)
}

# Glyph lookup (Unicode by default; "ascii" overrides for
# font-coverage fallback).
.shape_glyph <- function(category, style = c("unicode", "ascii")) {
  style <- rlang::arg_match(style)
  tax   <- .shape_taxonomy()
  idx   <- match(category, tax$category)
  col   <- if (style == "ascii") "ascii" else "glyph"
  ifelse(is.na(idx), "?", tax[[col]][idx])
}

# ---------------------------------------------------------------
# Rule-based classifier. Inputs must be finite scalars or NA.
# ---------------------------------------------------------------

# Dispatch for the monotone (T=0, I=0) cell — no turning points, no
# inflections. Uses the (M, C) indices to separate strict-linear from
# curved monotone, with orientation from sign(M) and curvature from
# sign(C).
.classify_monotone <- function(M, C, cutoffs) {
  if (abs(M) > cutoffs$mono_strong && abs(C) < cutoffs$curv_low) {
    return(if (M > 0) "linear_up" else "linear_down")
  }
  if (M >  cutoffs$mono_mod && C >  cutoffs$curv_mod) return("convex_up")
  if (M >  cutoffs$mono_mod && C < -cutoffs$curv_mod) return("concave_up")
  if (M < -cutoffs$mono_mod && C >  cutoffs$curv_mod) return("convex_down")
  if (M < -cutoffs$mono_mod && C < -cutoffs$curv_mod) return("concave_down")
  if (M >  0) return("linear_up")
  if (M <  0) return("linear_down")
  "flat"
}

# Primary classifier: dispatch on (T, I) = (n_turning_points,
# n_inflections). (M, C) only disambiguates the `(0, 0)` and `(1, 0)`
# cells. See .shape_taxonomy() for the category list + glosses.
.classify_shape <- function(M, C, n_turn, n_inflect, flat_ratio, cutoffs) {
  if (is.na(M) || is.na(C)) return("indeterminate")
  if (!is.na(flat_ratio) && flat_ratio < cutoffs$flat) return("flat")
  n_T <- if (is.na(n_turn))    0L else as.integer(n_turn)
  n_I <- if (is.na(n_inflect)) 0L else as.integer(n_inflect)
  if (n_T >= 5L || n_I >= 5L) return("complex")

  if (n_T == 0L) {
    if (n_I == 0L) return(.classify_monotone(M, C, cutoffs))
    if (n_I == 1L) return("s_shape")
    return("rippled_monotone")        # n_I >= 2
  }
  if (n_T == 1L) {
    if (n_I == 0L) return(if (C > 0) "u_shape" else "inverted_u")
    if (n_I == 1L) return("skewed_peak")
    if (n_I == 2L) return("broad_peak")
    return("rippled_peak")             # n_I >= 3
  }
  if (n_T == 2L) {
    if (n_I <= 1L) return("wave")
    if (n_I == 2L) return("warped_wave")
    if (n_I == 3L) return("rippled_wave")
    return("complex_wave")             # n_I >= 4
  }
  if (n_T == 3L) {
    if (n_I >= 3L) return("bimodal_ripple")
    return("bimodal")                  # n_I in {0, 1, 2}
  }
  if (n_T == 4L) {
    if (n_I >= 3L) return("bi_wave_ripple")
    return("bi_wave")                  # n_I in {0, 1, 2}
  }
  "complex"
}

# ---------------------------------------------------------------
# Numerical derivatives on an equally-spaced grid.
# Central differences inside, forward/backward at the endpoints.
# ---------------------------------------------------------------

.central_diff <- function(x, y) {
  n <- length(y)
  if (n < 3L) return(rep(NA_real_, n))
  d <- numeric(n)
  d[1L]  <- (y[2L] - y[1L])     / (x[2L] - x[1L])
  d[n]   <- (y[n]  - y[n - 1L]) / (x[n]  - x[n - 1L])
  idx    <- 2:(n - 1L)
  d[idx] <- (y[idx + 1L] - y[idx - 1L]) / (x[idx + 1L] - x[idx - 1L])
  d
}

# Robust sign-change count for a derivative vector. A candidate sign
# change at index j is only counted when the integrated |v| in the
# lobe on either side exceeds `min_lobe_frac` of the total integrated
# |v|. This discards tail-noise wiggles that would otherwise inflate
# inflection / turning-point counts for spline-smoothed saturating
# curves and similar shapes.
.n_sign_changes <- function(v, min_lobe_frac = 0.1, rel_tol = 1e-2) {
  v <- v[is.finite(v)]
  n <- length(v)
  if (n < 2L) return(0L)
  scale_v <- max(abs(v))
  if (!is.finite(scale_v) || scale_v == 0) return(0L)
  total <- sum(abs(v))
  if (!is.finite(total) || total == 0) return(0L)

  # Suppress noise floor before sign extraction.
  v_clean <- ifelse(abs(v) < rel_tol * scale_v, 0, v)
  s <- sign(v_clean)
  # Forward-fill zero runs so a candidate crossing sits between two
  # non-zero sign regions rather than swallowing the boundary.
  prev <- 0
  for (i in seq_along(s)) {
    if (s[i] == 0) s[i] <- prev else prev <- s[i]
  }
  if (all(s == 0)) return(0L)

  change_idx <- which(diff(s) != 0) + 1L
  if (!length(change_idx)) return(0L)

  # Integrate |v| between consecutive change boundaries.
  boundaries <- c(0L, change_idx - 1L, n)
  lobe_mass  <- numeric(length(boundaries) - 1L)
  for (k in seq_along(lobe_mass)) {
    lo <- boundaries[k] + 1L
    hi <- boundaries[k + 1L]
    if (hi < lo) next
    lobe_mass[k] <- sum(abs(v[lo:hi]))
  }
  lobe_frac <- lobe_mass / total

  # Count only crossings where BOTH adjacent lobes carry substantial
  # mass.
  real <- 0L
  for (j in seq_along(change_idx)) {
    if (lobe_frac[j] >= min_lobe_frac &&
        lobe_frac[j + 1L] >= min_lobe_frac) {
      real <- real + 1L
    }
  }
  real
}

# Empirical x-density evaluated on the prediction grid. Trapezoidal
# integration weights built in: we normalise so sum(w) = 1.
.grid_weights <- function(raw_x, grid_x) {
  raw_x <- raw_x[is.finite(raw_x)]
  if (length(raw_x) < 2L || diff(range(raw_x)) == 0) {
    w <- rep(1, length(grid_x))
    return(w / sum(w))
  }
  bw <- tryCatch(stats::bw.nrd0(raw_x), error = function(e) {
    stats::sd(raw_x) * (4 / (3 * length(raw_x)))^(1 / 5)
  })
  if (!is.finite(bw) || bw <= 0) bw <- diff(range(raw_x)) / 20
  dens <- tryCatch(
    stats::density(raw_x, bw = bw,
                   from = min(grid_x), to = max(grid_x),
                   n = max(512L, length(grid_x))),
    error = function(e) NULL
  )
  if (is.null(dens)) {
    w <- rep(1, length(grid_x))
    return(w / sum(w))
  }
  w <- stats::approx(dens$x, dens$y, xout = grid_x, rule = 2L)$y
  w[!is.finite(w) | w < 0] <- 0
  if (sum(w) <= 0) {
    w <- rep(1, length(grid_x))
  }
  w / sum(w)
}

# ---------------------------------------------------------------
# Raw metric computation (no classification). Cached on each pair
# fit so render-time reclassification under different cutoffs is
# free. Returns M, C, counts, flat ratio. Never errors.
# ---------------------------------------------------------------

.na_shape_raw <- function() {
  list(
    monotonicity_index = NA_real_,
    convexity_index    = NA_real_,
    n_turning_points   = NA_integer_,
    n_inflections      = NA_integer_,
    flat_range_ratio   = NA_real_
  )
}

.compute_shape_metrics_raw <- function(fit_obj) {
  if (is.null(fit_obj) || !is.list(fit_obj)) return(.na_shape_raw())
  pred <- fit_obj$pred
  if (is.null(pred) || nrow(pred) < 3L) return(.na_shape_raw())
  if (all(is.na(pred$fit))) return(.na_shape_raw())

  x  <- pred$x
  y  <- pred$fit
  ok <- is.finite(x) & is.finite(y)
  if (sum(ok) < 3L) return(.na_shape_raw())
  x <- x[ok]
  y <- y[ok]

  raw_x <- if (!is.null(fit_obj$raw) && nrow(fit_obj$raw) > 0L) {
    fit_obj$raw[[fit_obj$x_name]]
  } else {
    x
  }
  w <- .grid_weights(raw_x, x)

  d1 <- .central_diff(x, y)
  d2 <- .central_diff(x, d1)

  abs_d1_sum <- sum(w * abs(d1), na.rm = TRUE)
  abs_d2_sum <- sum(w * abs(d2), na.rm = TRUE)
  # Zero integrated slope / curvature means a truly flat or truly
  # linear smooth; in both cases the corresponding index is 0, not
  # NA. Only return NA when the raw derivatives themselves were NA.
  M <- if (is.finite(abs_d1_sum) && abs_d1_sum > 0) {
    sum(w * d1, na.rm = TRUE) / abs_d1_sum
  } else if (is.finite(abs_d1_sum)) {
    0
  } else {
    NA_real_
  }
  C <- if (is.finite(abs_d2_sum) && abs_d2_sum > 0) {
    sum(w * d2, na.rm = TRUE) / abs_d2_sum
  } else if (is.finite(abs_d2_sum)) {
    0
  } else {
    NA_real_
  }
  # Dimensionless curvature-signal: compares integrated |d2| against
  # integrated |d1| scaled by the x-range, so the test is invariant
  # to absolute units. When the second derivative is numerical noise
  # (linear smooth), this collapses to zero and we suppress both C
  # and the inflection count.
  x_range <- diff(range(x, na.rm = TRUE))
  y_range <- diff(range(y, na.rm = TRUE))
  curv_signal <- if (is.finite(x_range) && x_range > 0 &&
                     is.finite(abs_d1_sum) && abs_d1_sum > 0) {
    abs_d2_sum * x_range / abs_d1_sum
  } else {
    NA_real_
  }
  slope_signal <- if (is.finite(x_range) && x_range > 0 &&
                      is.finite(y_range) && y_range > 0) {
    abs_d1_sum * x_range / y_range
  } else {
    NA_real_
  }
  if (is.finite(curv_signal) && curv_signal < 1e-3 && is.finite(C)) C <- 0

  # Count sign changes only when the underlying derivative carries
  # meaningful signal; otherwise noise-driven sign flips inflate
  # counts and mis-classify linear / flat smooths as `complex`.
  n_turn <- if (is.finite(slope_signal) && slope_signal < 1e-3) {
    0L
  } else {
    .n_sign_changes(d1)
  }
  n_inflect <- if (is.finite(curv_signal) && curv_signal < 1e-3) {
    0L
  } else {
    .n_sign_changes(d2)
  }

  raw_y <- if (!is.null(fit_obj$raw) && !is.null(fit_obj$y_name) &&
               !is.null(fit_obj$raw[[fit_obj$y_name]])) {
    fit_obj$raw[[fit_obj$y_name]]
  } else {
    y
  }
  sd_y <- stats::sd(raw_y, na.rm = TRUE)
  flat_ratio <- if (!is.finite(sd_y) || sd_y == 0) {
    0
  } else {
    y_range / sd_y
  }

  list(
    monotonicity_index = unname(M),
    convexity_index    = unname(C),
    n_turning_points   = as.integer(n_turn),
    n_inflections      = as.integer(n_inflect),
    flat_range_ratio   = unname(flat_ratio)
  )
}

# Classified variant — raw + shape_category via user cutoffs.
.compute_shape_metrics <- function(fit_obj,
                                   cutoffs = janusplot_shape_cutoffs()) {
  raw <- .compute_shape_metrics_raw(fit_obj)
  raw$shape_category <- .classify_shape(
    raw$monotonicity_index, raw$convexity_index,
    raw$n_turning_points, raw$n_inflections,
    raw$flat_range_ratio, cutoffs
  )
  raw
}

# ---------------------------------------------------------------
# Public wrapper: compute shape metrics directly from a fitted
# mgcv::gam on a single predictor, or from the list returned by a
# pair-fit helper.
# ---------------------------------------------------------------

#' Shape metrics for a fitted univariate smooth
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Compute the continuous monotonicity and convexity indices, inflection
#' and turning-point counts, and rule-based shape category for a fitted
#' univariate smooth. Works on either a per-pair fit object returned
#' from the janusplot internal machinery or a freshly fitted
#' [mgcv::gam()] with a single `s()` term.
#'
#' Both indices are bounded in `[-1, 1]` and weighted by the empirical
#' density of the predictor:
#' * `monotonicity_index` (paper symbol `M`). Let `f` be the fitted
#'   smooth evaluated on a dense grid of `n_grid` equally-spaced
#'   points across the predictor range, `f'` its numerical first
#'   derivative, and `w` the empirical density of the predictor on
#'   the same grid with `sum(w) = 1`. Then
#'   `monotonicity_index = sum(w * f') / sum(w * |f'|) in [-1, 1]`.
#'   `+1` is strictly increasing, `-1` strictly decreasing, `0`
#'   non-monotone.
#' * `convexity_index` (paper symbol `C`). With `f''` the numerical
#'   second derivative on the same grid,
#'   `convexity_index = sum(w * f'') / sum(w * |f''|) in [-1, 1]`.
#'   `+1` is globally convex (bowl-up), `-1` globally concave
#'   (bowl-down), `0` inflection-dominated (S-curve, sine, flat).
#'
#' Both indices are scale-invariant (replacing `y -> a*y + b` leaves
#' them unchanged) and density-weighted so they describe the smooth
#' *where the data actually live*, not extrapolated tails.
#'
#' @param fit Either a list returned by a janusplot pair-fit helper
#'   (must contain `pred` and `raw`), or a fitted [mgcv::gam()] with
#'   a single `s(x)` term.
#' @param x_name Character. Column name of the predictor when `fit` is
#'   a [mgcv::gam()] object. Ignored for pair-fit lists.
#' @param newdata Optional data frame supplying the raw predictor values
#'   used for density weighting when `fit` is a [mgcv::gam()] object.
#'   If `NULL`, the model frame is used.
#' @param n_grid Integer. Prediction grid length when `fit` is a
#'   [mgcv::gam()] object. Default `200L`.
#' @param cutoffs Named list of classification thresholds; see
#'   [janusplot_shape_cutoffs()]. Default uses package defaults.
#'
#' @returns A named list with components:
#'   \describe{
#'     \item{`monotonicity_index`}{`M` in `[-1, 1]`. See Description.}
#'     \item{`convexity_index`}{`C` in `[-1, 1]`. See Description.}
#'     \item{`n_turning_points`}{Integer count of lobe-mass-weighted
#'       sign changes of `f'`. Equals the number of interior extrema.}
#'     \item{`n_inflections`}{Integer count of lobe-mass-weighted
#'       sign changes of `f''`.}
#'     \item{`flat_range_ratio`}{`range(f) / sd(y)` — small values
#'       indicate a degenerate flat smooth.}
#'     \item{`shape_category`}{One of 24 labels from
#'       [janusplot_shape_hierarchy()] dispatched on
#'       `(n_turning_points, n_inflections)` with
#'       `(monotonicity_index, convexity_index)` disambiguation for
#'       the monotone case.}
#'   }
#'
#' @seealso [janusplot_shape_cutoffs()], [janusplot()], [janusplot_data()].
#'
#' @examples
#' # On a fitted gam
#' set.seed(2026L)
#' n  <- 200L
#' x  <- stats::runif(n, 0, 10)
#' y  <- log1p(x) + stats::rnorm(n, sd = 0.3)
#' d  <- data.frame(x = x, y = y)
#' fit <- mgcv::gam(y ~ s(x), data = d, method = "REML")
#' janusplot_shape_metrics(fit, x_name = "x", newdata = d)
#' @family shape-metrics
#' @export
janusplot_shape_metrics <- function(fit,
                                    x_name  = NULL,
                                    newdata = NULL,
                                    n_grid  = 200L,
                                    cutoffs = janusplot_shape_cutoffs()) {
  if (is.list(fit) && !inherits(fit, "gam") &&
      !is.null(fit$pred) && !is.null(fit$x_name)) {
    return(.compute_shape_metrics(fit, cutoffs = cutoffs))
  }
  if (!inherits(fit, "gam")) {
    cli::cli_abort(
      "{.arg fit} must be a fitted {.cls gam} or a janusplot pair-fit list."
    )
  }
  if (is.null(x_name) || !is.character(x_name) || length(x_name) != 1L) {
    cli::cli_abort(
      "{.arg x_name} must be a single character naming the predictor."
    )
  }
  mf <- stats::model.frame(fit)
  if (is.null(newdata)) newdata <- mf
  if (!x_name %in% names(newdata)) {
    cli::cli_abort(
      "Column {.val {x_name}} not found in {.arg newdata} or model frame."
    )
  }
  xv <- newdata[[x_name]]
  xv <- xv[is.finite(xv)]
  if (length(xv) < 3L) {
    cli::cli_abort("Need at least 3 finite values of {.val {x_name}}.")
  }
  x_grid <- seq(min(xv), max(xv), length.out = as.integer(n_grid))
  nd     <- data.frame(placeholder = x_grid)
  names(nd) <- x_name
  # Hold any additional covariates at typical values (mean for numeric,
  # mode for factor). Mirrors .typical_newdata() from internal-fit.R.
  for (nm in setdiff(names(mf), c(x_name, all.vars(fit$formula)[1L]))) {
    col <- mf[[nm]]
    nd[[nm]] <- if (is.numeric(col)) mean(col, na.rm = TRUE) else
      if (is.factor(col)) factor(levels(col)[which.max(tabulate(col))],
                                 levels = levels(col)) else
                                   names(sort(table(col), decreasing = TRUE))[1L]
  }
  preds <- stats::predict(fit, newdata = nd, se.fit = TRUE)
  y_name <- all.vars(fit$formula)[1L]
  pred_df <- data.frame(
    x   = x_grid,
    fit = as.numeric(preds$fit),
    se  = as.numeric(preds$se.fit),
    lo  = as.numeric(preds$fit) - 1.96 * as.numeric(preds$se.fit),
    hi  = as.numeric(preds$fit) + 1.96 * as.numeric(preds$se.fit)
  )
  raw_df <- newdata[, intersect(c(x_name, y_name), names(newdata)),
                    drop = FALSE]
  synth <- list(
    x_name = x_name, y_name = y_name,
    pred   = pred_df, raw = raw_df
  )
  .compute_shape_metrics(synth, cutoffs = cutoffs)
}

# Shape-metric unit tests.
# We build synthetic fit_obj-like lists directly (no mgcv fit needed)
# so tests are deterministic, fast, and isolate the metric logic from
# GAM smoothing noise.

# Build a minimal fit object matching what .compute_shape_metrics_raw()
# expects: pred data frame (x, fit, se, lo, hi) + raw data frame.
.mk_fit <- function(x, y, y_name = "y", x_name = "x") {
  list(
    x_name = x_name,
    y_name = y_name,
    pred   = data.frame(x = x, fit = y,
                        se = 0, lo = y, hi = y),
    raw    = setNames(data.frame(x, y), c(x_name, y_name))
  )
}

test_that("linear increasing recovers M ~ 1, C ~ 0, linear_up", {
  x <- seq(0, 10, length.out = 200L)
  y <- 2 * x
  m <- janusplot:::.compute_shape_metrics(.mk_fit(x, y))
  expect_equal(m$monotonicity_index, 1, tolerance = 1e-6)
  expect_lt(abs(m$convexity_index), 0.1)
  expect_equal(m$shape_category, "linear_up")
})

test_that("linear decreasing recovers M ~ -1, linear_down", {
  x <- seq(0, 10, length.out = 200L)
  y <- -1.5 * x + 7
  m <- janusplot:::.compute_shape_metrics(.mk_fit(x, y))
  expect_equal(m$monotonicity_index, -1, tolerance = 1e-6)
  expect_equal(m$shape_category, "linear_down")
})

test_that("y = x^2 on [0, 10] recovers convex_up", {
  x <- seq(0, 10, length.out = 200L)
  y <- x^2
  m <- janusplot:::.compute_shape_metrics(.mk_fit(x, y))
  expect_gt(m$monotonicity_index, 0.5)
  expect_gt(m$convexity_index, 0.5)
  expect_equal(m$shape_category, "convex_up")
})

test_that("y = log1p(x) recovers concave_up (saturating growth)", {
  x <- seq(0.01, 10, length.out = 200L)
  y <- log1p(x)
  m <- janusplot:::.compute_shape_metrics(.mk_fit(x, y))
  expect_gt(m$monotonicity_index, 0.5)
  expect_lt(m$convexity_index, -0.3)
  expect_equal(m$shape_category, "concave_up")
})

test_that("inverted parabola on [0, 10] recovers inverted_u", {
  x <- seq(0, 10, length.out = 200L)
  y <- -(x - 5)^2
  m <- janusplot:::.compute_shape_metrics(.mk_fit(x, y))
  expect_lt(abs(m$monotonicity_index), 0.3)
  expect_lt(m$convexity_index, -0.5)
  expect_equal(m$shape_category, "inverted_u")
})

test_that("upright parabola on [0, 10] recovers u_shape", {
  x <- seq(0, 10, length.out = 200L)
  y <- (x - 5)^2
  m <- janusplot:::.compute_shape_metrics(.mk_fit(x, y))
  expect_lt(abs(m$monotonicity_index), 0.3)
  expect_gt(m$convexity_index, 0.5)
  expect_equal(m$shape_category, "u_shape")
})

test_that("tanh on [-5, 5] recovers s_shape", {
  x <- seq(-5, 5, length.out = 200L)
  y <- tanh(x)
  m <- janusplot:::.compute_shape_metrics(.mk_fit(x, y))
  expect_equal(m$n_inflections, 1L)
  expect_gt(abs(m$monotonicity_index), 0.5)
  expect_equal(m$shape_category, "s_shape")
})

test_that("flat line is labelled flat", {
  x <- seq(0, 10, length.out = 200L)
  y <- rep(3, 200L)
  m <- janusplot:::.compute_shape_metrics(.mk_fit(x, y))
  expect_equal(m$shape_category, "flat")
})

test_that("NA fit yields indeterminate category", {
  x <- seq(0, 10, length.out = 10L)
  y <- rep(NA_real_, 10L)
  m <- janusplot:::.compute_shape_metrics(.mk_fit(x, y))
  expect_equal(m$shape_category, "indeterminate")
})

test_that("janusplot_shape_cutoffs accepts valid overrides", {
  co <- janusplot_shape_cutoffs(curv_mod = 0.6, flat = 0.01)
  expect_equal(co$curv_mod, 0.6)
  expect_equal(co$flat, 0.01)
  expect_equal(co$mono_strong, 0.9)  # default retained
})

test_that("janusplot_shape_cutoffs rejects unknown names and bad types", {
  expect_error(janusplot_shape_cutoffs(not_a_cutoff = 0.5))
  expect_error(janusplot_shape_cutoffs(curv_mod = "big"))
  expect_error(janusplot_shape_cutoffs(curv_mod = c(0.4, 0.6)))
})

test_that("janusplot_shape_metrics() works on a freshly fitted gam", {
  withr::with_seed(2026L, {
    n <- 200L
    x <- stats::runif(n, 0.01, 10)
    y <- log1p(x) + stats::rnorm(n, sd = 0.2)
    d <- data.frame(x = x, y = y)
  })
  fit <- mgcv::gam(y ~ s(x), data = d, method = "REML")
  m   <- janusplot_shape_metrics(fit, x_name = "x", newdata = d)
  # Saturating growth: monotone increasing; mgcv may recover this
  # as linear_up (if smoothing penalty dominates), concave_up (if
  # curvature is detected), or s_shape (if it overshoots once).
  expect_true(m$shape_category %in%
                c("linear_up", "concave_up", "s_shape"))
  expect_gt(m$monotonicity_index, 0.5)
})

test_that("janusplot_shape_metrics() rejects bogus inputs", {
  expect_error(janusplot_shape_metrics("not a fit"))
  fit <- mgcv::gam(mpg ~ s(wt), data = mtcars, method = "REML")
  expect_error(janusplot_shape_metrics(fit, x_name = NULL))
  expect_error(janusplot_shape_metrics(fit, x_name = "no_such_column"))
})

test_that(".shape_taxonomy has unique category names and 24 rows", {
  tax <- janusplot:::.shape_taxonomy()
  expect_equal(nrow(tax), 24L)
  expect_equal(length(unique(tax$category)), 24L)
  ascii_bytes <- vapply(tax$ascii, function(s) all(charToRaw(s) < as.raw(128L)),
                        logical(1L))
  expect_true(all(ascii_bytes))
})

test_that("taxonomy hierarchy columns are complete, unique, and ASCII", {
  tax <- janusplot:::.shape_taxonomy()
  # 2-letter codes must be unique and ASCII
  expect_equal(length(unique(tax$code)), 24L)
  expect_true(all(nchar(tax$code) == 2L))
  code_bytes <- vapply(tax$code, function(s) all(charToRaw(s) < as.raw(128L)),
                       logical(1L))
  expect_true(all(code_bytes))
  # Archetype must be one of 7 values
  expect_true(all(tax$archetype %in%
                    c("monotone_linear", "monotone_curved",
                      "unimodal", "wave", "multimodal",
                      "chaotic", "degenerate")))
  # monotonic column restricted to 3 values
  expect_true(all(tax$monotonic %in%
                    c("monotone", "non_monotone", "degenerate")))
  # linear column restricted to 3 values
  expect_true(all(tax$linear %in%
                    c("linear", "non_linear", "degenerate")))
  # No NA anywhere in the hierarchy columns
  expect_false(any(is.na(tax$code)))
  expect_false(any(is.na(tax$archetype)))
  expect_false(any(is.na(tax$monotonic)))
  expect_false(any(is.na(tax$linear)))
})

test_that("janusplot_shape_hierarchy() is the public mirror of the taxonomy", {
  expect_identical(janusplot_shape_hierarchy(), janusplot:::.shape_taxonomy())
  expect_true(all(c("category", "code", "archetype", "monotonic",
                    "linear", "label", "gloss") %in%
                    names(janusplot_shape_hierarchy())))
})

test_that("degenerate categories flagged as degenerate in all tiers", {
  tax <- janusplot_shape_hierarchy()
  deg <- tax[tax$category %in% c("flat", "indeterminate"), ]
  expect_true(all(deg$archetype == "degenerate"))
  expect_true(all(deg$monotonic == "degenerate"))
  expect_true(all(deg$linear    == "degenerate"))
})

test_that("linear_up / linear_down are the only `linear` linear-tier rows", {
  tax <- janusplot_shape_hierarchy()
  expect_setequal(tax$category[tax$linear == "linear"],
                  c("linear_up", "linear_down"))
})

test_that("skewed_peak recovered from y = x * exp(-3x)", {
  x <- seq(0, 1, length.out = 200L)
  y <- x * exp(-3 * x)
  m <- janusplot:::.compute_shape_metrics(.mk_fit(x, y))
  expect_equal(m$n_turning_points, 1L)
  # (1, 0) would be inverted_u/u_shape; a true skewed_peak wants
  # at least one inflection on the asymmetric descent.
  expect_true(m$shape_category %in%
                c("skewed_peak", "inverted_u", "broad_peak"))
})

test_that("broad_peak recovered from plateau exp(-((x-.5)*3)^4)", {
  x <- seq(0, 1, length.out = 200L)
  y <- exp(-((x - 0.5) * 3)^4)
  m <- janusplot:::.compute_shape_metrics(.mk_fit(x, y))
  expect_equal(m$n_turning_points, 1L)
  expect_true(m$shape_category %in% c("broad_peak", "inverted_u"))
})

test_that("wave recovered from sin(2pi x)", {
  x <- seq(0, 1, length.out = 200L)
  y <- sin(2 * pi * x)
  m <- janusplot:::.compute_shape_metrics(.mk_fit(x, y))
  expect_equal(m$n_turning_points, 2L)
  expect_equal(m$shape_category, "wave")
})

test_that("bi_wave recovered from sin(4pi x)", {
  x <- seq(0, 1, length.out = 400L)
  y <- sin(4 * pi * x)
  m <- janusplot:::.compute_shape_metrics(.mk_fit(x, y))
  expect_equal(m$n_turning_points, 4L)
  expect_true(m$shape_category %in%
                c("bi_wave", "bi_wave_ripple"))
})

test_that("bimodal recovered from two-gaussian mixture", {
  x <- seq(0, 1, length.out = 400L)
  y <- exp(-((x - 0.25) * 6)^2) + exp(-((x - 0.75) * 6)^2)
  m <- janusplot:::.compute_shape_metrics(.mk_fit(x, y))
  expect_equal(m$n_turning_points, 3L)
  expect_true(m$shape_category %in% c("bimodal", "bimodal_ripple"))
})

#' Asymmetric smoothed-association matrix
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Render a pairwise, asymmetric matrix of smoothed associations between
#' numeric variables. Each cell \[i, j\] where `i != j` shows the fitted
#' spline from [mgcv::gam()]:
#' * Upper triangle (`i < j`): `gam(x_j ~ s(x_i) + <adjust>)`.
#' * Lower triangle (`i > j`): `gam(x_i ~ s(x_j) + <adjust>)`.
#' * Diagonal: blank panel when labels live on the border (default),
#'   or a variable-name label when `labels = "diagonal"`.
#'
#' The two triangles intentionally differ — the asymmetry reveals
#' heteroscedasticity, leverage, and directional non-linearity that a
#' single scalar correlation hides.
#'
#' @param data A data frame with numeric columns to include.
#' @param vars Character vector of column names to use. `NULL` (default)
#'   uses all numeric columns in `data`. Non-numeric columns trigger an
#'   error listing offenders.
#' @param adjust A one-sided formula RHS giving additional covariates
#'   and/or random effects to include in every pairwise GAM. For example,
#'   `adjust = ~ s(age) + s(site, bs = "re")` fits
#'   `gam(y ~ s(x) + s(age) + s(site, bs = "re"))` for each pair.
#'   Default `NULL` fits unadjusted pairwise smooths.
#' @param method Smoothing-parameter estimation method passed to
#'   [mgcv::gam()]. Default `"REML"` per mgcv recommendation.
#' @param k Integer, or named list mapping variable names to integers.
#'   Basis dimension for `s()`. Default `-1L` (mgcv's automatic choice).
#' @param bs Basis type for `s()`. Default `"tp"` (thin plate).
#' @param order One of `"original"` (default), `"hclust"` (reorder by
#'   hierarchical clustering of Pearson correlations), or `"alphabetical"`.
#' @param show_data Logical. If `TRUE` (default), overlay raw data
#'   points (low alpha) behind each spline. Only applies when
#'   `display = "fit"`; derivative panels never overlay raw data.
#' @param show_ci Logical. If `TRUE` (default), overlay the 95%
#'   confidence envelope from `predict(gam, se.fit = TRUE)` on the
#'   fit panel (i.e. when `display = "fit"`). CI rendering on
#'   derivative panels is controlled separately by `derivative_ci`.
#' @param display One of `"fit"` (default), `"d1"`, or `"d2"`.
#'   Selects which single quantity is rendered in every
#'   off-diagonal cell of the matrix.
#'   * `"fit"` — the fitted smooth \eqn{\hat f(x)}; default,
#'     behaviour identical to the pre-derivative release.
#'   * `"d1"` — the first derivative \eqn{\hat f'(x)} of the
#'     fitted smooth. Zero crossings localise turning points of
#'     \eqn{\hat f}.
#'   * `"d2"` — the second derivative \eqn{\hat f''(x)}. Zero
#'     crossings localise inflection points of \eqn{\hat f}.
#'
#'   A single matrix shows a single quantity by design: stacked
#'   multi-panel cells crowd the matrix at any realistic variable
#'   count. To compare fit against derivative, render two or three
#'   `janusplot()` calls side-by-side; each call keeps its own
#'   `with_data = TRUE` summary table tagged with the `display`
#'   column.
#'
#'   Orders \eqn{k \ge 3} are not exposed — higher-order derivatives
#'   of penalised regression splines amplify noise and rarely carry
#'   usable signal at realistic sample sizes. See
#'   `vignette("janusplot")` for the theoretical justification and
#'   applied use-cases.
#' @param derivative_ci One of `"none"` (default), `"pointwise"`, or
#'   `"simultaneous"`. Controls whether — and how — a 95%
#'   confidence ribbon is drawn underneath the derivative curve when
#'   `display %in% c("d1", "d2")`. Ignored when `display = "fit"`.
#'   * `"none"` — no ribbon. The curve and the zero reference line
#'     are all you see. Default, because pointwise ribbons overshoot
#'     nominal coverage as a joint region and can invite
#'     over-reading of local features.
#'   * `"pointwise"` — 95% pointwise ribbon from
#'     \eqn{\sqrt{\mathrm{diag}(D V_p D^\top)}} (Wood 2017 §7.2.4).
#'     Valid marginally; not a simultaneous statement.
#'   * `"simultaneous"` — 95% simultaneous band via the Monte Carlo
#'     construction of Ruppert, Wand & Carroll (2003) popularised for
#'     GAMs by Simpson (2018, *Frontiers Ecol. Evol.* 6:149): draw
#'     \eqn{B} samples \eqn{\tilde{\boldsymbol\beta} \sim
#'     \mathcal{N}(\hat{\boldsymbol\beta}, V_p)}, compute
#'     \eqn{\max_x |D_i(\tilde{\boldsymbol\beta} -
#'     \hat{\boldsymbol\beta})| / \mathrm{se}_i}, and use the
#'     \eqn{(1-\alpha)} quantile as a critical multiplier on the
#'     pointwise SE. Valid for feature localisation ("where is
#'     \eqn{\hat f'(x)} significantly non-zero").
#' @param derivative_ci_nsim Integer. Number of Monte Carlo samples
#'   used when `derivative_ci = "simultaneous"`. Default `1000L` —
#'   a compromise between coverage accuracy (Simpson 2018 uses
#'   10000) and CPU budget across every pair in a medium-sized
#'   matrix. Ignored for any other `derivative_ci`.
#' @param n_grid Integer or `NULL`. Number of equally-spaced points
#'   used to evaluate each fitted smooth (and its derivatives).
#'   Default `NULL` resolves to `100` when `display = "fit"` and
#'   `200` otherwise, because finite-difference second derivatives
#'   visibly degrade below \eqn{\sim 150} points on moderate-`k`
#'   smooths. Supplying `n_grid` directly overrides both defaults.
#'   Larger grids shift the numerical shape-metric values
#'   (\eqn{M}, \eqn{C}, turning / inflection counts) slightly
#'   because they are computed on this same grid. Shapes and
#'   asymmetry are the primary reading; `M`, `C` and the counts are
#'   secondary diagnostics and the grid-induced drift is tolerable.
#' @param colour_by One of `"pearson"` (default), `"spearman"`,
#'   `"kendall"`, `"edf"`, `"deviance_gap"`, or `"none"`. Encodes the
#'   per-cell fill colour by the chosen scalar. Correlation choices use
#'   a diverging palette with limits `c(-1, 1)` and a shared `corr`
#'   colour-bar title; `"edf"` and `"deviance_gap"` use a sequential
#'   palette labelled by the metric.
#' @param fill_by Deprecated alias for `colour_by`. If supplied emits a
#'   single soft deprecation warning and is forwarded to `colour_by`.
#' @param palette Character. Colour palette for the cell fill scale.
#'   Defaults to `"RdBu"` when `colour_by` is a correlation and
#'   `"viridis"` otherwise. Sequential choices: `"viridis"`, `"magma"`,
#'   `"inferno"`, `"plasma"`, `"cividis"`, `"mako"`, `"rocket"`,
#'   `"turbo"` (not CB-safe), `"YlOrRd"`, `"YlGnBu"`, `"Blues"`,
#'   `"Greens"`. Diverging choices: `"RdYlBu"`, `"RdBu"`, `"PuOr"`,
#'   `"Spectral"` (not CB-safe). Passing a sequential palette while
#'   `colour_by` is a correlation silently upgrades to the default
#'   diverging palette.
#' @param annotations Character vector, a subset of
#'   `c("edf", "A", "shape", "code")`. Controls which corner
#'   annotations appear on each off-diagonal cell:
#'   * `"code"` — 2-letter ASCII shape code, **top-left** corner.
#'   * `"A"` and `"edf"` — asymmetry index and effective degrees of
#'     freedom, stacked **bottom-left**.
#'   * `"shape"` — shape glyph (Unicode or ASCII per `glyph_style`),
#'     **bottom-right** corner.
#'
#'   Default `c("edf", "A")`. `"code"` and `"shape"` occupy distinct
#'   corners so both can be requested together. See
#'   [janusplot_shape_hierarchy()] for the full code list.
#' @param shape_cutoffs Named list of classification thresholds used to
#'   map the continuous shape indices into discrete `shape_category`
#'   labels; see [janusplot_shape_cutoffs()].
#' @param show_shape_legend Logical. If `TRUE` (default), attach a
#'   standing shape-types legend plate below the matrix that
#'   illustrates every category in the taxonomy as a canonical
#'   thumbnail spline. Independent of `annotations`.
#' @param glyph_style One of `"ascii"` (default) or `"unicode"`.
#'   Controls how cell shape glyphs render when `"shape"` is included
#'   in `annotations`. Default is `"ascii"` for maximum portability
#'   across typesetting pipelines; switch to `"unicode"` only when
#'   the target font is known to cover the curve glyph set.
#' @param labels One of `"border"` (default), `"diagonal"`, or
#'   `"none"`. Controls where variable names are rendered:
#'   * `"border"` — names along the top (rotated per `label_srt`) and
#'     left margins of the matrix; diagonal cells are left blank.
#'     Mirrors `corrplot`'s `tl.pos = "lt"` convention.
#'   * `"diagonal"` — names centred on the diagonal cells (the
#'     pre-0.1 layout).
#'   * `"none"` — labels suppressed entirely; diagonal cells blank.
#' @param diagonal One of `"auto"` (default), `"blank"`, `"name"`,
#'   or `"density"`. Controls what is rendered in the diagonal
#'   cells of the matrix.
#'   * `"auto"` — preserves the historical behaviour: variable name
#'     when `labels = "diagonal"`, blank otherwise.
#'   * `"blank"` — empty bordered panel (uniform grid reading).
#'   * `"name"` — variable name centred in the cell, bold.
#'   * `"density"` — kernel density of the variable filled in
#'     translucent grey, with a rug of raw values along the bottom
#'     edge. Mirrors the `GGally::ggpairs` convention; surfaces
#'     tail weight, bimodality, and support clipping that the
#'     pairwise smooths alone cannot reveal. Variable names should
#'     come from the border (`labels = "border"`, the default) when
#'     this mode is on.
#' @param label_srt Numeric. Rotation (degrees) of top labels when
#'   `labels = "border"`. Default `45`; set to `0` for horizontal or
#'   `90` for vertical. Ignored when `labels != "border"`.
#' @param label_cex Positive numeric multiplier on the border-label
#'   font size. Default `1`. Ignored when `labels = "none"`.
#' @param signif_glyph Logical. If `TRUE` (default), annotate cells
#'   with `·` / `*` / `**` reflecting the smooth's F-test p-value.
#' @param show_asymmetry Deprecated. Use `annotations` instead
#'   (`"A" %in% annotations`). When supplied, a soft deprecation
#'   warning fires and the argument is merged into `annotations`.
#' @param na_action One of `"pairwise"` (default; per-cell complete
#'   observations) or `"complete"` (listwise; all cells use the same
#'   rows).
#' @param parallel Logical. If `TRUE`, use `future.apply::future_mapply()`
#'   to fit pairs in parallel. Requires the `future.apply` package and a
#'   user-configured `future::plan()`. Default `FALSE`.
#' @param with_data Logical. If `TRUE`, return a two-element list
#'   `list(plot, data)` where `data` is a flat per-cell summary
#'   (one row per off-diagonal cell) of everything the plot displays.
#'   The `data` element is always a plain `data.frame` (base R — no
#'   `data.table` dependency). Default `FALSE` — in which case only
#'   the ggplot is returned.
#' @param text_scale_diag Positive numeric multiplier applied to the
#'   diagonal variable-name labels. Default `1`. Diagonal labels
#'   additionally auto-shrink for long variable names
#'   (`nchar(var) > 10`) so they fit the cell regardless of this value.
#' @param text_scale_off_diag Positive numeric multiplier applied to
#'   all off-diagonal annotations (`n` / `EDF` readouts, significance
#'   glyphs, asymmetry-index labels). Default `1`. Use `< 1` when
#'   cells are small and the annotations crowd the fit line; use
#'   `> 1` for presentation plots.
#' @param show_glossary Logical. If `TRUE` (default), attach a
#'   multi-line caption below the matrix describing the on-plot
#'   abbreviations (`n`, `EDF`, `A`, fill encoding, significance
#'   glyphs). Only keys actually displayed are listed.
#' @param glossary_scale Positive numeric multiplier on the glossary
#'   caption font size. Default `1`.
#' @param ... Additional arguments passed to [mgcv::gam()].
#'
#' @returns If `with_data = FALSE` (default), a [ggplot2::ggplot] object
#'   (via [patchwork::wrap_plots()]) carrying a top-of-matrix title
#'   that names the displayed quantity (`"Direct fit"`,
#'   `"First derivative f'"`, or `"Second derivative f''"`). If
#'   `with_data = TRUE`, a list with two elements: `plot` (the
#'   ggplot) and `data` (a tidy table with columns `var_x`, `var_y`,
#'   `position`, `n_used`, `edf`, `pvalue`, `signif`, `dev_exp`,
#'   `asymmetry_index`, `cor_pearson`, `cor_spearman`,
#'   `cor_kendall`, `tie_ratio`, `monotonicity_index`,
#'   `convexity_index`, `n_turning_points`, `n_inflections`,
#'   `flat_range_ratio`, `shape_category`, `colour_value`,
#'   `display`, one row per off-diagonal cell). The `display`
#'   column tags which quantity the call rendered, so separate
#'   calls for fit / d1 / d2 yield comparable, stackable tables.
#'   Derivative *curves* themselves (grid of \eqn{x}, fitted
#'   \eqn{\hat f^{(k)}}, SE) live on `janusplot_data()` — see there.
#'
#' @family smooth-associations
#' @seealso [janusplot_data()] for the raw per-cell fits + metrics.
#'
#' @examples
#' # Minimal runnable example — 3 variables, 6 asymmetric pairwise GAM fits.
#' janusplot(mtcars[, c("mpg", "hp", "wt")])
#'
#' \donttest{
#' # Heteroscedastic DGP: Pearson r is ~ 0.9 but the inverse fit is
#' # clearly non-linear, yielding asymmetry index > 0.5.
#' set.seed(2026L)
#' n  <- 200L
#' x1 <- stats::runif(n, 0, 10)
#' x2 <- x1 + stats::rnorm(n, sd = 0.2 * x1)
#' janusplot(data.frame(x1 = x1, x2 = x2, x3 = stats::rnorm(n)))
#'
#' # A single matrix renders a single quantity. To compare the fit
#' # against its derivatives, render three calls and place them
#' # side-by-side; each call's title makes the quantity explicit.
#' set.seed(2026L)
#' xs <- stats::runif(300L, -3, 3)
#' df <- data.frame(
#'   x  = xs,
#'   y1 = sin(xs)  + stats::rnorm(300L, sd = 0.3),
#'   y2 = xs^2     + stats::rnorm(300L, sd = 0.6)
#' )
#' janusplot(df, display = "fit")
#' janusplot(df, display = "d1")
#' janusplot(df, display = "d2")
#'
#' # Simultaneous CI bands on a derivative panel, per Simpson (2018).
#' janusplot(df, display = "d1", derivative_ci = "simultaneous")
#' }
#' @export
janusplot <- function(
    data,
    vars = NULL,
    adjust = NULL,
    method = "REML",
    k = -1L,
    bs = "tp",
    order = c("original", "hclust", "alphabetical"),
    show_data = TRUE,
    show_ci = TRUE,
    display = c("fit", "d1", "d2"),
    derivative_ci = c("none", "pointwise", "simultaneous"),
    derivative_ci_nsim = 1000L,
    n_grid = NULL,
    colour_by = c("pearson", "spearman", "kendall",
                  "edf", "deviance_gap", "none"),
    fill_by = NULL,
    palette = NULL,
    annotations = c("edf", "A"),
    shape_cutoffs = janusplot_shape_cutoffs(),
    show_shape_legend = TRUE,
    glyph_style = c("ascii", "unicode"),
    labels = c("border", "diagonal", "none"),
    diagonal = c("auto", "blank", "name", "density"),
    label_srt = 45,
    label_cex = 1,
    signif_glyph = TRUE,
    show_asymmetry = NULL,
    na_action = c("pairwise", "complete"),
    parallel = FALSE,
    with_data = FALSE,
    text_scale_diag     = 1,
    text_scale_off_diag = 1,
    show_glossary       = TRUE,
    glossary_scale      = 1,
    ...) {
  order         <- rlang::arg_match(order)
  na_action     <- rlang::arg_match(na_action)
  glyph_style   <- rlang::arg_match(glyph_style)
  labels        <- rlang::arg_match(labels)
  diagonal      <- rlang::arg_match(diagonal)
  display       <- rlang::arg_match(display)
  derivative_ci <- rlang::arg_match(derivative_ci)
  # `diagonal = "auto"` resolves to the historical behaviour: name
  # in the diagonal cell when labels == "diagonal", blank otherwise.
  diagonal_eff <- if (diagonal == "auto") {
    if (labels == "diagonal") "name" else "blank"
  } else {
    diagonal
  }

  if (!is.numeric(derivative_ci_nsim) ||
      length(derivative_ci_nsim) != 1L ||
      !is.finite(derivative_ci_nsim) ||
      derivative_ci_nsim < 100) {
    cli::cli_abort(
      "{.arg derivative_ci_nsim} must be a single integer >= 100."
    )
  }
  derivative_ci_nsim <- as.integer(derivative_ci_nsim)

  if (!is.numeric(label_srt) || length(label_srt) != 1L ||
      is.na(label_srt)) {
    cli::cli_abort("{.arg label_srt} must be a single numeric value.")
  }
  .check_scalar_positive(label_cex, "label_cex")

  # Dual-alias: fill_by -> colour_by (soft deprecation).
  if (!is.null(fill_by)) {
    lifecycle::deprecate_warn(
      when = "0.0.0.9000",
      what = "janusplot(fill_by = )",
      with = "janusplot(colour_by = )",
      details = paste0("Forwarding \"", fill_by, "\" to colour_by.")
    )
    colour_by <- fill_by
  }
  colour_by <- rlang::arg_match(colour_by, .colour_choices())

  if (is.null(palette)) palette <- .default_palette(colour_by)
  palette <- rlang::arg_match(palette, .palette_choices())

  if (!is.character(annotations)) {
    cli::cli_abort(
      "{.arg annotations} must be a character vector (subset of edf/A/shape/code)."
    )
  }
  valid_annot <- c("edf", "A", "shape", "code")
  bad <- setdiff(annotations, valid_annot)
  if (length(bad)) {
    cli::cli_abort(c(
      "Unknown {.arg annotations} value{?s}: {.val {bad}}.",
      i = "Allowed: {.val {valid_annot}}."
    ))
  }

  # Resolve a single integer derivative order from scalar display.
  # "fit" → no derivatives computed; "d1" / "d2" → the matching order.
  derivative_orders <- switch(display,
    fit = integer(),
    d1  = 1L,
    d2  = 2L
  )

  # Validate and resolve n_grid. NULL means "100 if no derivatives,
  # 200 otherwise"; a user value overrides both defaults. Flag very
  # large grids — fit time scales as O(n_grid) and derivative SE as
  # O(n_grid) via the D %*% Vp %*% t(D) product.
  if (!is.null(n_grid)) {
    if (!is.numeric(n_grid) || length(n_grid) != 1L ||
        !is.finite(n_grid) || n_grid < 10) {
      cli::cli_abort(
        "{.arg n_grid} must be a single finite number >= 10 or NULL."
      )
    }
    n_grid <- as.integer(n_grid)
    if (n_grid > 500L) {
      cli::cli_inform(c(
        "!" = paste0(
          "{.arg n_grid} = {n_grid} is unusually large \u2014 ",
          "each pair pays an O(n_grid) predict cost."
        ),
        i = "Grids above ~300 rarely improve readability."
      ))
    }
  } else {
    n_grid <- if (length(derivative_orders)) 200L else 100L
  }

  # Dual-alias: show_asymmetry -> annotations (soft deprecation).
  if (!is.null(show_asymmetry)) {
    lifecycle::deprecate_warn(
      when = "0.0.0.9000",
      what = "janusplot(show_asymmetry = )",
      with = "janusplot(annotations = )"
    )
    if (isTRUE(show_asymmetry)) {
      annotations <- union(annotations, "A")
    } else if (isFALSE(show_asymmetry)) {
      annotations <- setdiff(annotations, "A")
    }
  }

  .check_scalar_positive(text_scale_diag,     "text_scale_diag")
  .check_scalar_positive(text_scale_off_diag, "text_scale_off_diag")
  .check_scalar_positive(glossary_scale,      "glossary_scale")
  if (!is.logical(show_glossary) || length(show_glossary) != 1L ||
      is.na(show_glossary)) {
    cli::cli_abort("{.arg show_glossary} must be TRUE or FALSE.")
  }
  if (!is.logical(show_shape_legend) || length(show_shape_legend) != 1L ||
      is.na(show_shape_legend)) {
    cli::cli_abort("{.arg show_shape_legend} must be TRUE or FALSE.")
  }

  .validate_inputs(data, vars, adjust, na_action)
  vars <- .resolve_vars(data, vars)

  if (order == "hclust") {
    vars <- .reorder_hclust(data, vars)
  } else if (order == "alphabetical") {
    vars <- sort(vars)
  }

  fits <- .fit_all_pairs(
    data = data, vars = vars, adjust = adjust,
    method = method, k = k, bs = bs,
    na_action = na_action, parallel = parallel,
    n_grid = n_grid, derivatives = derivative_orders,
    derivative_ci = derivative_ci,
    derivative_ci_nsim = derivative_ci_nsim, ...
  )

  # Colour-scale limits pooled across off-diagonal cells. Correlation
  # encodings use fixed symmetric [-1, 1]; non-linearity indices use
  # the observed data range so we get full use of the palette.
  colour_vals <- vapply(
    fits, function(f) .colour_value(f, colour_by), numeric(1L)
  )
  colour_limits <- if (colour_by == "none" || all(is.na(colour_vals))) {
    c(NA_real_, NA_real_)
  } else if (.is_correlation_colour(colour_by)) {
    c(-1, 1)
  } else {
    range(colour_vals, na.rm = TRUE)
  }

  asym_tbl <- .asymmetry_table(fits, vars)

  k_n <- length(vars)
  text_sizes <- .cell_text_sizes(
    k_n,
    text_scale_diag     = text_scale_diag,
    text_scale_off_diag = text_scale_off_diag
  )
  cells_by_ij <- vector("list", length(fits))
  names(cells_by_ij) <- names(fits)
  for (key in names(fits)) {
    f <- fits[[key]]
    ij <- as.integer(strsplit(key, "_", fixed = TRUE)[[1L]])
    is_upper <- ij[1L] < ij[2L]
    asym_key <- paste(sort(ij), collapse = "_")
    cells_by_ij[[key]] <- .build_cell(
      fit_obj = f,
      show_data = show_data, show_ci = show_ci,
      colour_by = colour_by, palette = palette,
      signif_glyph = signif_glyph,
      annotations = annotations,
      shape_cutoffs = shape_cutoffs,
      glyph_style = glyph_style,
      asym_val = asym_tbl[[asym_key]] %||% NA_real_,
      colour_limits = colour_limits,
      is_upper = is_upper,
      text_sizes = text_sizes,
      display = display,
      derivative_ci = derivative_ci
    )
  }

  diag_cells <- lapply(vars, function(v) {
    switch(diagonal_eff,
      blank   = .build_blank_diagonal_cell(),
      name    = .build_diagonal_cell(v, text_sizes = text_sizes),
      density = .build_density_rug_cell(v, data[[v]],
                                        text_sizes = text_sizes)
    )
  })

  # Shape legend is a standing reference: it now renders the full
  # taxonomy regardless of which categories are present in the
  # matrix or whether the user opted into cell-level shape glyphs.
  shape_legend_plot <- if (isTRUE(show_shape_legend)) {
    .build_shape_legend_plot()
  } else {
    NULL
  }

  composite <- .assemble_matrix(
    cells_by_ij, diag_cells, vars,
    colour_by = colour_by, colour_limits = colour_limits,
    palette = palette, shape_legend_plot = shape_legend_plot,
    labels = labels, label_srt = label_srt, label_cex = label_cex
  )
  plot_out <- .finalize_plot(
    composite,
    title   = .display_title(display, glyph_style),
    caption = if (isTRUE(show_glossary)) {
      .build_glossary_text(signif_glyph, annotations, colour_by, display,
                           derivative_ci)
    } else {
      NULL
    },
    glossary_scale = glossary_scale
  )

  if (!isTRUE(with_data)) {
    return(plot_out)
  }

  list(
    plot = plot_out,
    data = .build_summary_table(fits, vars, colour_by, asym_tbl,
                                shape_cutoffs, display = display)
  )
}

# ---------------------------------------------------------------
# Flat per-cell summary table — mirrors what the plot displays.
# ---------------------------------------------------------------

.build_summary_table <- function(fits, vars, colour_by, asym_tbl,
                                 shape_cutoffs, display = "fit") {
  rows <- lapply(names(fits), function(key) {
    f <- fits[[key]]
    ij <- as.integer(strsplit(key, "_", fixed = TRUE)[[1L]])
    position <- if (ij[1L] < ij[2L]) "upper" else "lower"
    asym_key <- paste(sort(ij), collapse = "_")
    shape_cat <- .classify_shape(
      f$shape$monotonicity_index, f$shape$convexity_index,
      f$shape$n_turning_points, f$shape$n_inflections,
      f$shape$flat_range_ratio, shape_cutoffs
    )
    data.frame(
      var_x              = f$x_name,
      var_y              = f$y_name,
      position           = position,
      n_used             = f$n_used,
      edf                = f$edf,
      pvalue             = f$pvalue,
      signif             = .pvalue_to_glyph(f$pvalue),
      dev_exp            = f$dev_exp,
      asymmetry_index    = asym_tbl[[asym_key]] %||% NA_real_,
      cor_pearson        = f$corr$cor_pearson,
      cor_spearman       = f$corr$cor_spearman,
      cor_kendall        = f$corr$cor_kendall,
      tie_ratio          = f$corr$tie_ratio,
      monotonicity_index = f$shape$monotonicity_index,
      convexity_index    = f$shape$convexity_index,
      n_turning_points   = f$shape$n_turning_points,
      n_inflections      = f$shape$n_inflections,
      flat_range_ratio   = f$shape$flat_range_ratio,
      shape_category     = shape_cat,
      shape_code         = .shape_lookup(shape_cat, "code"),
      shape_archetype    = .shape_lookup(shape_cat, "archetype"),
      shape_monotonic    = .shape_lookup(shape_cat, "monotonic"),
      shape_linear       = .shape_lookup(shape_cat, "linear"),
      colour_value       = .colour_value(f, colour_by),
      display            = display,
      stringsAsFactors   = FALSE
    )
  })
  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  out
}

# ---------------------------------------------------------------
# Display-mode title: the label at the top of the assembled matrix
# naming what is rendered in every cell. Unicode primes by default,
# ASCII fallback when glyph_style = "ascii".
# ---------------------------------------------------------------

.display_title <- function(display, glyph_style = c("ascii", "unicode")) {
  glyph_style <- rlang::arg_match(glyph_style)
  if (glyph_style == "unicode") {
    switch(display,
      fit = "Direct fit",
      d1  = "First derivative  f\u2032(x)",
      d2  = "Second derivative  f\u2033(x)",
      display
    )
  } else {
    switch(display,
      fit = "Direct fit",
      d1  = "First derivative  df/dx",
      d2  = "Second derivative  d^2 f / dx^2",
      display
    )
  }
}

# ---------------------------------------------------------------
# Glossary caption body — only lists keys actually shown on the
# plot. Extended to mention the derivative CI mode when a
# derivative is displayed.
# ---------------------------------------------------------------

.build_glossary_text <- function(signif_glyph, annotations, colour_by,
                                 display = "fit", derivative_ci = "none") {
  parts <- character()
  if (display == "fit" && "A" %in% annotations) {
    parts <- c(parts,
      "A = asymmetry index |EDF_yx - EDF_xy| / (EDF_yx + EDF_xy), in [0, 1]")
  }
  if (display == "fit" && "edf" %in% annotations) {
    parts <- c(parts,
      "EDF = effective degrees of freedom of the smooth")
  }
  if (display == "fit" && "shape" %in% annotations) {
    parts <- c(parts,
      "Shape glyph = objective shape category (see shape-types legend below)")
  }
  colour_msg <- switch(colour_by,
    spearman     = "Cell colour encodes Spearman rank correlation (corr)",
    pearson      = "Cell colour encodes Pearson correlation (corr)",
    kendall      = "Cell colour encodes Kendall correlation (corr)",
    edf          = "Cell colour encodes EDF (see right-margin legend)",
    deviance_gap = "Cell colour encodes deviance explained",
    none         = NULL
  )
  if (!is.null(colour_msg)) parts <- c(parts, colour_msg)
  if (display == "fit" && signif_glyph) {
    parts <- c(parts,
      "Signif. of smooth F-test: *** p<.001, ** p<.01, * p<.05, \u00b7 p<.1")
  }
  if (display %in% c("d1", "d2")) {
    parts <- c(parts, sprintf(
      "Derivative panel: curve from D %%*%% coef(gam), dashed line at zero; CI mode = %s",
      derivative_ci
    ))
  }
  paste(parts, collapse = "\n")
}

# ---------------------------------------------------------------
# Finalise composite plot — add title + optional caption in a
# single plot_annotation call so patchwork lays them out cleanly.
# ---------------------------------------------------------------

.finalize_plot <- function(composite, title, caption = NULL,
                           glossary_scale = 1) {
  composite +
    patchwork::plot_annotation(
      title   = title,
      caption = caption,
      theme = ggplot2::theme(
        plot.title   = ggplot2::element_text(
          size       = 13,
          face       = "bold",
          hjust      = 0,
          colour     = "grey10",
          margin     = ggplot2::margin(t = 2, b = 6)
        ),
        plot.caption = ggplot2::element_text(
          size       = 9 * glossary_scale,
          hjust      = 0,
          colour     = "grey25",
          lineheight = 1.1,
          margin     = ggplot2::margin(t = 8, b = 2)
        )
      )
    )
}

# Helper used only by janusplot(): asymmetry index lookup by unordered pair
.asymmetry_table <- function(fits, vars) {
  k <- length(vars)
  tbl <- list()
  for (i in seq_len(k - 1L)) {
    for (j in (i + 1L):k) {
      f_ij <- fits[[sprintf("%d_%d", i, j)]]
      f_ji <- fits[[sprintf("%d_%d", j, i)]]
      if (is.null(f_ij) || is.null(f_ji)) next
      tbl[[paste(i, j, sep = "_")]] <- .compute_asymmetry_index(
        f_ij$edf, f_ji$edf
      )
    }
  }
  tbl
}


#' Raw GAM fits and per-cell metrics for a smoothed-association matrix
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Companion to [janusplot()] returning the raw list of GAM fits plus
#' per-cell metrics (EDF, F-test p-value, deviance explained, asymmetry
#' index, pairwise correlations, shape descriptors) without constructing
#' the ggplot. Useful for custom rendering or downstream analysis.
#'
#' @inheritParams janusplot
#' @param keep_fits Logical. If `TRUE`, retain full [mgcv::gam()] model
#'   objects in the return (large memory footprint for `k` above ~15).
#'   Default `FALSE` — retains summary metrics and prediction grids only.
#' @param derivatives Integer vector of derivative orders to compute
#'   on every pair (subset of `1:2`). Default `integer()` — no
#'   derivatives. Unlike `janusplot()`, the data companion can
#'   return multiple orders from a single call for programmatic
#'   analysis; pass `c(1L, 2L)` to surface both.
#' @param shape_cutoffs Named list of classification thresholds used to
#'   map the continuous shape indices (`monotonicity_index`,
#'   `convexity_index`) into discrete
#'   `shape_category` labels. Defaults from [janusplot_shape_cutoffs()].
#'
#' @returns A list with components:
#' \describe{
#'   \item{`vars`}{Character vector of variables used, in plotted order.}
#'   \item{`pairs`}{List of per-pair results. Each element has `i`, `j`,
#'     `var_i`, `var_j`, `fit_yx`, `fit_xy` (NULL if `keep_fits = FALSE`),
#'     `pred_yx`, `pred_xy` (data frames with `x`, `fit`, `se`, `lo`,
#'     `hi`), `edf_yx`, `edf_xy`, `pvalue_yx`, `pvalue_xy`, `dev_exp_yx`,
#'     `dev_exp_xy`, `n_used`, `asymmetry_index`, plus Pearson /
#'     Spearman / Kendall correlations (`cor_pearson`, `cor_spearman`,
#'     `cor_kendall`), the maximum tie ratio across `x` and `y`
#'     (`tie_ratio`), and per-direction shape descriptors
#'     (`monotonicity_index_yx`, `convexity_index_yx`,
#'     `monotonicity_index_xy`, `convexity_index_xy`,
#'     `n_turning_yx`, `n_inflect_yx`, `n_turning_xy`,
#'     `n_inflect_xy`, `shape_yx`, `shape_xy`).
#'     When `derivatives` is non-empty, each pair additionally
#'     carries `deriv_yx` and `deriv_xy`, each a named list keyed by
#'     order (`"1"`, `"2"`) whose entries are data frames with
#'     columns `x`, `fit`, `se`, `lo`, `hi`, `ci_type` matching the
#'     schema of `pred_yx` / `pred_xy`. The `ci_type` column records
#'     whether the `lo` / `hi` columns are `"pointwise"` (default),
#'     `"simultaneous"` (Ruppert--Wand--Carroll / Simpson 2018
#'     critical-multiplier bands), or `"none"`. When
#'     `derivative_ci = "simultaneous"`, each derivative frame also
#'     carries a `"crit_multiplier"` attribute giving the MC-derived
#'     critical multiplier used.
#'     See [janusplot_shape_metrics()] for the definition of the
#'     monotonicity and convexity indices.}
#'   \item{`call`}{Match call.}
#' }
#'
#' @family smooth-associations
#' @seealso [janusplot()] for the ggplot front-end,
#'   [janusplot_shape_metrics()] for the shape-metric primitives.
#'
#' @examples
#' # Per-pair fits + metrics on a small mtcars slice
#' out <- janusplot_data(mtcars[, c("mpg", "hp", "wt")])
#' out$pairs[[1L]]$asymmetry_index
#' out$pairs[[1L]]$cor_spearman
#' out$pairs[[1L]]$shape_yx
#' @export
janusplot_data <- function(
    data,
    vars = NULL,
    adjust = NULL,
    method = "REML",
    k = -1L,
    bs = "tp",
    na_action = c("pairwise", "complete"),
    parallel = FALSE,
    keep_fits = FALSE,
    derivatives = integer(),
    derivative_ci = c("pointwise", "none", "simultaneous"),
    derivative_ci_nsim = 1000L,
    n_grid = NULL,
    shape_cutoffs = janusplot_shape_cutoffs(),
    ...) {
  na_action     <- rlang::arg_match(na_action)
  derivative_ci <- rlang::arg_match(derivative_ci)
  .validate_inputs(data, vars, adjust, na_action)
  vars <- .resolve_vars(data, vars)

  if (length(derivatives)) {
    if (!is.numeric(derivatives) || anyNA(derivatives)) {
      cli::cli_abort(
        "{.arg derivatives} must be an integer vector of orders in 1:2."
      )
    }
    derivatives <- as.integer(derivatives)
    if (any(derivatives < 1L | derivatives > 2L)) {
      cli::cli_abort(
        "{.arg derivatives} entries must be in 1:2 \u2014 higher orders are not supported."
      )
    }
    if (anyDuplicated(derivatives)) {
      cli::cli_abort("{.arg derivatives} must not contain duplicates.")
    }
  }

  if (!is.numeric(derivative_ci_nsim) ||
      length(derivative_ci_nsim) != 1L ||
      !is.finite(derivative_ci_nsim) ||
      derivative_ci_nsim < 100) {
    cli::cli_abort(
      "{.arg derivative_ci_nsim} must be a single integer >= 100."
    )
  }
  derivative_ci_nsim <- as.integer(derivative_ci_nsim)

  if (!is.null(n_grid)) {
    if (!is.numeric(n_grid) || length(n_grid) != 1L ||
        !is.finite(n_grid) || n_grid < 10) {
      cli::cli_abort(
        "{.arg n_grid} must be a single finite number >= 10 or NULL."
      )
    }
    n_grid <- as.integer(n_grid)
  } else {
    n_grid <- if (length(derivatives)) 200L else 100L
  }

  fits <- .fit_all_pairs(
    data = data, vars = vars, adjust = adjust,
    method = method, k = k, bs = bs,
    na_action = na_action, parallel = parallel,
    n_grid = n_grid, derivatives = derivatives,
    derivative_ci = derivative_ci,
    derivative_ci_nsim = derivative_ci_nsim, ...
  )

  k_n <- length(vars)
  pairs_out <- list()
  for (i in seq_len(k_n - 1L)) {
    for (j in (i + 1L):k_n) {
      f_yx <- fits[[sprintf("%d_%d", i, j)]]   # y = vars[j] ~ s(vars[i])
      f_xy <- fits[[sprintf("%d_%d", j, i)]]   # y = vars[i] ~ s(vars[j])
      shape_yx <- .classify_shape(
        f_yx$shape$monotonicity_index, f_yx$shape$convexity_index,
        f_yx$shape$n_turning_points, f_yx$shape$n_inflections,
        f_yx$shape$flat_range_ratio, shape_cutoffs
      )
      shape_xy <- .classify_shape(
        f_xy$shape$monotonicity_index, f_xy$shape$convexity_index,
        f_xy$shape$n_turning_points, f_xy$shape$n_inflections,
        f_xy$shape$flat_range_ratio, shape_cutoffs
      )
      corr <- f_yx$corr %||% f_xy$corr
      pairs_out[[length(pairs_out) + 1L]] <- list(
        i                      = i,
        j                      = j,
        var_i                  = vars[i],
        var_j                  = vars[j],
        fit_yx                 = if (keep_fits) f_yx$fit else NULL,
        fit_xy                 = if (keep_fits) f_xy$fit else NULL,
        pred_yx                = f_yx$pred,
        pred_xy                = f_xy$pred,
        edf_yx                 = f_yx$edf,
        edf_xy                 = f_xy$edf,
        pvalue_yx              = f_yx$pvalue,
        pvalue_xy              = f_xy$pvalue,
        dev_exp_yx             = f_yx$dev_exp,
        dev_exp_xy             = f_xy$dev_exp,
        n_used                 = min(f_yx$n_used, f_xy$n_used),
        asymmetry_index        = .compute_asymmetry_index(f_yx$edf, f_xy$edf),
        cor_pearson            = corr$cor_pearson,
        cor_spearman           = corr$cor_spearman,
        cor_kendall            = corr$cor_kendall,
        tie_ratio              = corr$tie_ratio,
        monotonicity_index_yx  = f_yx$shape$monotonicity_index,
        convexity_index_yx     = f_yx$shape$convexity_index,
        monotonicity_index_xy  = f_xy$shape$monotonicity_index,
        convexity_index_xy     = f_xy$shape$convexity_index,
        n_turning_yx           = f_yx$shape$n_turning_points,
        n_inflect_yx           = f_yx$shape$n_inflections,
        n_turning_xy           = f_xy$shape$n_turning_points,
        n_inflect_xy           = f_xy$shape$n_inflections,
        shape_yx           = shape_yx,
        shape_xy           = shape_xy,
        shape_code_yx      = .shape_lookup(shape_yx, "code"),
        shape_code_xy      = .shape_lookup(shape_xy, "code"),
        shape_archetype_yx = .shape_lookup(shape_yx, "archetype"),
        shape_archetype_xy = .shape_lookup(shape_xy, "archetype"),
        shape_monotonic_yx = .shape_lookup(shape_yx, "monotonic"),
        shape_monotonic_xy = .shape_lookup(shape_xy, "monotonic"),
        shape_linear_yx    = .shape_lookup(shape_yx, "linear"),
        shape_linear_xy    = .shape_lookup(shape_xy, "linear"),
        deriv_yx           = f_yx$deriv,
        deriv_xy           = f_xy$deriv
      )
    }
  }
  list(vars = vars, pairs = pairs_out, call = match.call())
}

test_that("non-data-frame input is rejected", {
  expect_error(janusplot(1:10), class = "rlang_error")
})

test_that("too few rows is rejected", {
  expect_error(janusplot(data.frame(a = 1:3, b = 4:6)), class = "rlang_error")
})

test_that("vars must exist in data", {
  d <- make_linear_data(n = 50L)
  expect_error(janusplot(d, vars = c("x1", "missing")),
               class = "rlang_error")
})

test_that("vars must be numeric", {
  d <- data.frame(a = 1:30, g = factor(rep(c("a", "b"), 15L)))
  expect_error(janusplot(d, vars = c("a", "g")), class = "rlang_error")
})

test_that("adjust must be a one-sided formula", {
  d <- make_linear_data(n = 50L)
  expect_error(janusplot(d, adjust = "not a formula"),
               class = "rlang_error")
  expect_error(janusplot(d, adjust = y ~ x), class = "rlang_error")
})

test_that(".resolve_vars picks numeric columns by default", {
  d <- data.frame(a = 1:30, b = 1:30 + 0.5,
                  g = factor(rep(c("a", "b"), 15L)))
  out <- janusplot:::.resolve_vars(d, NULL)
  expect_equal(out, c("a", "b"))
})

# --- Shared scalar-argument validation --------------------------------------
# Characterisation tests pinning the message + that each check fires (before
# any model is fit). These guard the .validate_shared_scalars() delegation
# shared by janusplot() and janusplot_data().

test_that("discrete must be a single non-NA logical", {
  d <- make_linear_data(n = 50L)
  expect_error(janusplot_data(d, discrete = "yes"), "discrete")
  expect_error(janusplot_data(d, discrete = c(TRUE, FALSE)), "discrete")
  expect_error(janusplot_data(d, discrete = NA), "discrete")
})

test_that("nthreads must be a single positive integer", {
  d <- make_linear_data(n = 50L)
  expect_error(janusplot_data(d, nthreads = 0), "nthreads")
  expect_error(janusplot_data(d, nthreads = -1L), "nthreads")
  expect_error(janusplot_data(d, nthreads = c(1L, 2L)), "nthreads")
})

test_that("auto_refit_k must be a single non-NA logical", {
  d <- make_linear_data(n = 50L)
  expect_error(janusplot_data(d, auto_refit_k = "x"), "auto_refit_k")
  expect_error(janusplot_data(d, auto_refit_k = NA), "auto_refit_k")
})

test_that("k_max_iter must be a single non-negative integer", {
  d <- make_linear_data(n = 50L)
  expect_error(janusplot_data(d, k_max_iter = -1L), "k_max_iter")
  expect_error(janusplot_data(d, k_max_iter = c(1L, 2L)), "k_max_iter")
})

test_that("derivative_ci_nsim must be a single integer >= 100", {
  d <- make_linear_data(n = 50L)
  expect_error(janusplot_data(d, derivative_ci_nsim = 50), "derivative_ci_nsim")
  expect_error(
    janusplot_data(d, derivative_ci_nsim = c(100L, 200L)),
    "derivative_ci_nsim"
  )
})

test_that("shared scalar checks also fire from janusplot()", {
  d <- make_linear_data(n = 50L)
  expect_error(janusplot(d, discrete = "yes"), "discrete")
  expect_error(janusplot(d, nthreads = 0), "nthreads")
})

# CI-safe numeric-validation canary.
# Full 500-rep validation study lives under the parent
# bidirplot_dev/simulation/ workspace (out of package scope).
# This in-CI canary asserts: on a pure-linear DGP, both the
# asymmetry index and the recovered EDFs stay close to their
# theoretical floor.

test_that("asymmetry index near zero on pure-linear DGP", {
  set.seed(2026L)
  n  <- 200L
  x1 <- stats::rnorm(n)
  x2 <- 0.8 * x1 + stats::rnorm(n, sd = 0.5)
  x3 <- 0.5 * x1 + stats::rnorm(n, sd = 0.5)
  d  <- data.frame(x1 = x1, x2 = x2, x3 = x3)

  out <- janusplot_data(d)
  a   <- vapply(out$pairs, function(p) p$asymmetry_index, numeric(1L))

  # Cap deliberately generous — this is a canary, not a
  # methodological claim. REML smoothing-parameter bounce pushes
  # individual A values up to ~0.4 on finite samples even when the
  # DGP is exactly linear. The test asserts neither pair exceeds
  # 0.6 AND the mean sits below 0.4, which together catch a
  # genuine regression without false-alarming on REML noise.
  expect_true(
    all(a < 0.6) && mean(a) < 0.4,
    info = paste("asymmetry indices:", paste(round(a, 3), collapse = ", "))
  )
})

test_that("EDF recovers near 1 on linear DGP within REML bounce", {
  set.seed(2026L)
  n  <- 300L
  x1 <- stats::rnorm(n)
  x2 <- 1.5 * x1 + stats::rnorm(n, sd = 0.4)
  d  <- data.frame(x1 = x1, x2 = x2)

  out  <- janusplot_data(d)
  edfs <- c(out$pairs[[1L]]$edf_yx, out$pairs[[1L]]$edf_xy)

  expect_true(
    all(edfs < 1.5),
    info = paste("EDFs:", paste(round(edfs, 3), collapse = ", "))
  )
})

skip_on_cran()  # heavy GAM-fit suite: full on CI, skipped on CRAN for time budget

# Tests for Feature 1 — per-cell k-checking + opt-in auto-refit.
# Covers: diagnostic always-on; unreliable detection; flag plumbing;
# refit loop respects k_max_iter + per-cell unique-x cap; public API
# validation; annotations vocabulary accepts "k_warn".

skip_if_no_mgcv <- function() {
  testthat::skip_if_not_installed("mgcv")
}

# ---------------------------------------------------------------
# Diagnostic always-on — no opt-in needed
# ---------------------------------------------------------------

test_that("k-check diagnostics populate every cell on a normal dataset", {
  skip_if_no_mgcv()
  dat <- make_linear_data(n = 200L, seed = 1L)
  res <- janusplot_data(dat, vars = c("x1", "x2", "x3"))
  expect_true(length(res$pairs) > 0L)
  for (p in res$pairs) {
    expect_named(p$k_check_yx,
                 c("k_prime", "k_index", "k_p", "k_flag",
                   "k_check_status", "k_initial", "k_final",
                   "k_iterations", "k_at_cap"))
    expect_true(p$k_check_yx$k_check_status %in% c("ok", "flagged"))
    expect_equal(p$k_check_yx$k_iterations, 0L)
    expect_false(p$k_check_yx$k_at_cap)
  }
})

test_that("with_data table carries the 9 k-check columns", {
  skip_if_no_mgcv()
  dat <- make_linear_data(n = 200L, seed = 1L)
  out <- janusplot(dat, vars = c("x1", "x2", "x3"), with_data = TRUE)
  expected <- c("k_prime", "k_index", "k_p", "k_flag",
                "k_check_status", "k_initial", "k_final",
                "k_iterations", "k_at_cap")
  expect_true(all(expected %in% names(out$data)))
  expect_type(out$data$k_flag, "logical")
  expect_type(out$data$k_iterations, "integer")
})

# ---------------------------------------------------------------
# Unreliable status on cells with very few unique x values
# ---------------------------------------------------------------

test_that("k-check marks discrete-x cells with < 10 unique values as unreliable", {
  skip_if_no_mgcv()
  # 8 unique x values: mgcv fits (k = 8 allowed) but n_unique < 10
  # means k.check's simulation p-value is meaningless. Use k = 6 so
  # the fit definitely succeeds.
  withr::with_seed(7L, {
    n <- 80L
    df <- data.frame(
      g  = sample(seq_len(8L), n, replace = TRUE),
      y1 = stats::rnorm(n),
      y2 = stats::rnorm(n)
    )
  })
  res <- janusplot_data(df, vars = c("g", "y1", "y2"), k = 6L)
  # Cells where g is the predictor (i.e. s(g)) should be unreliable.
  statuses <- character()
  for (p in res$pairs) {
    if (p$var_i == "g") {
      statuses <- c(statuses, p$k_check_yx$k_check_status %||% NA_character_)
    }
    if (p$var_j == "g") {
      statuses <- c(statuses, p$k_check_xy$k_check_status %||% NA_character_)
    }
  }
  expect_true(any(statuses == "unreliable", na.rm = TRUE))
})

# ---------------------------------------------------------------
# Auto-refit loop
# ---------------------------------------------------------------

test_that("auto_refit_k = FALSE does not refit even on flagged cells", {
  skip_if_no_mgcv()
  # A noisy oscillation that will flag k underfit at the default k = 10.
  withr::with_seed(11L, {
    n <- 400L
    x <- stats::runif(n, 0, 10)
    df <- data.frame(
      x  = x,
      y  = sin(3 * x) + stats::rnorm(n, sd = 0.2),
      z  = stats::rnorm(n)
    )
  })
  res <- janusplot_data(df, vars = c("x", "y", "z"))
  for (p in res$pairs) {
    expect_equal(p$k_check_yx$k_iterations, 0L)
    expect_equal(p$k_check_xy$k_iterations, 0L)
    expect_false(p$k_check_yx$k_at_cap)
    expect_false(p$k_check_xy$k_at_cap)
  }
})

test_that("auto_refit_k = TRUE doubles k on flagged cells, respects k_max_iter", {
  skip_if_no_mgcv()
  withr::with_seed(13L, {
    n <- 400L
    x <- stats::runif(n, 0, 10)
    df <- data.frame(
      x = x,
      y = sin(3 * x) + stats::rnorm(n, sd = 0.2)
    )
  })
  res_no  <- janusplot_data(df, vars = c("x", "y"),
                            auto_refit_k = FALSE)
  res_yes <- janusplot_data(df, vars = c("x", "y"),
                            auto_refit_k = TRUE, k_max_iter = 2L)
  was_flagged <- FALSE
  for (p in res_no$pairs) {
    if (isTRUE(p$k_check_yx$k_flag) || isTRUE(p$k_check_xy$k_flag)) {
      was_flagged <- TRUE
    }
  }
  if (was_flagged) {
    iters <- c()
    for (p in res_yes$pairs) {
      iters <- c(iters, p$k_check_yx$k_iterations, p$k_check_xy$k_iterations)
    }
    expect_true(any(iters > 0L))
    expect_true(all(iters <= 2L))
  } else {
    succeed("no cells flagged on this DGP — refit loop has nothing to do")
  }
})

test_that("k_max_iter = 0 disables refit even with auto_refit_k = TRUE", {
  skip_if_no_mgcv()
  withr::with_seed(17L, {
    n <- 400L
    x <- stats::runif(n, 0, 10)
    df <- data.frame(
      x = x,
      y = sin(3 * x) + stats::rnorm(n, sd = 0.2)
    )
  })
  res <- janusplot_data(df, vars = c("x", "y"),
                        auto_refit_k = TRUE, k_max_iter = 0L)
  for (p in res$pairs) {
    expect_equal(p$k_check_yx$k_iterations, 0L)
    expect_equal(p$k_check_xy$k_iterations, 0L)
  }
})

# ---------------------------------------------------------------
# Public API validation
# ---------------------------------------------------------------

test_that("k_check_thresholds defaults match Wood (2017) and gam.check()", {
  expect_identical(
    janusplot:::.default_k_thresholds(),
    list(edf_ratio = 0.9, k_index = 1.0, p = 0.05)
  )
})

test_that("invalid k_check_thresholds errors loudly", {
  dat <- make_linear_data(n = 60L, seed = 1L)
  expect_error(
    janusplot_data(dat, vars = c("x1", "x2"),
                   k_check_thresholds = list(edf_ratio = 0.9)),
    "missing entries"
  )
  expect_error(
    janusplot_data(dat, vars = c("x1", "x2"),
                   k_check_thresholds = list(edf_ratio = -1, k_index = 1, p = 0.05)),
    "positive finite"
  )
  expect_error(
    janusplot_data(dat, vars = c("x1", "x2"),
                   k_check_thresholds = "edf_ratio = 0.9"),
    "named list"
  )
})

test_that("invalid auto_refit_k / k_max_iter error loudly", {
  dat <- make_linear_data(n = 60L, seed = 1L)
  expect_error(
    janusplot_data(dat, vars = c("x1", "x2"), auto_refit_k = NA),
    "TRUE or FALSE"
  )
  expect_error(
    janusplot_data(dat, vars = c("x1", "x2"), k_max_iter = -1),
    "non-negative integer"
  )
})

# ---------------------------------------------------------------
# Annotations vocabulary accepts "k_warn"
# ---------------------------------------------------------------

test_that("annotations = c('k_warn') is accepted and produces a ggplot", {
  skip_if_no_mgcv()
  dat <- make_linear_data(n = 60L, seed = 1L)
  p <- janusplot(dat, vars = c("x1", "x2"),
                 annotations = c("edf", "A", "k_warn"))
  expect_s3_class(p, "ggplot")
})

test_that("invalid annotation entry surfaces full vocabulary in error", {
  dat <- make_linear_data(n = 60L, seed = 1L)
  expect_error(
    janusplot(dat, vars = c("x1", "x2"), annotations = c("edf", "nonsense")),
    "k_warn"
  )
})

#' janusplot: Asymmetric Smoothed-Association Matrices via GAM Fits
#'
#' @description
#' `janusplot` renders pairwise, asymmetric smoothed-association matrices
#' of continuous variables. Each cell shows the fitted spline from an
#' [mgcv::gam()] model, with upper and lower triangles encoding the two
#' directional regressions `y ~ s(x)` and `x ~ s(y)` respectively.
#'
#' Unlike a Pearson correlation matrix (one scalar per pair, symmetric),
#' a smoothed-association matrix gives two curves per pair and is
#' intentionally asymmetric. Heteroscedasticity, leverage, and directional
#' non-linearity become visually evident.
#'
#' # Main functions
#'
#' * [janusplot()] — returns a ggplot of the matrix.
#' * [janusplot_data()] — returns the raw GAM fits + per-cell metrics
#'   (for custom plotting or downstream analysis).
#'
#' # Asymmetry index
#'
#' For each pair, the asymmetry index `A_ij = |EDF_yx - EDF_xy| /
#' (EDF_yx + EDF_xy)` is bounded in \[0, 1\]. Values near 0 indicate
#' symmetric complexity; values near 1 indicate the two directional
#' fits differ sharply in effective degrees of freedom.
#'
#' Under the additive noise model (Hoyer et al. 2009; Peters et al.
#' 2014), the two directional regressions are generally asymmetric
#' when the data-generating process is non-linear, and this
#' asymmetry identifies the causal direction under mild conditions.
#' The asymmetry index is offered here as a *visual pre-discovery
#' diagnostic* rather than a causal inference procedure; see the
#' package vignette and accompanying paper for full scope and
#' limitations (in particular the failure modes under
#' heteroscedasticity, confounding, and Gaussian-linear DGPs).
#'
#' @keywords internal
#' @aliases janusplot-package
#' @importFrom stats as.formula complete.cases predict quantile sd cor as.dist hclust
#' @importFrom rlang arg_match check_installed is_installed .data
"_PACKAGE"

#' Precomputed shape-recognition sensitivity results (demo)
#'
#' @description
#' Raw output from a small-footprint invocation of
#' [janusplot_shape_sensitivity()]. Shipped so users can explore the
#' sensitivity API and regenerate every figure in the
#' `shape-recognition-sensitivity` vignette without having to re-run
#' the sweep themselves. Regenerated via
#' `data-raw/shape_sensitivity_demo.R`.
#'
#' Design:
#' \itemize{
#'   \item **Shapes** (6, one per non-degenerate archetype):
#'     `linear_up`, `concave_up`, `u_shape`, `inverted_u`, `wave`,
#'     `bimodal`.
#'   \item **Sample sizes** (3): `c(100, 200, 500)`.
#'   \item **Noise levels** (4): `c(0.05, 0.10, 0.20, 0.40)` fraction
#'     of y-range.
#'   \item **Replicates**: 30.
#'   \item **Total fits**: 2160.
#'   \item **Seed**: 2026.
#' }
#'
#' @format A data frame with 2160 rows and 14 columns — see the
#'   "Value" section of [janusplot_shape_sensitivity()] for the
#'   column schema.
#'
#' @seealso [janusplot_shape_sensitivity()],
#'   [janusplot_shape_sensitivity_plot()],
#'   [janusplot_shape_sensitivity_summary()].
#'
#' @examples
#' data("shape_sensitivity_demo", package = "janusplot")
#' head(shape_sensitivity_demo)
#' janusplot_shape_sensitivity_plot(shape_sensitivity_demo,
#'                                  "recovery_curves")
"shape_sensitivity_demo"

# tests/testthat/setup.R — env sanitisation + seed discipline.

# Deterministic RNG across tests. Seed pinned to 2026L to match paper
# replication bundle.
set.seed(2026L, kind = "default", normal.kind = "default",
         sample.kind = "default")

# Locale sanitisation — number formatting and string collation must be
# deterministic across machines to keep vdiffr + snapshot tests stable.
tryCatch({
  Sys.setlocale("LC_COLLATE", "C")
  Sys.setlocale("LC_NUMERIC", "C")
  Sys.setlocale("LC_TIME",    "C")
}, error = function(e) invisible(NULL))

# Env-var sanitisation — clear anything that could influence knitr,
# rmarkdown, future.apply, or tinytex during test runs.
Sys.unsetenv(c(
  "R_CHECK_LENGTH_1_LOGIC2",
  "R_CHECK_LENGTH_1_CONDITION",
  "KNITR_IN_PROGRESS",
  "RMARKDOWN_PREVIEW_DIR"
))

# Disable parallelism in tests unless a specific test opts in.
# (future.apply respects options(future.plan = "sequential").)
options(future.plan = "sequential")

# vdiffr snapshot directory lives under tests/testthat/_snaps/ (default).

# Internal helpers — NOT EXPORTED.
# Fit-layer logic for janusplot(): input validation, formula assembly,
# GAM fitting, asymmetry index, and hclust reorder.

# Local null-coalesce (R >= 4.4 has this in base; keep local for 4.3 support).
`%||%` <- function(x, y) if (is.null(x)) y else x

# Argument validation — single positive scalar.
.check_scalar_positive <- function(x, arg_name) {
  if (!is.numeric(x) || length(x) != 1L || !is.finite(x) || x <= 0) {
    cli::cli_abort("{.arg {arg_name}} must be a single positive number.")
  }
  invisible(NULL)
}

# Diagnose parallel = TRUE requests that would silently fall back to
# sequential execution. Throttled per session so simulation loops don't
# spam the console; tests can pass .throttle = FALSE to force emission.
.check_parallel_plan <- function(parallel, .throttle = TRUE) {
  if (!isTRUE(parallel)) return(invisible(NULL))
  freq <- if (isTRUE(.throttle)) "regularly" else "always"
  if (!rlang::is_installed("future.apply")) {
    cli::cli_inform(
      c(
        "!" = paste(
          "{.arg parallel} is {.val TRUE} but {.pkg future.apply} is",
          "not installed \u2014 falling back to sequential dispatch."
        ),
        i = "Install with {.code pak::pak(\"future.apply\")}."
      ),
      .frequency    = freq,
      .frequency_id = "janusplot_no_future_apply"
    )
    return(invisible(NULL))
  }
  if (rlang::is_installed("future") &&
      future::nbrOfWorkers() < 2L) {
    cli::cli_inform(
      c(
        "!" = paste(
          "{.arg parallel} is {.val TRUE} but the current",
          "{.code future::plan()} is sequential (1 worker)."
        ),
        i = paste(
          "To parallelise, run e.g.",
          "{.code future::plan(future::multisession, workers = 4)}",
          "before {.fn janusplot}."
        )
      ),
      .frequency    = freq,
      .frequency_id = "janusplot_sequential_plan"
    )
  }
  invisible(NULL)
}

# ---------------------------------------------------------------
# Validation
# ---------------------------------------------------------------

.validate_inputs <- function(data, vars, adjust, na_action) {
  if (!is.data.frame(data)) {
    cli::cli_abort("{.arg data} must be a data frame.")
  }
  if (nrow(data) < 5L) {
    cli::cli_abort(c(
      "{.arg data} has {nrow(data)} row{?s}.",
      i = "A meaningful GAM needs at least ~5 rows per pair."
    ))
  }
  if (!is.null(vars)) {
    if (!is.character(vars)) {
      cli::cli_abort("{.arg vars} must be a character vector or NULL.")
    }
    missing <- setdiff(vars, names(data))
    if (length(missing)) {
      cli::cli_abort(c(
        "Column{?s} not found in {.arg data}: {.val {missing}}."
      ))
    }
  }
  if (!is.null(adjust)) {
    if (!inherits(adjust, "formula") || length(adjust) != 2L) {
      cli::cli_abort(
        "{.arg adjust} must be a one-sided formula (e.g., {.code ~ s(g, bs = 're')})."
      )
    }
  }
  invisible(NULL)
}

# ---------------------------------------------------------------
# Resolve vars — default to all numeric columns, validate otherwise
# ---------------------------------------------------------------

.resolve_vars <- function(data, vars) {
  if (is.null(vars)) {
    vars <- names(data)[vapply(data, is.numeric, logical(1L))]
    if (length(vars) < 2L) {
      cli::cli_abort(c(
        "{.arg data} has fewer than 2 numeric columns.",
        i = "Supply {.arg vars} explicitly, or add numeric columns."
      ))
    }
    return(vars)
  }
  if (length(vars) < 2L) {
    cli::cli_abort("{.arg vars} must have length >= 2.")
  }
  not_numeric <- vars[!vapply(data[vars], is.numeric, logical(1L))]
  if (length(not_numeric)) {
    cli::cli_abort(c(
      "Non-numeric column{?s} in {.arg vars}: {.val {not_numeric}}.",
      i = "janusplot() models smooth associations only over numeric variables."
    ))
  }
  vars
}

# ---------------------------------------------------------------
# Resolve k: either scalar or named list per variable
# ---------------------------------------------------------------

.resolve_k <- function(k, var_name) {
  if (is.list(k)) {
    k[[var_name]] %||% -1L
  } else {
    k
  }
}

# ---------------------------------------------------------------
# Extract names of variables referenced in an adjust formula RHS
# ---------------------------------------------------------------

.adjust_vars <- function(adjust) {
  if (is.null(adjust)) return(character())
  all.vars(adjust)
}

# ---------------------------------------------------------------
# Build the pairwise GAM formula
# ---------------------------------------------------------------

.build_formula <- function(y_name, x_name, k_val, bs, adjust) {
  rhs <- sprintf("s(`%s`, k = %d, bs = %s)", x_name, as.integer(k_val),
                 deparse(bs))
  if (!is.null(adjust)) {
    adj_rhs <- deparse(adjust[[2L]], width.cutoff = 500L)
    adj_rhs <- paste(adj_rhs, collapse = " ")
    rhs <- paste(rhs, "+", adj_rhs)
  }
  stats::as.formula(sprintf("`%s` ~ %s", y_name, rhs))
}

# ---------------------------------------------------------------
# Subset data to complete cases on the variables a pair uses
# ---------------------------------------------------------------

.complete_pair <- function(data, x_name, y_name, adjust) {
  cols <- unique(c(x_name, y_name, .adjust_vars(adjust)))
  cols <- intersect(cols, names(data))
  sub <- data[, cols, drop = FALSE]
  sub[stats::complete.cases(sub), , drop = FALSE]
}

# ---------------------------------------------------------------
# Hold adjust-covariates at "typical" values for prediction
# numeric -> mean; factor -> most common level
# ---------------------------------------------------------------

.typical_newdata <- function(train_data, x_name, x_grid, adjust) {
  nd <- data.frame(placeholder = x_grid)
  names(nd) <- x_name
  adj_names <- setdiff(.adjust_vars(adjust), x_name)
  for (nm in adj_names) {
    col <- train_data[[nm]]
    if (is.null(col)) next
    nd[[nm]] <- if (is.numeric(col)) {
      mean(col, na.rm = TRUE)
    } else if (is.factor(col)) {
      tb <- sort(table(col), decreasing = TRUE)
      factor(names(tb)[1L], levels = levels(col))
    } else {
      tb <- sort(table(col), decreasing = TRUE)
      names(tb)[1L]
    }
  }
  nd
}

# ---------------------------------------------------------------
# Analytic pointwise derivatives of a univariate mgcv::gam smooth via
# the LP (linear-predictor) matrix. Let X_p be the LP matrix evaluated
# on the plotting grid, with beta = coef(fit) and V_p = fit$Vp the
# Bayesian posterior covariance of the coefficients. If D is a matrix
# whose rows finite-difference the rows of X_p at order k, then
#   f^{(k)}_hat = D %*% beta,   Var(f^{(k)}_hat) = D %*% V_p %*% t(D),
# and the pointwise pairwise SE is sqrt(diag(.)). See Wood (2017)
# Generalized Additive Models, 2nd ed., 7.2.4; Simpson (2018)
# Frontiers in Ecology and Evolution for the simultaneous-CI extension.
#
# When the model also includes `adjust` terms held at typical values,
# the corresponding columns of X_p are identical across grid rows,
# their finite differences are zero, and they contribute nothing to
# either estimate or variance. So the derivative is with respect to
# x_i of the (partial) fit shown in the cell — i.e. exactly what the
# user sees in the fit panel.
#
# Returns a named list keyed by order ("1", "2", ...) of data frames
# with the same schema as fit_obj$pred so the cell renderer can
# consume fit and derivative panels uniformly.
# ---------------------------------------------------------------

.derivatives_lpmatrix <- function(fit, newdata, x_grid, orders) {
  if (!length(orders)) return(list())
  if (is.null(fit) || !inherits(fit, "gam")) return(list())
  if (!is.numeric(x_grid) || length(x_grid) < 3L) return(list())
  h <- mean(diff(x_grid))
  if (!is.finite(h) || h <= 0) return(list())
  Xp <- tryCatch(
    stats::predict(fit, newdata = newdata, type = "lpmatrix"),
    error = function(e) NULL
  )
  if (is.null(Xp) || !is.matrix(Xp) || nrow(Xp) != length(x_grid)) {
    return(list())
  }
  Vp   <- fit$Vp
  beta <- stats::coef(fit)
  if (is.null(Vp) || is.null(beta) ||
      ncol(Xp) != length(beta) || any(dim(Vp) != length(beta))) {
    return(list())
  }
  out <- list()
  for (k in orders) {
    k_int <- as.integer(k)
    if (!is.finite(k_int) || k_int < 1L) next
    D <- .diff_stencil(Xp, h = h, order = k_int)
    if (is.null(D)) next
    d_hat <- as.numeric(D %*% beta)
    # diag(D %*% Vp %*% t(D)) without materialising the full product:
    se_sq <- rowSums((D %*% Vp) * D)
    se_sq[!is.finite(se_sq) | se_sq < 0] <- 0
    se    <- sqrt(se_sq)
    out[[as.character(k_int)]] <- data.frame(
      x   = x_grid,
      fit = d_hat,
      se  = se,
      lo  = d_hat - 1.96 * se,
      hi  = d_hat + 1.96 * se
    )
  }
  out
}

# Simultaneous confidence bands for the derivative curve via the
# Ruppert, Wand & Carroll (2003) / Simpson (2018) Monte Carlo
# construction: draw beta* ~ N(beta_hat, V_p), compute the
# normalised max-deviation statistic max_x |D_i (beta* - beta_hat)|
# / se_i across the plotting grid, and use the (1 - alpha) quantile
# as a critical multiplier on the pointwise SE. Returns only the
# replacement `lo` and `hi` columns (plus the critical multiplier
# for diagnostic record) — the point estimate and pointwise SE are
# unchanged and are reused from `.derivatives_lpmatrix()`.

.derivatives_simultaneous_bands <- function(fit, newdata, x_grid, orders,
                                            n_sim = 1000L, alpha = 0.05) {
  if (!length(orders)) return(list())
  if (is.null(fit) || !inherits(fit, "gam")) return(list())
  if (!is.numeric(x_grid) || length(x_grid) < 3L) return(list())
  h <- mean(diff(x_grid))
  if (!is.finite(h) || h <= 0) return(list())
  Xp <- tryCatch(
    stats::predict(fit, newdata = newdata, type = "lpmatrix"),
    error = function(e) NULL
  )
  if (is.null(Xp) || !is.matrix(Xp) || nrow(Xp) != length(x_grid)) {
    return(list())
  }
  Vp   <- fit$Vp
  beta <- stats::coef(fit)
  p    <- length(beta)
  if (is.null(Vp) || is.null(beta) ||
      ncol(Xp) != p || any(dim(Vp) != p)) return(list())

  # Cholesky of V_p with small diagonal jitter fallback for the
  # occasional near-singular case at high basis dimension.
  L <- tryCatch(
    chol(Vp),
    error = function(e) {
      tryCatch(
        chol(Vp + diag(sqrt(.Machine$double.eps), p)),
        error = function(e2) NULL
      )
    }
  )
  if (is.null(L)) return(list())

  # dB[b, ] = beta*_b - beta_hat, drawn as Z %*% L with Z standard
  # normal (n_sim x p). Keeps memory O(n_sim * p), fast for
  # n_sim = 1000 and basis ranks in the low tens.
  Z  <- matrix(stats::rnorm(n_sim * p), nrow = n_sim)
  dB <- Z %*% L

  out <- list()
  for (k in orders) {
    k_int <- as.integer(k)
    D <- .diff_stencil(Xp, h = h, order = k_int)
    if (is.null(D)) next
    d_hat <- as.numeric(D %*% beta)
    se_sq <- rowSums((D %*% Vp) * D)
    se_sq[!is.finite(se_sq) | se_sq < 0] <- 0
    se_pw <- sqrt(se_sq)
    # Deviation matrix: (n_grid x p) %*% (p x n_sim) = (n_grid x n_sim).
    dev_mat <- D %*% t(dB)
    ok <- se_pw > 0 & is.finite(se_pw)
    if (!any(ok)) next
    dev_std <- matrix(0, nrow = nrow(dev_mat), ncol = ncol(dev_mat))
    dev_std[ok, ] <- dev_mat[ok, ] / se_pw[ok]
    max_abs_dev <- apply(abs(dev_std), 2L, max, na.rm = TRUE)
    crit <- as.numeric(stats::quantile(
      max_abs_dev, probs = 1 - alpha, na.rm = TRUE, names = FALSE
    ))
    if (!is.finite(crit)) next
    out[[as.character(k_int)]] <- list(
      lo = d_hat - crit * se_pw,
      hi = d_hat + crit * se_pw,
      crit_multiplier = crit
    )
  }
  out
}

# Finite-difference stencil on the rows of an LP matrix. Central
# differences in the interior; second-order-accurate forward/backward
# stencils at the two endpoints so the derivative is defined on the
# full plotting grid. Orders beyond 2 iterate the first-order stencil;
# accuracy degrades rapidly and we advise against k >= 3 in the docs.
.diff_stencil <- function(X, h, order) {
  n <- nrow(X)
  if (is.null(n) || n < 3L) return(NULL)
  if (order == 1L) {
    D <- matrix(0, nrow = n, ncol = ncol(X))
    # forward at the left endpoint (second-order): (-3*X[1] + 4*X[2] - X[3]) / (2h)
    D[1L, ]  <- (-3 * X[1L, ] + 4 * X[2L, ] - X[3L, ]) / (2 * h)
    # backward at the right endpoint (second-order)
    D[n, ]   <- (3 * X[n, ]  - 4 * X[n - 1L, ] + X[n - 2L, ]) / (2 * h)
    idx      <- 2:(n - 1L)
    D[idx, ] <- (X[idx + 1L, ] - X[idx - 1L, ]) / (2 * h)
    return(D)
  }
  if (order == 2L) {
    if (n < 4L) return(NULL)
    D <- matrix(0, nrow = n, ncol = ncol(X))
    # Four-point second-order forward/backward stencils at endpoints;
    # three-point central in the interior.
    D[1L, ] <- (2 * X[1L, ] - 5 * X[2L, ] + 4 * X[3L, ] - X[4L, ]) / h^2
    D[n, ]  <- (2 * X[n, ]  - 5 * X[n - 1L, ] +
                4 * X[n - 2L, ] - X[n - 3L, ]) / h^2
    idx      <- 2:(n - 1L)
    D[idx, ] <- (X[idx + 1L, ] - 2 * X[idx, ] + X[idx - 1L, ]) / h^2
    return(D)
  }
  # Order >= 3: iterate. Not exposed in janusplot() by default; present
  # for diagnostic completeness.
  D <- .diff_stencil(X, h, 1L)
  for (i in 2:order) D <- .diff_stencil(D, h, 1L)
  D
}

# ---------------------------------------------------------------
# Pairwise correlations on the complete-case (x, y) subset.
# Computed once per fit and cached on the fit object so downstream
# rendering + data-table construction is free. Also records the
# tie ratio (max of x, y) so callers can flag Spearman coefficients
# computed on heavily tied variables.
# ---------------------------------------------------------------

.compute_correlations <- function(dat, x_name, y_name) {
  na_out <- list(
    cor_pearson  = NA_real_,
    cor_spearman = NA_real_,
    cor_kendall  = NA_real_,
    tie_ratio    = NA_real_
  )
  if (is.null(dat) || nrow(dat) < 3L) return(na_out)
  x <- dat[[x_name]]
  y <- dat[[y_name]]
  ok <- is.finite(x) & is.finite(y)
  x <- x[ok]
  y <- y[ok]
  if (length(x) < 3L) return(na_out)
  safe_cor <- function(m) {
    tryCatch(
      stats::cor(x, y, method = m),
      error = function(e) NA_real_,
      warning = function(w) suppressWarnings(stats::cor(x, y, method = m))
    )
  }
  n <- length(x)
  tie_x <- 1 - length(unique(x)) / n
  tie_y <- 1 - length(unique(y)) / n
  list(
    cor_pearson  = safe_cor("pearson"),
    cor_spearman = safe_cor("spearman"),
    cor_kendall  = safe_cor("kendall"),
    tie_ratio    = max(tie_x, tie_y)
  )
}

# ---------------------------------------------------------------
# Fit a single pairwise GAM and summarise
# ---------------------------------------------------------------

.fit_pair <- function(x_name, y_name, data_full, adjust, method, k, bs,
                     na_action, n_grid = 100L,
                     derivatives = integer(),
                     derivative_ci = "pointwise",
                     derivative_ci_nsim = 1000L, ...) {
  k_val <- .resolve_k(k, x_name)
  if (na_action == "pairwise") {
    dat <- .complete_pair(data_full, x_name, y_name, adjust)
  } else {
    dat <- data_full
  }
  n_used <- nrow(dat)
  if (n_used < 5L) {
    return(.empty_fit_result(x_name, y_name, n_used))
  }

  fml <- .build_formula(y_name, x_name, k_val, bs, adjust)
  fit <- tryCatch(
    mgcv::gam(fml, data = dat, method = method, ...),
    error = function(e) e
  )
  if (inherits(fit, "error")) {
    out <- .empty_fit_result(x_name, y_name, n_used)
    out$error <- conditionMessage(fit)
    return(out)
  }

  smry <- summary(fit)
  # First smooth term is s(x); paranoid guard in case mgcv reorders
  s_row <- which(rownames(smry$s.table) ==
                   sprintf("s(%s)", x_name))[1L]
  if (is.na(s_row)) s_row <- 1L
  edf <- unname(smry$s.table[s_row, "edf"])
  pval <- unname(smry$s.table[s_row, "p-value"])
  dev_exp <- unname(smry$dev.expl %||% NA_real_)

  # Prediction grid on the response scale, holding adjust covariates at typical values
  x_vals <- dat[[x_name]]
  x_grid <- seq(min(x_vals), max(x_vals), length.out = n_grid)
  nd <- .typical_newdata(dat, x_name, x_grid, adjust)
  preds <- tryCatch(
    stats::predict(fit, newdata = nd, se.fit = TRUE),
    error = function(e) NULL
  )
  if (is.null(preds)) {
    pred_df <- data.frame(
      x = x_grid, fit = NA_real_, se = NA_real_,
      lo = NA_real_, hi = NA_real_
    )
  } else {
    pred_df <- data.frame(
      x  = x_grid,
      fit = as.numeric(preds$fit),
      se  = as.numeric(preds$se.fit),
      lo  = as.numeric(preds$fit) - 1.96 * as.numeric(preds$se.fit),
      hi  = as.numeric(preds$fit) + 1.96 * as.numeric(preds$se.fit)
    )
  }

  deriv_list <- if (length(derivatives)) {
    .derivatives_lpmatrix(fit, nd, x_grid, derivatives)
  } else {
    list()
  }
  # If the caller asked for simultaneous bands, replace the
  # pointwise lo/hi in the derivative data frames with the
  # simultaneous ones. Stamp a `ci_type` column on every derivative
  # frame so the renderer and downstream callers can tell them apart.
  if (length(deriv_list)) {
    for (k_nm in names(deriv_list)) {
      deriv_list[[k_nm]]$ci_type <- derivative_ci
    }
    if (identical(derivative_ci, "simultaneous")) {
      sim <- .derivatives_simultaneous_bands(
        fit, nd, x_grid, derivatives,
        n_sim = derivative_ci_nsim
      )
      for (k_nm in names(sim)) {
        if (!is.null(deriv_list[[k_nm]])) {
          deriv_list[[k_nm]]$lo <- sim[[k_nm]]$lo
          deriv_list[[k_nm]]$hi <- sim[[k_nm]]$hi
          attr(deriv_list[[k_nm]], "crit_multiplier") <-
            sim[[k_nm]]$crit_multiplier
        }
      }
    }
  }

  out <- list(
    x_name   = x_name,
    y_name   = y_name,
    fit      = fit,
    pred     = pred_df,
    deriv    = deriv_list,
    raw      = dat[, c(x_name, y_name), drop = FALSE],
    edf      = edf,
    pvalue   = pval,
    dev_exp  = dev_exp,
    n_used   = n_used,
    error    = NA_character_
  )
  out$corr  <- .compute_correlations(dat, x_name, y_name)
  out$shape <- .compute_shape_metrics_raw(out)
  out
}

.empty_fit_result <- function(x_name, y_name, n_used) {
  list(
    x_name = x_name, y_name = y_name, fit = NULL,
    pred = data.frame(x = numeric(), fit = numeric(),
                      se = numeric(), lo = numeric(), hi = numeric()),
    deriv = list(),
    raw = data.frame(),
    edf = NA_real_, pvalue = NA_real_, dev_exp = NA_real_,
    n_used = n_used, error = NA_character_,
    corr  = list(cor_pearson = NA_real_, cor_spearman = NA_real_,
                 cor_kendall = NA_real_, tie_ratio = NA_real_),
    shape = .na_shape_raw()
  )
}

# ---------------------------------------------------------------
# Fit every ordered pair (i != j)
# ---------------------------------------------------------------

.fit_all_pairs <- function(data, vars, adjust, method, k, bs,
                           na_action, parallel,
                           n_grid = 100L,
                           derivatives = integer(),
                           derivative_ci = "pointwise",
                           derivative_ci_nsim = 1000L, ...) {
  .check_parallel_plan(parallel)
  if (na_action == "complete") {
    data <- data[stats::complete.cases(
      data[, unique(c(vars, .adjust_vars(adjust))), drop = FALSE]
    ), , drop = FALSE]
  }
  grid <- expand.grid(i = seq_along(vars), j = seq_along(vars),
                      KEEP.OUT.ATTRS = FALSE)
  grid <- grid[grid$i != grid$j, , drop = FALSE]

  fit_one <- function(i, j) {
    .fit_pair(
      x_name = vars[i], y_name = vars[j],
      data_full = data, adjust = adjust,
      method = method, k = k, bs = bs,
      na_action = na_action,
      n_grid = n_grid, derivatives = derivatives,
      derivative_ci = derivative_ci,
      derivative_ci_nsim = derivative_ci_nsim, ...
    )
  }

  if (isTRUE(parallel) && rlang::is_installed("future.apply")) {
    fits <- future.apply::future_mapply(
      fit_one, grid$i, grid$j, SIMPLIFY = FALSE,
      future.seed = TRUE
    )
  } else {
    fits <- mapply(fit_one, grid$i, grid$j, SIMPLIFY = FALSE)
  }
  names(fits) <- sprintf("%d_%d", grid$i, grid$j)
  fits
}

# ---------------------------------------------------------------
# Asymmetry index in [0, 1]
# ---------------------------------------------------------------

.compute_asymmetry_index <- function(edf_yx, edf_xy) {
  if (is.na(edf_yx) || is.na(edf_xy)) return(NA_real_)
  denom <- edf_yx + edf_xy
  if (!is.finite(denom) || denom <= 0) return(NA_real_)
  abs(edf_yx - edf_xy) / denom
}

# ---------------------------------------------------------------
# Reorder variables by hierarchical clustering of |cor|
# ---------------------------------------------------------------

.reorder_hclust <- function(data, vars) {
  mat <- stats::cor(data[, vars, drop = FALSE],
                    use = "pairwise.complete.obs")
  mat[is.na(mat)] <- 0
  d <- stats::as.dist(1 - abs(mat))
  if (length(vars) < 3L) return(vars)
  h <- stats::hclust(d, method = "average")
  vars[h$order]
}

# ---------------------------------------------------------------
# p-value → glyph
# ---------------------------------------------------------------

.pvalue_to_glyph <- function(p) {
  if (is.na(p)) return("")
  if (p < 0.001) "***"
  else if (p < 0.01) "**"
  else if (p < 0.05) "*"
  else if (p < 0.1) "\u00b7"  # middle dot
  else ""
}

# Tests for the single-display derivative view in janusplot().
# The matrix renders one quantity at a time: "fit", "d1", or "d2".
# Derivative CI rendering is off by default; opt-in via
# derivative_ci = "pointwise" or "simultaneous" (Simpson 2018 MC).

# ---------------------------------------------------------------
# Numerical correctness of .derivatives_lpmatrix()
# ---------------------------------------------------------------

test_that(".derivatives_lpmatrix recovers the true d1 of sin(3x)", {
  withr::with_seed(2026L, {
    n  <- 500L
    x  <- stats::runif(n, -pi, pi)
    y  <- sin(3 * x) + stats::rnorm(n, sd = 0.15)
  })
  d    <- data.frame(x = x, y = y)
  fit  <- mgcv::gam(y ~ s(x, k = 40L), data = d, method = "REML")
  x_g  <- seq(-pi, pi, length.out = 200L)
  nd   <- data.frame(x = x_g)
  dv   <- janusplot:::.derivatives_lpmatrix(fit, nd, x_g, orders = 1:2)

  expect_named(dv, c("1", "2"))
  expect_s3_class(dv[["1"]], "data.frame")
  expect_identical(names(dv[["1"]]), c("x", "fit", "se", "lo", "hi"))
  expect_identical(nrow(dv[["1"]]), 200L)
  expect_true(all(dv[["1"]]$se >= 0))
  expect_true(all(dv[["2"]]$se >= 0))

  int  <- which(x_g > -0.8 * pi & x_g < 0.8 * pi)
  truth_d1 <- 3 * cos(3 * x_g[int])
  truth_d2 <- -9 * sin(3 * x_g[int])
  rmse_d1  <- sqrt(mean((dv[["1"]]$fit[int] - truth_d1)^2))
  rmse_d2  <- sqrt(mean((dv[["2"]]$fit[int] - truth_d2)^2))

  expect_lt(rmse_d1, 0.5)
  expect_lt(rmse_d2, 3.0)
})

# ---------------------------------------------------------------
# Simultaneous-CI construction (Simpson 2018)
# ---------------------------------------------------------------

test_that("simultaneous bands are wider than pointwise, critical >= 1.96", {
  withr::with_seed(2026L, {
    n <- 300L
    x <- stats::runif(n, -pi, pi)
    y <- sin(x) + stats::rnorm(n, sd = 0.2)
  })
  d <- data.frame(x = x, y = y)
  pw  <- janusplot_data(d, derivatives = 1L,
                        derivative_ci = "pointwise")
  sim <- janusplot_data(d, derivatives = 1L,
                        derivative_ci = "simultaneous",
                        derivative_ci_nsim = 2000L)
  dfp <- pw$pairs[[1L]]$deriv_yx[["1"]]
  dfs <- sim$pairs[[1L]]$deriv_yx[["1"]]
  expect_identical(unique(dfp$ci_type), "pointwise")
  expect_identical(unique(dfs$ci_type), "simultaneous")
  w_pw  <- mean(dfp$hi - dfp$lo)
  w_sim <- mean(dfs$hi - dfs$lo)
  expect_gt(w_sim, w_pw)
  crit <- attr(dfs, "crit_multiplier")
  expect_true(is.finite(crit))
  expect_gt(crit, 1.96)
  expect_lt(crit, 6)   # sanity cap
})

test_that("derivative_ci = 'none' returns pointwise lo/hi but is tagged none", {
  d <- make_nonlinear_data(n = 150L, seed = 2L)
  out <- janusplot_data(d, vars = c("x1", "x2"),
                        derivatives = 1L, derivative_ci = "none")
  df <- out$pairs[[1L]]$deriv_yx[["1"]]
  expect_identical(unique(df$ci_type), "none")
  # lo/hi are still pointwise; the renderer honours the tag.
  expect_true(all(is.finite(df$lo)))
  expect_true(all(is.finite(df$hi)))
})

# ---------------------------------------------------------------
# Basis-robustness
# ---------------------------------------------------------------

test_that(".derivatives_lpmatrix works across bases (tp/cr/cs)", {
  d <- make_nonlinear_data(n = 200L, seed = 2L)
  for (bs_type in c("tp", "cr", "cs")) {
    fit <- mgcv::gam(x2 ~ s(x1, bs = bs_type), data = d, method = "REML")
    x_g <- seq(min(d$x1), max(d$x1), length.out = 120L)
    nd  <- data.frame(x1 = x_g)
    dv  <- janusplot:::.derivatives_lpmatrix(fit, nd, x_g, 1L)
    expect_true(
      is.data.frame(dv[["1"]]),
      info = sprintf("basis = %s produced no derivative", bs_type)
    )
    expect_true(all(is.finite(dv[["1"]]$fit)))
    expect_true(all(dv[["1"]]$se >= 0))
  }
})

# ---------------------------------------------------------------
# Edge cases
# ---------------------------------------------------------------

test_that(".derivatives_lpmatrix returns empty list on bad input", {
  expect_identical(
    janusplot:::.derivatives_lpmatrix(NULL, data.frame(), numeric(), 1L),
    list()
  )
  expect_identical(
    janusplot:::.derivatives_lpmatrix(NULL, data.frame(), 1:5, integer()),
    list()
  )
})

test_that(".diff_stencil returns NULL when the grid is too short", {
  X <- matrix(1:4, nrow = 2L)
  expect_null(janusplot:::.diff_stencil(X, h = 1, order = 1L))
  X4 <- matrix(1, nrow = 3L, ncol = 2L)
  expect_null(janusplot:::.diff_stencil(X4, h = 1, order = 2L))
})

test_that(".derivatives_simultaneous_bands returns empty list cleanly", {
  expect_identical(
    janusplot:::.derivatives_simultaneous_bands(
      NULL, data.frame(), numeric(), 1L, n_sim = 100L
    ),
    list()
  )
})

# ---------------------------------------------------------------
# adjust-term invariance
# ---------------------------------------------------------------

test_that("derivative estimates drop adjust covariates held at typical values", {
  d <- make_linear_data(n = 300L, seed = 4L)
  d$site <- factor(sample(letters[1:3], 300L, replace = TRUE))
  f_plain <- janusplot:::.fit_pair(
    x_name = "x1", y_name = "x2", data_full = d,
    adjust = NULL, method = "REML", k = -1L, bs = "tp",
    na_action = "pairwise", derivatives = 1L
  )
  f_adj <- janusplot:::.fit_pair(
    x_name = "x1", y_name = "x2", data_full = d,
    adjust = ~ s(site, bs = "re"),
    method = "REML", k = -1L, bs = "tp",
    na_action = "pairwise", derivatives = 1L
  )
  d1a <- f_plain$deriv[["1"]]$fit
  d1b <- f_adj$deriv[["1"]]$fit
  int <- seq(20, 80)
  rms_diff <- sqrt(mean((d1a[int] - d1b[int])^2))
  rms_lvl  <- sqrt(mean(d1a[int]^2))
  expect_lt(rms_diff / rms_lvl, 0.25)
})

# ---------------------------------------------------------------
# janusplot() integration — scalar display
# ---------------------------------------------------------------

test_that("janusplot(display = 'fit') returns the legacy single-panel cell type", {
  # Each off-diagonal cell must still be a plain ggplot (not a
  # patchwork) — this is the invariant that keeps legacy visual
  # snapshots interpretable.
  d <- make_linear_data(n = 120L, seed = 1L)
  f <- janusplot:::.fit_pair(
    x_name = "x1", y_name = "x2", data_full = d,
    adjust = NULL, method = "REML", k = -1L, bs = "tp",
    na_action = "pairwise"
  )
  cell <- janusplot:::.build_cell(
    fit_obj = f, show_data = TRUE, show_ci = TRUE,
    colour_by = "spearman", palette = "RdBu",
    signif_glyph = TRUE, annotations = c("edf", "A"),
    shape_cutoffs = janusplot_shape_cutoffs(),
    glyph_style = "ascii", asym_val = 0.1,
    colour_limits = c(-1, 1), is_upper = TRUE,
    text_sizes = janusplot:::.cell_text_sizes(3L),
    display = "fit"
  )
  expect_s3_class(cell, "ggplot")
  expect_false(inherits(cell, "patchwork"))
})

test_that("janusplot(display = 'd1') renders a derivative panel cell", {
  d <- make_nonlinear_data(n = 150L, seed = 2L)
  p <- janusplot(d, vars = c("x1", "x2"), display = "d1")
  expect_true(inherits(p, "ggplot") || inherits(p, "patchwork"))
})

test_that("janusplot(display = 'd2', derivative_ci = 'pointwise') runs", {
  d <- make_nonlinear_data(n = 150L, seed = 2L)
  p <- janusplot(d, vars = c("x1", "x2"),
                 display = "d2", derivative_ci = "pointwise")
  expect_true(inherits(p, "ggplot") || inherits(p, "patchwork"))
})

test_that("display validation rejects unknown values", {
  d <- make_linear_data(n = 60L, seed = 1L)
  expect_error(janusplot(d, vars = c("x1", "x2"), display = "d3"))
  expect_error(janusplot(d, vars = c("x1", "x2"),
                         display = c("fit", "d1")))   # vector no longer allowed
})

test_that("derivative_ci validation + nsim validation", {
  d <- make_linear_data(n = 60L, seed = 1L)
  expect_error(janusplot(d, vars = c("x1", "x2"),
                         display = "d1", derivative_ci = "banana"))
  expect_error(janusplot(d, vars = c("x1", "x2"),
                         display = "d1",
                         derivative_ci = "simultaneous",
                         derivative_ci_nsim = 10))
})

test_that("n_grid validation is strict and auto-resolves", {
  d <- make_linear_data(n = 60L, seed = 1L)
  expect_error(janusplot(d, vars = c("x1", "x2"), n_grid = 5))
  expect_error(janusplot(d, vars = c("x1", "x2"), n_grid = "fifty"))
  out <- janusplot_data(d, vars = c("x1", "x2"))
  expect_identical(nrow(out$pairs[[1L]]$pred_yx), 100L)
  out2 <- janusplot_data(d, vars = c("x1", "x2"), derivatives = 1L)
  expect_identical(nrow(out2$pairs[[1L]]$pred_yx), 200L)
  expect_identical(nrow(out2$pairs[[1L]]$deriv_yx[["1"]]), 200L)
})

# ---------------------------------------------------------------
# with_data = TRUE carries a `display` column
# ---------------------------------------------------------------

test_that("summary table tags display mode on every call", {
  d <- make_nonlinear_data(n = 120L, seed = 2L)
  out_fit <- janusplot(d, vars = c("x1", "x2"), with_data = TRUE)
  out_d1  <- janusplot(d, vars = c("x1", "x2"),
                       display = "d1", with_data = TRUE)
  out_d2  <- janusplot(d, vars = c("x1", "x2"),
                       display = "d2", with_data = TRUE)
  expect_true("display" %in% colnames(out_fit$data))
  expect_identical(unique(out_fit$data$display), "fit")
  expect_identical(unique(out_d1$data$display), "d1")
  expect_identical(unique(out_d2$data$display), "d2")
})

# ---------------------------------------------------------------
# janusplot_data() surfaces derivatives (multi-order allowed)
# ---------------------------------------------------------------

test_that("janusplot_data surfaces deriv_yx / deriv_xy with ci_type tag", {
  d <- make_nonlinear_data(n = 120L, seed = 2L)
  out <- janusplot_data(d, vars = c("x1", "x2"),
                        derivatives = c(1L, 2L),
                        derivative_ci = "pointwise")
  p   <- out$pairs[[1L]]
  expect_named(p$deriv_yx, c("1", "2"))
  expect_named(p$deriv_xy, c("1", "2"))
  expect_true(all(c("x", "fit", "se", "lo", "hi", "ci_type") %in%
                  names(p$deriv_yx[["1"]])))
  expect_identical(unique(p$deriv_yx[["1"]]$ci_type), "pointwise")
})

test_that("janusplot_data derivatives validation rejects bad orders", {
  d <- make_linear_data(n = 60L, seed = 1L)
  expect_error(janusplot_data(d, vars = c("x1", "x2"), derivatives = 3L))
  expect_error(janusplot_data(d, vars = c("x1", "x2"),
                              derivatives = c(1L, 1L)))
})

# ---------------------------------------------------------------
# Matrix title — what display mode labels the top of the plot
# ---------------------------------------------------------------

test_that("diagonal = 'density' renders a density+rug ggplot per variable", {
  d <- make_nonlinear_data(n = 200L, seed = 2L)
  p <- janusplot(d, vars = c("x1", "x2"), diagonal = "density",
                 show_shape_legend = FALSE, show_glossary = FALSE)
  expect_true(inherits(p, "ggplot") || inherits(p, "patchwork"))

  cell <- janusplot:::.build_density_rug_cell("x1", d$x1)
  expect_s3_class(cell, "ggplot")
  # Two layers in the data frame source (geom_area + geom_line +
  # geom_rug = 3 layers); the cell function must produce a real plot
  # not the blank fallback.
  expect_gte(length(cell$layers), 3L)
})

test_that("diagonal = 'density' falls back to blank when n < 5", {
  cell <- janusplot:::.build_density_rug_cell("x", c(1, 2, NA, NA))
  expect_s3_class(cell, "ggplot")
})

test_that("diagonal validation rejects unknown value", {
  d <- make_linear_data(n = 60L, seed = 1L)
  expect_error(janusplot(d, vars = c("x1", "x2"), diagonal = "violin"))
})

test_that(".display_title emits the three mode labels", {
  expect_match(janusplot:::.display_title("fit"), "Direct fit")
  expect_match(janusplot:::.display_title("d1"), "First derivative")
  expect_match(janusplot:::.display_title("d2"), "Second derivative")
  # Unicode primes when glyph_style = "unicode"
  expect_match(janusplot:::.display_title("d1", "unicode"), "\u2032")
  expect_match(janusplot:::.display_title("d2", "unicode"), "\u2033")
})

# ---------------------------------------------------------------
# vdiffr snapshots for the three display modes
# ---------------------------------------------------------------

test_that("janusplot display = 'd1' snapshot", {
  skip_if_not_installed("vdiffr")
  d <- make_nonlinear_data(n = 150L, seed = 2L)
  p <- janusplot(d, vars = c("x1", "x2"),
                 display = "d1",
                 colour_by = "none",
                 show_shape_legend = FALSE,
                 show_glossary = FALSE)
  vdiffr::expect_doppelganger("matrix-2var-d1-plain", p)
})

test_that("janusplot display = 'd1' simultaneous-CI snapshot", {
  skip_if_not_installed("vdiffr")
  withr::with_seed(2026L, {
    d <- make_nonlinear_data(n = 150L, seed = 2L)
    p <- janusplot(d, vars = c("x1", "x2"),
                   display = "d1",
                   derivative_ci = "simultaneous",
                   derivative_ci_nsim = 500L,
                   colour_by = "none",
                   show_shape_legend = FALSE,
                   show_glossary = FALSE)
  })
  vdiffr::expect_doppelganger("matrix-2var-d1-simultaneous", p)
})

test_that("asymmetry index is zero when EDFs match", {
  expect_equal(janusplot:::.compute_asymmetry_index(1.5, 1.5), 0)
})

test_that("asymmetry index is one when one EDF is zero", {
  expect_equal(janusplot:::.compute_asymmetry_index(0, 3), 1)
})

test_that("asymmetry index lies in [0, 1]", {
  vals <- vapply(1:50, function(i) {
    a <- stats::runif(1, 0.5, 8)
    b <- stats::runif(1, 0.5, 8)
    janusplot:::.compute_asymmetry_index(a, b)
  }, numeric(1L))
  expect_true(all(vals >= 0 & vals <= 1))
})

test_that("asymmetry index is NA when either input is NA", {
  expect_true(is.na(janusplot:::.compute_asymmetry_index(NA_real_, 1)))
  expect_true(is.na(janusplot:::.compute_asymmetry_index(1, NA_real_)))
})

test_that("p-value glyphs follow conventional cutoffs", {
  expect_equal(janusplot:::.pvalue_to_glyph(0.0001), "***")
  expect_equal(janusplot:::.pvalue_to_glyph(0.005),  "**")
  expect_equal(janusplot:::.pvalue_to_glyph(0.03),   "*")
  expect_equal(janusplot:::.pvalue_to_glyph(0.08),   "\u00b7")
  expect_equal(janusplot:::.pvalue_to_glyph(0.5),    "")
  expect_equal(janusplot:::.pvalue_to_glyph(NA_real_), "")
})

---
title: "Asymmetric Smoothed-Association Matrices"
output:
  rmarkdown::html_vignette:
    toc: true
    fig_width: 7
    fig_height: 7
vignette: >
  %\VignetteIndexEntry{Asymmetric Smoothed-Association Matrices}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse   = TRUE,
  comment    = "#>",
  fig.alt    = paste(
    "Asymmetric smoothed-association matrix produced by",
    "janusplot(); diagonal cells hold variable labels,",
    "off-diagonal cells show the fitted mgcv::gam spline",
    "with a 95% confidence envelope, raw-data scatter,",
    "and per-cell annotations for n, EDF, and smooth",
    "significance glyph."
  ),
  fig.width  = 7,
  fig.height = 7,
  dpi        = 110,
  eval       = TRUE
)
library(ggplot2)
set.seed(2026L)
```

## Why Pearson is not enough

A Pearson correlation matrix gives one scalar per pair of variables.
Two numbers are discarded in that collapse:

1. The *shape* of the association — linear, monotone non-linear,
   U-shaped, or irregular.
2. The *direction* — whether $y$ is a smooth function of $x$
   differs in general from whether $x$ is a smooth function of $y$,
   because leverage and noise are directional.

`janusplot()` renders both recoveries visually for every pair in a
matrix layout, using proper `mgcv::gam()` fits (not loess) so EDF,
F-tests, and random effects are available.

## Quick start

```{r quickstart}
library(janusplot)

# Four numeric columns from mtcars (32 rows: small but illustrative)
janusplot(mtcars[, c("mpg", "hp", "wt", "qsec")])
```

Each off-diagonal cell shows:

- raw scatter (light grey),
- the fitted spline (blue line) and 95% CI ribbon,
- EDF (effective degrees of freedom) in the bottom-right,
- *n* used in the bottom-left,
- a signif-glyph in the top-right (`***` / `**` / `*` / `·`).

The cell fill is keyed to EDF: darker = more non-linear.

## Non-linear detection

A synthetic quadratic + sinusoidal example. The matrix makes it
obvious which variables are genuinely non-linearly related to which.

```{r nonlinear}
n <- 300
x1 <- runif(n, -3, 3)
x2 <- x1^2 + rnorm(n, sd = 0.6)   # quadratic on x1
x3 <- sin(x1) + rnorm(n, sd = 0.4) # sinusoidal on x1
x4 <- rnorm(n)                     # independent
d  <- data.frame(x1 = x1, x2 = x2, x3 = x3, x4 = x4)

janusplot(d)
```

EDF for `x2 ~ s(x1)` and `x3 ~ s(x1)` should clearly exceed 1; the
cell fills reflect that. Cells involving `x4` should be close to EDF =
1 (linear / flat).

## Asymmetry — a heteroscedastic example

When the noise scale depends on a predictor, the two directional smooths
diverge: $y \sim s(x)$ recovers the mean relationship; $x \sim s(y)$ is
distorted by the variance asymmetry.

```{r asym}
n <- 400
x <- runif(n, 0, 5)
y <- 0.5 * x + rnorm(n, sd = 0.3 + 0.4 * x)   # variance grows with x
d <- data.frame(x = x, y = y, z = rnorm(n))

janusplot(d)
```

The `A = ...` label per cell reports the asymmetry index
$A_{ij} = |EDF_{y|x} - EDF_{x|y}| / (EDF_{y|x} + EDF_{x|y}) \in [0, 1]$,
shown by default in the bottom-left corner alongside `EDF = ...`.

## Partial smooths (controlling for covariates)

Pass `adjust =` as a one-sided formula RHS to include fixed covariates
and/or random effects in every pairwise GAM.

```{r adjust, eval = requireNamespace("palmerpenguins", quietly = TRUE)}
library(palmerpenguins)
pp <- na.omit(penguins)

# Without covariate
janusplot(pp[, c("bill_length_mm", "bill_depth_mm",
                 "flipper_length_mm", "body_mass_g")])

# With species as a fixed effect — resolves Simpson's-paradox geometry
janusplot(pp, vars = c("bill_length_mm", "bill_depth_mm",
                       "flipper_length_mm", "body_mass_g"),
         adjust = ~ species)
```

## Changing the palette

The cell fill encodes the EDF (or deviance-explained) of the smooth and
is accompanied by a shared colourbar legend. Choose a palette with
`palette = `.

```{r palette-viridis, fig.height = 5, fig.width = 6}
d <- data.frame(
  x1 = runif(200, -3, 3),
  x2 = rnorm(200),
  x3 = rnorm(200)
)
d$x2 <- d$x1^2 + rnorm(200, sd = 0.8)  # non-linear

janusplot(d, palette = "viridis")  # default, colourblind-safe
```

```{r palette-brewer, fig.height = 5, fig.width = 6}
janusplot(d, palette = "RdYlBu")   # diverging, colourblind-safe
```

```{r palette-turbo, fig.height = 5, fig.width = 6}
janusplot(d, palette = "turbo")    # high-contrast, NOT colourblind-safe
```

**Colourblind-safe choices:**

- *Sequential:* `viridis` (default), `magma`, `inferno`, `plasma`,
  `cividis`, `mako`, `rocket`, `YlOrRd`, `YlGnBu`, `Blues`, `Greens`.
- *Diverging:* `RdYlBu`, `RdBu`, `PuOr`.

**High-contrast but not colourblind-safe:** `turbo`, `Spectral`.

## Handling missing data

```{r missing}
# airquality has genuine NAs in Ozone and Solar.R
janusplot(airquality[, c("Ozone", "Solar.R", "Wind", "Temp")],
          na_action = "pairwise")
```

`na_action = "pairwise"` uses all rows for which *that* pair is
complete; `"complete"` restricts to rows complete across every
variable (matching listwise deletion).

## Scaling up — `order = "hclust"`

For k large, reorder the axes by hierarchical clustering on
|correlation|:

```{r hclust, eval = requireNamespace("MASS", quietly = TRUE)}
data(Boston, package = "MASS")
janusplot(Boston[, c("medv", "lstat", "rm", "age",
                     "indus", "nox", "dis")],
          order = "hclust")
```

## Programmatic access — `janusplot_data()`

Returns raw GAM fits and per-cell metrics without constructing a
ggplot — useful for custom rendering or downstream analysis.

```{r data}
# Re-create the heteroscedastic example
n <- 400
het <- data.frame(
  x = runif(n, 0, 5),
  y = NA_real_
)
het$y <- 0.5 * het$x + rnorm(n, sd = 0.3 + 0.4 * het$x)

out <- janusplot_data(het, vars = c("x", "y"))
out$pairs[[1L]]$edf_yx
out$pairs[[1L]]$edf_xy
out$pairs[[1L]]$asymmetry_index
```

## Shape metrics explained

Every fitted smooth is summarised by two continuous indices and two
discrete counts. These drive the 24-category classifier and appear
as columns in `janusplot(..., with_data = TRUE)$data` and as fields
on each entry of `janusplot_data()$pairs`.

Let `f(x)` be the fitted smooth on a dense grid of 200 equally-spaced
points across the predictor range, with `f'` and `f''` the numerical
first and second derivatives. Let `w(x)` be the empirical density of
the predictor on the same grid, normalised to `sum(w) = 1`.

- **`monotonicity_index`** (paper symbol `M`):

    `M = sum(w * f') / sum(w * |f'|)  in  [-1, 1]`

    `+1` means strictly increasing, `-1` strictly decreasing, `0` a
    non-monotone curve (bowl, dome, wave).

- **`convexity_index`** (paper symbol `C`):

    `C = sum(w * f'') / sum(w * |f''|)  in  [-1, 1]`

    `+1` means globally convex (bowl-up), `-1` globally concave
    (bowl-down), `0` inflection-dominated (S-curve, sine, flat).

Both indices are density-weighted so they describe the smooth
*where the data actually live*, not extrapolated tails, and are
scale-invariant: replacing `y` with `a * y + b` leaves them unchanged.

- **`n_turning_points`** — count of interior extrema (sign changes
  of `f'`), robust to noise via lobe-mass weighting.
- **`n_inflections`** — count of interior curvature flips (sign
  changes of `f''`), same robust counting.

Together the pair `(n_turning_points, n_inflections)` drives the
primary `shape_category` dispatch; `(monotonicity_index,
convexity_index)` disambiguate within the monotone `(0, 0)` and
single-extremum `(1, 0)` cells. The full taxonomy with 2-letter
codes, archetypes, and thumbnail curves is available from
`janusplot_shape_hierarchy()` and is rendered as the standing legend
below every `janusplot()` call.

Tune the thresholds applied to these indices via
`janusplot_shape_cutoffs()`. See the
[shape-recognition-sensitivity](shape-recognition-sensitivity.html)
vignette for how faithfully the classifier recovers ground-truth
shapes across sample-size and noise regimes.

## Derivative views: theoretical justification and applied use

Each matrix renders one quantity. `display = "fit"` (default) shows
the fitted smooth; `display = "d1"` shows \eqn{\hat f'(x)};
`display = "d2"` shows \eqn{\hat f''(x)}. A top-of-matrix title
names the mode, so side-by-side calls compare unambiguously.
Orders beyond two are not exposed — see *Noise amplification*
below. Derivative CI rendering is **off by default**; opt in with
`derivative_ci = "pointwise"` or `"simultaneous"`.

```{r derivs-fit, fig.height = 4.5}
set.seed(2026L)
n  <- 300L
xs <- runif(n, -pi, pi)
df <- data.frame(
  x  = xs,
  y1 = xs + sin(3 * xs) + rnorm(n, sd = 0.15),
  y2 = 0.5 * xs^2       + rnorm(n, sd = 0.8)
)
janusplot(df, display = "fit", show_shape_legend = FALSE)
```

```{r derivs-d1, fig.height = 4.5}
janusplot(df, display = "d1", show_shape_legend = FALSE)
```

```{r derivs-d2, fig.height = 4.5}
janusplot(df, display = "d2", show_shape_legend = FALSE)
```

Turn on simultaneous bands — a single call gets the Monte Carlo
critical multiplier per Simpson (2018):

```{r derivs-sim, fig.height = 4.5}
janusplot(df, display = "d1",
          derivative_ci = "simultaneous",
          derivative_ci_nsim = 2000L,
          show_shape_legend = FALSE)
```

### What derivatives reveal that the fit hides

The fitted smooth $\hat f(x) = \mathbb{E}[y\mid x]$ is a level
description. Its derivatives are different statistical objects with
their own interpretations:

- $\hat f'(x)$ — the **local rate of change** of $y$ in $x$. Zero
  crossings localise the turning points of $\hat f$; the sign of
  $\hat f'$ gives the direction of monotonicity; the magnitude gives
  the sensitivity at the operating point $x$. In control engineering
  this is literally the process gain $K(x) = \partial y / \partial u$
  that gain-scheduled controllers are built around (Rugh & Shamma,
  2000; Leith & Leithead, 2000). In causal analysis of a continuous
  treatment it is the derivative of the dose--response curve
  $\mu'(t) = \partial \mathbb{E}[Y(t)] / \partial t$, which Zhang &
  Chen (2025) argue is often *the* treatment-effect object of
  interest, not the curve itself.
- $\hat f''(x)$ — the **local curvature**. Zero crossings localise
  the inflection points of $\hat f$; a persistently positive
  second derivative flags accelerating growth, persistently negative
  flags saturation (diminishing returns). $\hat f''$ is the input to
  the convexity index $C$ defined earlier in this vignette, so the
  derivative panel exposes the *local* signal behind that scalar
  summary.

The asymmetric matrix layout sharpens this. `janusplot()` fits both
$\hat f_{y\mid x}(x)$ and $\hat f_{x\mid y}(y)$, so derivative panels
on the two triangles answer genuinely different questions: the upper
triangle is "how steeply does $y$ respond to a nudge in $x$ at this
operating point" (forward gain); the lower triangle is "how steeply
must $x$ change to induce a unit change in $y$" (inverse
sensitivity). For an asymmetric process these do not transpose into
each other, and the directional asymmetry is a diagnostic the
symmetric correlation matrix cannot expose (Janzing & Schölkopf,
2010).

### Estimation — the LP matrix

Let $X_p = X_p(\mathbf{x}_g)$ denote the design (linear predictor)
matrix of the fitted GAM evaluated on the plotting grid
$\mathbf{x}_g$, obtained from
`predict(gam_fit, newdata = ..., type = "lpmatrix")` (Wood, 2017,
§7.2.4). With penalised posterior mean $\hat{\boldsymbol\beta} =
\mathtt{coef(gam\_fit)}$ and posterior covariance $V_p =
\mathtt{gam\_fit\$Vp}$, we construct a finite-difference operator
$D^{(k)}$ on the *rows* of $X_p$ (central differences in the interior,
second-order forward / backward stencils at the endpoints) and read
off

$$
\hat f^{(k)}(x_i) = \bigl[D^{(k)} \hat{\boldsymbol\beta}\bigr]_i,
\qquad
\widehat{\mathrm{Var}}\bigl(\hat f^{(k)}(x_i)\bigr) = \bigl[D^{(k)} V_p
\bigl(D^{(k)}\bigr)^{\!\top}\bigr]_{ii}.
$$

Pointwise $95\%$ intervals are $\hat f^{(k)}(x) \pm 1.96\,\sqrt{\cdot}$.
This is the standard Wood (2017) construction, and is what
`gratia::derivatives()` implements in its default mode (Simpson,
2014; Simpson, 2018). Columns of $X_p$ corresponding to `adjust`
terms held at typical values contribute identical rows across the
grid, so their finite differences are zero and they drop out of both
$\hat f^{(k)}$ and its variance — the derivative in the panel is
therefore the derivative of the *partial smooth actually shown in
the fit panel*, as expected.

For simultaneous intervals over the full grid (a stricter question
than pointwise, and what you want for formal feature localisation),
`janusplot()` implements the Monte Carlo construction of Ruppert,
Wand & Carroll (2003, §6.5), popularised for GAMs by Simpson (2018):
draw $\tilde{\boldsymbol\beta}_b \sim
\mathcal{N}(\hat{\boldsymbol\beta}, V_p)$ for $b = 1, \ldots, B$ and
take the $(1-\alpha)$ quantile of $\max_i |D^{(k)}_i
(\tilde{\boldsymbol\beta}_b - \hat{\boldsymbol\beta})| /
\mathtt{se}_i$ across the plotting grid as a critical multiplier
$c_\alpha$ on the pointwise SE, so the simultaneous band is
$\hat f^{(k)}(x) \pm c_\alpha\,\mathtt{se}(x)$. Opt in via
`derivative_ci = "simultaneous"` on either `janusplot()` or
`janusplot_data()`; the default is `derivative_ci = "none"` so that
no CI is drawn by default — derivative ribbons invite over-reading
of local features and should be a deliberate choice, not a
default. The implementation uses $B = 1000$ (see
`derivative_ci_nsim`); Simpson (2018) uses $10\,000$, which is
affordable if you need tighter quantile estimation.

### Noise amplification and why we cap at $k = 2$

Finite differencing of raw data amplifies noise; penalised splines
do not eliminate that amplification, they trade it against bias via
the REML-selected smoothing parameter. `mgcv`'s default thin-plate
penalty is on $\int (f'')^2$, which directly regularises $\hat f'$
and bounds (but does not penalise) $\hat f''$ only via the basis
rank (Wood, 2017, §5.3; Eilers & Marx, 1996). In practice we find
$\hat f^{(3)}$ is dominated by noise for $n < 10^4$ at moderate $k$,
and so janusplot refuses $k \ge 3$ by design. If you have a
domain-specific reason to need a higher-order derivative, specify a
matching-order P-spline penalty explicitly (Eilers, Marx & Durbán,
2015) and extract it yourself from `janusplot_data()`.

### Applied use: gain estimation and dose--response

Two strands in which the asymmetric derivative view is not a
cosmetic add-on but the analytical primitive the practitioner
actually wants.

- **Process-gain scheduling.** In adaptive and gain-scheduled
  control, the controller is indexed by the local process gain
  $K(x) = \partial y / \partial u$ (Rugh & Shamma, 2000). For a
  steady-state input-output dataset, $\hat f_{y\mid u}'(u)$ is a
  direct data-driven estimate of $K(u)$, and its simultaneous CI
  tells the engineer whether the local gain is distinguishable from
  a reference gain over an operating envelope. The inverse panel
  $\hat f_{u\mid y}'(y)$ is the feedforward-linearisation
  sensitivity; a large divergence between the two panels flags that
  a naive inverse controller will under-perform (Korda & Mezić,
  2018). The matrix view makes a fleet of such pairs inspectable at
  once.
- **Derivative of the dose--response curve as the causal estimand.**
  For a continuous treatment $T$ with unconfoundedness, the
  dose--response $\mu(t) = \mathbb{E}[Y(t)]$ and its derivative
  $\mu'(t)$ are both estimable, and recent work (Zhang & Chen, 2025)
  argues $\mu'(t)$ is often the more directly interpretable
  quantity — it answers "how much does the expected outcome change
  per unit shift in treatment at this dose?" This is structurally
  the same estimand as the process gain above; the
  asymmetric-matrix derivative panel delivers both forward and
  reverse-conditioned derivative curves in the same frame, which is
  a direct diagnostic for Simpson's-paradox-style conditioning
  reversals (visible, for example, in the penguins
  `bill_depth_mm` $\times$ `body_mass_g` pair once `species` is
  adjusted for).

### References cited in this section

Eilers, P. H. C., & Marx, B. D. (1996). Flexible smoothing with
B-splines and penalties. *Statistical Science*, **11**(2), 89--121.
<https://doi.org/10.1214/ss/1038425655>

Eilers, P. H. C., Marx, B. D., & Durbán, M. (2015). Twenty years of
P-splines. *SORT*, **39**(2), 149--186.

Janzing, D., & Schölkopf, B. (2010). Causal inference using the
algorithmic Markov condition. *IEEE Transactions on Information
Theory*, **56**(10), 5168--5194.
<https://doi.org/10.1109/TIT.2010.2060095>

Korda, M., & Mezić, I. (2018). Linear predictors for nonlinear
dynamical systems: Koopman operator meets model predictive control.
*Automatica*, **93**, 149--160.
<https://doi.org/10.1016/j.automatica.2018.03.046>

Leith, D. J., & Leithead, W. E. (2000). Survey of gain-scheduling
analysis and design. *International Journal of Control*, **73**(11),
1001--1025. <https://doi.org/10.1080/002071700411304>

Rugh, W. J., & Shamma, J. S. (2000). Research on gain scheduling.
*Automatica*, **36**(10), 1401--1425.
<https://doi.org/10.1016/S0005-1098(00)00058-3>

Ruppert, D., Wand, M. P., & Carroll, R. J. (2003). *Semiparametric
Regression*. Cambridge University Press.

Simpson, G. L. (2014). Simultaneous confidence intervals for
derivatives of smooth terms in a GAM. *From the Bottom of the Heap*
(blog post).

Simpson, G. L. (2018). Modelling palaeoecological time series using
generalised additive models. *Frontiers in Ecology and Evolution*,
**6**, 149. <https://doi.org/10.3389/fevo.2018.00149>

Wood, S. N. (2017). *Generalized Additive Models: An Introduction
with R* (2nd ed.). Chapman and Hall/CRC.
<https://doi.org/10.1201/9781315370279>

Zhang, Y., & Chen, Y.-C. (2025). Doubly robust inference on causal
derivative effects for continuous treatments. arXiv preprint
[2501.06969].

## Limitations

- Pairwise view, not conditional — always complement with a proper
  multivariate model.
- EDF depends on basis dimension `k`; defaults are sensible but
  domain-specific tuning is encouraged.
- The asymmetry index should not be interpreted causally without
  strong assumptions.
- `monotonicity_index` and `convexity_index` are scale-invariant in
  `y` but sensitive to the predictor-density weighting — they
  describe the smooth on the observed support of `x`, not outside
  it.
- `display` is scalar — a single `janusplot()` call renders a
  single quantity (fit, d1, or d2). To compare fit against
  derivative, issue two or three calls; each carries its own
  matrix-level title and, when `with_data = TRUE`, its own
  `display`-tagged summary table.
- Derivative panels show **no confidence ribbon by default**
  (`derivative_ci = "none"`). Opt in explicitly: `"pointwise"` for
  marginally-valid pointwise 95% bands, `"simultaneous"` for
  Simpson (2018) Monte Carlo bands valid for feature localisation.
- Requesting `display %in% c("d1", "d2")` raises the default
  prediction-grid resolution from 100 to 200 points, which slightly
  shifts the numeric shape-metric values (`M`, `C`, turning and
  inflection counts) reported alongside the fit. Shapes and
  asymmetry — the primary reading of the matrix — are robust to
  this drift; `M`, `C` and the counts are secondary diagnostics.
  The precomputed `shape_sensitivity_demo` dataset was generated
  under `n_grid = 100` and is preserved as-is for reproducibility.

## Citation

```{r cite}
citation("janusplot")
```

---
title: "Shape-recognition sensitivity study"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Shape-recognition sensitivity study}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.align = "center",
  fig.width = 8,
  fig.height = 6,
  dpi       = 110
)
```

`janusplot()` assigns every fitted smooth to one of 24 shape
categories via a `(n_turning_points, n_inflections)` dispatch with
additional `(monotonicity_index, convexity_index)` disambiguation
for the monotone cases (see the `janusplot` vignette for the full
definition of the indices). How reliably does this classifier
recover the ground-truth shape of a noisy sample? This vignette
answers the question with a
full-factorial sensitivity sweep.

## Design

For each combination of ground-truth shape, sample size `n`, and
noise level `sigma`, the sweep:

1. Generates `n` points from the noiseless canonical curve on
   `x ∈ [0, 1]`, with `y` normalised to `[0, 1]` so that `sigma` is
   the fraction of y-range that Gaussian noise contributes — an
   SNR-comparable scale across shapes.
2. Fits `mgcv::gam(y ~ s(x), method = "REML")`.
3. Classifies the fit via `janusplot_shape_metrics()`.
4. Records correctness at the **fine** (24-category) and
   **archetype** (7-family) levels.

The design factors are orthogonal and replicated. See
`?janusplot_shape_sensitivity` for the function surface. The 14
canonical ground-truth shapes cover five of the seven archetypes
(`chaotic` and `degenerate` have no realistic deterministic
generator).

```{r setup-pkg}
library(janusplot)
library(ggplot2)

janusplot_shape_sensitivity_shapes()
```

## Pre-registered hypotheses

The sweep's hypotheses are pinned in
`simulation/PLAN.md` (Scenario 4):

- **H1.** At `n = 500`, `sigma = 0.05`, archetype accuracy
  exceeds 0.90 for every shape.
- **H2.** Fine-category accuracy exceeds 0.75 at `n = 500`,
  `sigma = 0.05` for monotone + unimodal shapes; wave and multimodal
  tolerate less noise.
- **H3.** Rippled variants require `n ≥ 200` and `sigma ≤ 0.10`
  to resolve.
- **H4.** At `sigma = 0.40`, archetype accuracy collapses below 0.50
  for all but the simplest shapes.

## Precomputed demo

The package ships a small-footprint precomputed sweep — 6 shapes
(one per non-degenerate archetype) × 3 sample sizes × 4 noise levels ×
30 replicates = 2160 fits — so you can explore the API without
running the full sweep yourself.

```{r demo-data}
data("shape_sensitivity_demo")
str(shape_sensitivity_demo, vec.len = 2)
```

### Recovery curves (headline figure)

```{r recovery-curves}
janusplot_shape_sensitivity_plot(shape_sensitivity_demo,
                                 "recovery_curves")
```

Every shape is recovered near-perfectly at low noise; the
informative picture is where each shape's curve falls off as sigma
grows. The unimodal and monotone-curved families tolerate more noise
than the multimodal ones.

### Archetype confusion

```{r archetype-confusion, fig.width = 6, fig.height = 5}
janusplot_shape_sensitivity_plot(shape_sensitivity_demo,
                                 "confusion_archetype")
```

The off-diagonals reveal the classifier's failure modes. A `unimodal`
truth misclassified as `wave` or `multimodal` means the spline
invented extra turning points under noise.

### Archetype-level accuracy grid

```{r accuracy-grid}
janusplot_shape_sensitivity_plot(shape_sensitivity_demo,
                                 "accuracy_grid")
```

Per-shape heatmap of `P(archetype correct)` across the `(n, sigma)`
design. Reading across a row shows the noise-tolerance profile of
one sample size; reading up a column shows the sample-size
sensitivity at one noise level.

### Numerical summary

```{r summary}
head(janusplot_shape_sensitivity_summary(shape_sensitivity_demo,
                                         level = "archetype"), 10)
```

## Running your own sweep

The demo is a starting point. For the publication-grade figure
use the full default grid (14 shapes × 4 sample sizes × 5 noise levels
× 200 reps = 56 000 fits):

```{r full-sweep, eval = FALSE}
# Configure parallel execution (optional) — you control the plan.
future::plan(future::multisession, workers = 4L)

res <- janusplot_shape_sensitivity(parallel = TRUE)

# Save for your paper
saveRDS(res, "shape_sensitivity_full.rds")
janusplot_shape_sensitivity_plot(res, "recovery_curves")
```

### Custom shape subsets + cutoffs

Every argument is tunable. Below, we rerun only the bimodal/wave
family under stricter monotonicity thresholds to see whether
tightening `mono_strong` buys any fine-accuracy improvement for these
categories.

```{r custom-subset, eval = FALSE}
strict <- janusplot_shape_cutoffs(mono_strong = 0.95, curv_low = 0.1)

res_strict <- janusplot_shape_sensitivity(
  shapes     = c("wave", "bimodal", "bi_wave"),
  n_grid     = c(200L, 500L),
  sigma_grid = c(0.05, 0.10, 0.20),
  n_rep      = 100L,
  cutoffs    = strict
)

janusplot_shape_sensitivity_summary(res_strict, level = "fine")
```

## References

- Pya, N., & Wood, S. N. (2015). Shape constrained additive models.
  *Statistics and Computing*, 25(3), 543–559.
- Calabrese, E. J. (2008). Hormesis: why it is important to
  toxicology and toxicologists. *Environmental Toxicology and
  Chemistry*, 27(7), 1451–1474.
- Milnor, J. (1963). *Morse Theory*. Princeton University Press.
- Meyer, M. C. (2008). Inference using shape-restricted regression
  splines. *Annals of Applied Statistics*, 2(3), 1013–1033.

```{r session-info}
sessionInfo()
```

# p-value glyph ladder matches snapshot

    Code
      lapply(c(1e-04, 0.005, 0.03, 0.08, 0.5, NA_real_), janusplot:::.pvalue_to_glyph)
    Output
      [[1]]
      [1] "***"
      
      [[2]]
      [1] "**"
      
      [[3]]
      [1] "*"
      
      [[4]]
      [1] "·"
      
      [[5]]
      [1] ""
      
      [[6]]
      [1] ""
      

# palette choice vector matches snapshot

    Code
      janusplot:::.palette_choices()
    Output
       [1] "viridis"  "magma"    "inferno"  "plasma"   "cividis"  "mako"    
       [7] "rocket"   "turbo"    "RdYlBu"   "RdBu"     "PuOr"     "Spectral"
      [13] "YlOrRd"   "YlGnBu"   "Blues"    "Greens"  

# .cell_text_sizes output at k = 3 matches snapshot

    Code
      janusplot:::.cell_text_sizes(3L)
    Output
      $n_edf
      [1] 4
      
      $glyph
      [1] 4.6
      
      $asym
      [1] 3.8
      
      $shape
      [1] 6
      
      $empty
      [1] 3.2
      
      $diag
      [1] 5.2
      

# .cell_text_sizes output at k = 10 matches snapshot

    Code
      janusplot:::.cell_text_sizes(10L)
    Output
      $n_edf
      [1] 2.062893
      
      $glyph
      [1] 2.372327
      
      $asym
      [1] 1.959748
      
      $shape
      [1] 3.094339
      
      $empty
      [1] 1.650314
      
      $diag
      [1] 2.681761
      

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape-sensitivity.R
\name{janusplot_shape_sensitivity_shapes}
\alias{janusplot_shape_sensitivity_shapes}
\title{Canonical ground-truth shapes for the sensitivity study}
\usage{
janusplot_shape_sensitivity_shapes()
}
\value{
Character vector of length 14 — the generator names.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Return the names of every canonical ground-truth shape that
\code{\link[=janusplot_shape_sensitivity]{janusplot_shape_sensitivity()}} can simulate from. Fourteen shapes
spanning five archetypes (\code{monotone_linear}, \code{monotone_curved},
\code{unimodal}, \code{wave}, \code{multimodal}). The \code{chaotic} and \code{degenerate}
archetypes are out of scope (no realistic deterministic generator).
}
\examples{
janusplot_shape_sensitivity_shapes()
}
\seealso{
\code{\link[=janusplot_shape_sensitivity]{janusplot_shape_sensitivity()}}, \code{\link[=janusplot_shape_hierarchy]{janusplot_shape_hierarchy()}}.

Other shape-sensitivity: 
\code{\link{janusplot_shape_sensitivity}()},
\code{\link{janusplot_shape_sensitivity_plot}()},
\code{\link{janusplot_shape_sensitivity_summary}()}
}
\concept{shape-sensitivity}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape-sensitivity.R
\name{janusplot_shape_sensitivity_plot}
\alias{janusplot_shape_sensitivity_plot}
\title{Visualise a shape-sensitivity sweep}
\usage{
janusplot_shape_sensitivity_plot(
  results,
  type = c("confusion_fine", "confusion_archetype", "accuracy_grid", "recovery_curves")
)
}
\arguments{
\item{results}{Data frame from \code{\link[=janusplot_shape_sensitivity]{janusplot_shape_sensitivity()}} or the
precomputed \link{shape_sensitivity_demo}.}

\item{type}{One of \code{"confusion_fine"}, \code{"confusion_archetype"},
\code{"accuracy_grid"}, or \code{"recovery_curves"}.}
}
\value{
A \link[ggplot2:ggplot]{ggplot2::ggplot} object.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Produce one of four diagnostic plots from the raw data frame returned
by \code{\link[=janusplot_shape_sensitivity]{janusplot_shape_sensitivity()}}:
\describe{
\item{\code{"confusion_fine"}}{24 x (|shapes|) confusion matrix at the
fine category level — rows = ground truth, columns = predicted,
cells coloured by \code{P(pred | truth)}.}
\item{\code{"confusion_archetype"}}{7 x 7 confusion matrix at the
archetype level.}
\item{\code{"accuracy_grid"}}{per-shape heatmap of archetype-level
accuracy across the \verb{(n, sigma)} design.}
\item{\code{"recovery_curves"}}{accuracy as a function of sigma,
one line per sample size, faceted by shape.}
}
}
\examples{
data("shape_sensitivity_demo", package = "janusplot")
janusplot_shape_sensitivity_plot(shape_sensitivity_demo,
                                 "recovery_curves")
}
\seealso{
Other shape-sensitivity: 
\code{\link{janusplot_shape_sensitivity}()},
\code{\link{janusplot_shape_sensitivity_shapes}()},
\code{\link{janusplot_shape_sensitivity_summary}()}
}
\concept{shape-sensitivity}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/janusplot-package.R
\docType{package}
\name{janusplot-package}
\alias{janusplot-package}
\title{janusplot: Asymmetric Smoothed-Association Matrices via GAM Fits}
\description{
\code{janusplot} renders pairwise, asymmetric smoothed-association matrices
of continuous variables. Each cell shows the fitted spline from an
\code{\link[mgcv:gam]{mgcv::gam()}} model, with upper and lower triangles encoding the two
directional regressions \code{y ~ s(x)} and \code{x ~ s(y)} respectively.

Unlike a Pearson correlation matrix (one scalar per pair, symmetric),
a smoothed-association matrix gives two curves per pair and is
intentionally asymmetric. Heteroscedasticity, leverage, and directional
non-linearity become visually evident.
}
\section{Main functions}{
\itemize{
\item \code{\link[=janusplot]{janusplot()}} — returns a ggplot of the matrix.
\item \code{\link[=janusplot_data]{janusplot_data()}} — returns the raw GAM fits + per-cell metrics
(for custom plotting or downstream analysis).
}
}

\section{Asymmetry index}{
For each pair, the asymmetry index \verb{A_ij = |EDF_yx - EDF_xy| / (EDF_yx + EDF_xy)} is bounded in [0, 1]. Values near 0 indicate
symmetric complexity; values near 1 indicate the two directional
fits differ sharply in effective degrees of freedom.

Under the additive noise model (Hoyer et al. 2009; Peters et al.
2014), the two directional regressions are generally asymmetric
when the data-generating process is non-linear, and this
asymmetry identifies the causal direction under mild conditions.
The asymmetry index is offered here as a \emph{visual pre-discovery
diagnostic} rather than a causal inference procedure; see the
package vignette and accompanying paper for full scope and
limitations (in particular the failure modes under
heteroscedasticity, confounding, and Gaussian-linear DGPs).
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/max578/janusplot}
  \item \url{https://max578.github.io/janusplot/}
  \item Report bugs at \url{https://github.com/max578/janusplot/issues}
}

}
\author{
\strong{Maintainer}: Max Moldovan \email{max.moldovan@adelaide.edu.au} (\href{https://orcid.org/0000-0001-9680-8474}{ORCID}) [copyright holder]

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape-metrics.R
\name{janusplot_shape_cutoffs}
\alias{janusplot_shape_cutoffs}
\title{Default cutoff thresholds for \code{shape_category} classification}
\usage{
janusplot_shape_cutoffs(...)
}
\arguments{
\item{...}{Optional named overrides to merge into the defaults.}
}
\value{
A named list with numeric thresholds:
\describe{
\item{\code{mono_strong}}{\verb{|M|} threshold for a strictly monotone smooth (default \code{0.9}).}
\item{\code{mono_mod}}{\verb{|M|} threshold for a curved-but-monotone smooth (default \code{0.5}).}
\item{\code{mono_nonmono}}{\verb{|M|} below this is considered non-monotone (default \code{0.3}).}
\item{\code{mono_s}}{\verb{|M|} threshold for labelling an S-shape (default \code{0.5}).}
\item{\code{curv_low}}{\verb{|C|} below this is considered near-linear curvature (default \code{0.2}).}
\item{\code{curv_mod}}{\verb{|C|} threshold for a clearly curved monotone (default \code{0.5}).}
\item{\code{curv_strong}}{\verb{|C|} threshold for a U-shape / inverted-U shape
(default \code{0.5}).}
\item{\code{flat}}{\code{range(fit) / sd(y)} below this is called \code{flat} (default \code{0.05}).}
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Returns the named list of thresholds used to map the continuous
monotonicity (\code{M}) and convexity (\code{C}) indices (plus inflection
counts) into a discrete \code{shape_category}. Expose so callers can
override individual thresholds or pass a fully custom list to
\code{\link[=janusplot]{janusplot()}} / \code{\link[=janusplot_shape_metrics]{janusplot_shape_metrics()}}.
}
\examples{
janusplot_shape_cutoffs()
janusplot_shape_cutoffs(curv_mod = 0.6, flat = 0.02)
}
\seealso{
Other shape-metrics: 
\code{\link{janusplot_shape_hierarchy}()},
\code{\link{janusplot_shape_metrics}()}
}
\concept{shape-metrics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape-metrics.R
\name{janusplot_shape_hierarchy}
\alias{janusplot_shape_hierarchy}
\title{Shape-category taxonomy table}
\usage{
janusplot_shape_hierarchy()
}
\value{
A data frame with 24 rows and columns \code{category},
\code{code}, \code{archetype}, \code{monotonic}, \code{linear}, \code{glyph}, \code{ascii},
\code{label}, \code{gloss}.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Return the full janusplot shape taxonomy as a data frame with four
hierarchy columns plus presentation fields. The taxonomy is the
single source of truth consumed by the classifier, the cell
renderer, the legend plate, and the \code{janusplot_data()} output.

Hierarchy columns (finest → coarsest):
\describe{
\item{\code{category}}{24-way fine label (\code{linear_up}, \code{skewed_peak},
\code{bimodal}, …). Computed per cell by \code{\link[=janusplot]{janusplot()}}.}
\item{\code{code}}{Unique two-letter ASCII shorthand (safe on any
font or typesetting pipeline) — e.g. \code{lu} for \code{linear_up}.}
\item{\code{archetype}}{Seven-family grouping: \code{monotone_linear},
\code{monotone_curved}, \code{unimodal}, \code{wave}, \code{multimodal},
\code{chaotic}, \code{degenerate}.}
\item{\code{monotonic}}{Three-way coarse classification: \code{monotone}
/ \code{non_monotone} / \code{degenerate}.}
\item{\code{linear}}{Binary: \code{linear} / \code{non_linear} /
\code{degenerate}.}
}

The broader tiers (linear/non-linear, monotone/non-monotone) are
textbook calculus; the archetype layer maps cleanly to
shape-constrained regression vocabulary (Pya & Wood 2015;
Meyer 2008) and to dose-response shape categories (Calabrese
2008; Calabrese & Baldwin 2001). The \verb{(T, I)} dispatch
underlying each fine category is a coarsened Morse-theoretic
critical-point classification (Milnor 1963).
}
\examples{
tax <- janusplot_shape_hierarchy()
head(tax[, c("category", "code", "archetype", "monotonic", "linear")])
# Count how many categories live in each archetype
table(tax$archetype)
}
\references{
Calabrese, E. J. (2008). Hormesis: why it is important to
toxicology and toxicologists. \emph{Environmental Toxicology and
Chemistry}, \strong{27}(7), 1451–1474.

Meyer, M. C. (2008). Inference using shape-restricted regression
splines. \emph{Annals of Applied Statistics}, \strong{2}(3), 1013–1033.

Milnor, J. (1963). \emph{Morse Theory}. Princeton University Press.

Pya, N., & Wood, S. N. (2015). Shape constrained additive models.
\emph{Statistics and Computing}, \strong{25}(3), 543–559.
}
\seealso{
Other shape-metrics: 
\code{\link{janusplot_shape_cutoffs}()},
\code{\link{janusplot_shape_metrics}()}
}
\concept{shape-metrics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape_sensitivity_demo-doc.R
\docType{data}
\name{shape_sensitivity_demo}
\alias{shape_sensitivity_demo}
\title{Precomputed shape-recognition sensitivity results (demo)}
\format{
A data frame with 2160 rows and 14 columns — see the
"Value" section of \code{\link[=janusplot_shape_sensitivity]{janusplot_shape_sensitivity()}} for the
column schema.
}
\usage{
shape_sensitivity_demo
}
\description{
Raw output from a small-footprint invocation of
\code{\link[=janusplot_shape_sensitivity]{janusplot_shape_sensitivity()}}. Shipped so users can explore the
sensitivity API and regenerate every figure in the
\code{shape-recognition-sensitivity} vignette without having to re-run
the sweep themselves. Regenerated via
\code{data-raw/shape_sensitivity_demo.R}.

Design:
\itemize{
\item \strong{Shapes} (6, one per non-degenerate archetype):
\code{linear_up}, \code{concave_up}, \code{u_shape}, \code{inverted_u}, \code{wave},
\code{bimodal}.
\item \strong{Sample sizes} (3): \code{c(100, 200, 500)}.
\item \strong{Noise levels} (4): \code{c(0.05, 0.10, 0.20, 0.40)} fraction
of y-range.
\item \strong{Replicates}: 30.
\item \strong{Total fits}: 2160.
\item \strong{Seed}: 2026.
}
}
\examples{
data("shape_sensitivity_demo", package = "janusplot")
head(shape_sensitivity_demo)
janusplot_shape_sensitivity_plot(shape_sensitivity_demo,
                                 "recovery_curves")
}
\seealso{
\code{\link[=janusplot_shape_sensitivity]{janusplot_shape_sensitivity()}},
\code{\link[=janusplot_shape_sensitivity_plot]{janusplot_shape_sensitivity_plot()}},
\code{\link[=janusplot_shape_sensitivity_summary]{janusplot_shape_sensitivity_summary()}}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape-sensitivity.R
\name{janusplot_shape_sensitivity}
\alias{janusplot_shape_sensitivity}
\title{Shape-recognition sensitivity study}
\usage{
janusplot_shape_sensitivity(
  shapes = NULL,
  n_grid = c(50L, 100L, 200L, 500L),
  sigma_grid = c(0.02, 0.05, 0.1, 0.2, 0.4),
  n_rep = 200L,
  cutoffs = janusplot_shape_cutoffs(),
  parallel = FALSE,
  seed = 2026L,
  verbose = interactive()
)
}
\arguments{
\item{shapes}{Character vector of ground-truth names from
\code{\link[=janusplot_shape_sensitivity_shapes]{janusplot_shape_sensitivity_shapes()}}. Default \code{NULL} → all 14.}

\item{n_grid}{Integer vector of sample sizes. Default
\code{c(50L, 100L, 200L, 500L)}.}

\item{sigma_grid}{Numeric vector of noise levels (fraction of the
y-range). Default \code{c(0.02, 0.05, 0.10, 0.20, 0.40)}.}

\item{n_rep}{Integer. Replicates per cell. Default \code{200L}.}

\item{cutoffs}{Named list of classification thresholds; see
\code{\link[=janusplot_shape_cutoffs]{janusplot_shape_cutoffs()}}.}

\item{parallel}{Logical. If \code{TRUE} and \code{future.apply} is installed,
dispatch replicates in parallel. The caller is responsible for
configuring \code{future::plan()} (e.g. \code{future::plan(future::multisession)}).}

\item{seed}{Integer. Base seed — each fit uses \code{seed + row_index}
so results are reproducible and cell-permutation-invariant.}

\item{verbose}{Logical. Print progress messages to the console.
Default is \code{interactive()}.}
}
\value{
A data frame with one row per fit. Columns:
\describe{
\item{\code{truth}}{Ground-truth shape name.}
\item{\code{n}}{Sample size for this fit.}
\item{\code{sigma}}{Noise level for this fit.}
\item{\code{seed}}{RNG seed used.}
\item{\code{predicted}}{Classifier output at the fine (24-category) level.}
\item{\code{correct}}{Logical — does \code{predicted == truth}?}
\item{\code{archetype_truth}}{Expected archetype for \code{truth}.}
\item{\code{archetype_pred}}{Archetype of \code{predicted}.}
\item{\code{archetype_correct}}{Logical — archetype-level correctness.}
\item{\code{monotonicity_index}}{Monotonicity index \code{M} (see
\code{\link[=janusplot_shape_metrics]{janusplot_shape_metrics()}}).}
\item{\code{convexity_index}}{Convexity index \code{C} (see
\code{\link[=janusplot_shape_metrics]{janusplot_shape_metrics()}}).}
\item{\code{n_turn}, \code{n_inflect}}{Recovered turning-point and
inflection counts.}
\item{\code{error}}{\code{"gam_fit_failed"} when \code{mgcv::gam()} errored;
\code{NA} otherwise.}
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Run a full-factorial sensitivity sweep for the janusplot 24-category
shape classifier. For each combination of ground-truth shape, sample
size \code{n}, noise level \code{sigma}, and replicate, the sweep:
\enumerate{
\item Generates \code{n} points from the noiseless canonical curve on
\verb{[0, 1]} + Gaussian noise with SD = \code{sigma} (fraction of the
y-range, so signal-to-noise is comparable across shapes).
\item Fits \code{mgcv::gam(y ~ s(x), method = "REML")}.
\item Runs \code{\link[=janusplot_shape_metrics]{janusplot_shape_metrics()}} to classify the fitted smooth.
\item Records correctness at both the fine (24-category) and
archetype (7-family) levels.
}

The function is the package-native implementation of
\code{simulation/scripts/scenario_4_shape_recognition.R}. A small
precomputed dataset is shipped as \link{shape_sensitivity_demo} for
downstream examples without requiring users to re-run the sweep.
}
\examples{
# Tiny-run smoke test (< 2 seconds): 3 shapes x 2 n x 2 sigma x 5 reps.
res <- janusplot_shape_sensitivity(
  shapes     = c("linear_up", "u_shape", "wave"),
  n_grid     = c(100L, 200L),
  sigma_grid = c(0.05, 0.20),
  n_rep      = 5L,
  verbose    = FALSE
)
head(res)
janusplot_shape_sensitivity_summary(res, level = "archetype")
}
\seealso{
\code{\link[=janusplot_shape_sensitivity_summary]{janusplot_shape_sensitivity_summary()}},
\code{\link[=janusplot_shape_sensitivity_plot]{janusplot_shape_sensitivity_plot()}},
\code{\link[=janusplot_shape_sensitivity_shapes]{janusplot_shape_sensitivity_shapes()}}, \link{shape_sensitivity_demo}.

Other shape-sensitivity: 
\code{\link{janusplot_shape_sensitivity_plot}()},
\code{\link{janusplot_shape_sensitivity_shapes}()},
\code{\link{janusplot_shape_sensitivity_summary}()}
}
\concept{shape-sensitivity}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/janusplot.R
\name{janusplot_data}
\alias{janusplot_data}
\title{Raw GAM fits and per-cell metrics for a smoothed-association matrix}
\usage{
janusplot_data(
  data,
  vars = NULL,
  adjust = NULL,
  method = "REML",
  k = -1L,
  bs = "tp",
  na_action = c("pairwise", "complete"),
  parallel = FALSE,
  keep_fits = FALSE,
  derivatives = integer(),
  derivative_ci = c("pointwise", "none", "simultaneous"),
  derivative_ci_nsim = 1000L,
  n_grid = NULL,
  shape_cutoffs = janusplot_shape_cutoffs(),
  ...
)
}
\arguments{
\item{data}{A data frame with numeric columns to include.}

\item{vars}{Character vector of column names to use. \code{NULL} (default)
uses all numeric columns in \code{data}. Non-numeric columns trigger an
error listing offenders.}

\item{adjust}{A one-sided formula RHS giving additional covariates
and/or random effects to include in every pairwise GAM. For example,
\code{adjust = ~ s(age) + s(site, bs = "re")} fits
\code{gam(y ~ s(x) + s(age) + s(site, bs = "re"))} for each pair.
Default \code{NULL} fits unadjusted pairwise smooths.}

\item{method}{Smoothing-parameter estimation method passed to
\code{\link[mgcv:gam]{mgcv::gam()}}. Default \code{"REML"} per mgcv recommendation.}

\item{k}{Integer, or named list mapping variable names to integers.
Basis dimension for \code{s()}. Default \code{-1L} (mgcv's automatic choice).}

\item{bs}{Basis type for \code{s()}. Default \code{"tp"} (thin plate).}

\item{na_action}{One of \code{"pairwise"} (default; per-cell complete
observations) or \code{"complete"} (listwise; all cells use the same
rows).}

\item{parallel}{Logical. If \code{TRUE}, use \code{future.apply::future_mapply()}
to fit pairs in parallel. Requires the \code{future.apply} package and a
user-configured \code{future::plan()}. Default \code{FALSE}.}

\item{keep_fits}{Logical. If \code{TRUE}, retain full \code{\link[mgcv:gam]{mgcv::gam()}} model
objects in the return (large memory footprint for \code{k} above ~15).
Default \code{FALSE} — retains summary metrics and prediction grids only.}

\item{derivatives}{Integer vector of derivative orders to compute
on every pair (subset of \code{1:2}). Default \code{integer()} — no
derivatives. Unlike \code{janusplot()}, the data companion can
return multiple orders from a single call for programmatic
analysis; pass \code{c(1L, 2L)} to surface both.}

\item{derivative_ci}{One of \code{"none"} (default), \code{"pointwise"}, or
\code{"simultaneous"}. Controls whether — and how — a 95\%
confidence ribbon is drawn underneath the derivative curve when
\code{display \%in\% c("d1", "d2")}. Ignored when \code{display = "fit"}.
\itemize{
\item \code{"none"} — no ribbon. The curve and the zero reference line
are all you see. Default, because pointwise ribbons overshoot
nominal coverage as a joint region and can invite
over-reading of local features.
\item \code{"pointwise"} — 95\% pointwise ribbon from
\eqn{\sqrt{\mathrm{diag}(D V_p D^\top)}} (Wood 2017 §7.2.4).
Valid marginally; not a simultaneous statement.
\item \code{"simultaneous"} — 95\% simultaneous band via the Monte Carlo
construction of Ruppert, Wand & Carroll (2003) popularised for
GAMs by Simpson (2018, \emph{Frontiers Ecol. Evol.} 6:149): draw
\eqn{B} samples \eqn{\tilde{\boldsymbol\beta} \sim
    \mathcal{N}(\hat{\boldsymbol\beta}, V_p)}, compute
\eqn{\max_x |D_i(\tilde{\boldsymbol\beta} -
    \hat{\boldsymbol\beta})| / \mathrm{se}_i}, and use the
\eqn{(1-\alpha)} quantile as a critical multiplier on the
pointwise SE. Valid for feature localisation ("where is
\eqn{\hat f'(x)} significantly non-zero").
}}

\item{derivative_ci_nsim}{Integer. Number of Monte Carlo samples
used when \code{derivative_ci = "simultaneous"}. Default \code{1000L} —
a compromise between coverage accuracy (Simpson 2018 uses
10000) and CPU budget across every pair in a medium-sized
matrix. Ignored for any other \code{derivative_ci}.}

\item{n_grid}{Integer or \code{NULL}. Number of equally-spaced points
used to evaluate each fitted smooth (and its derivatives).
Default \code{NULL} resolves to \code{100} when \code{display = "fit"} and
\code{200} otherwise, because finite-difference second derivatives
visibly degrade below \eqn{\sim 150} points on moderate-\code{k}
smooths. Supplying \code{n_grid} directly overrides both defaults.
Larger grids shift the numerical shape-metric values
(\eqn{M}, \eqn{C}, turning / inflection counts) slightly
because they are computed on this same grid. Shapes and
asymmetry are the primary reading; \code{M}, \code{C} and the counts are
secondary diagnostics and the grid-induced drift is tolerable.}

\item{shape_cutoffs}{Named list of classification thresholds used to
map the continuous shape indices (\code{monotonicity_index},
\code{convexity_index}) into discrete
\code{shape_category} labels. Defaults from \code{\link[=janusplot_shape_cutoffs]{janusplot_shape_cutoffs()}}.}

\item{...}{Additional arguments passed to \code{\link[mgcv:gam]{mgcv::gam()}}.}
}
\value{
A list with components:
\describe{
\item{\code{vars}}{Character vector of variables used, in plotted order.}
\item{\code{pairs}}{List of per-pair results. Each element has \code{i}, \code{j},
\code{var_i}, \code{var_j}, \code{fit_yx}, \code{fit_xy} (NULL if \code{keep_fits = FALSE}),
\code{pred_yx}, \code{pred_xy} (data frames with \code{x}, \code{fit}, \code{se}, \code{lo},
\code{hi}), \code{edf_yx}, \code{edf_xy}, \code{pvalue_yx}, \code{pvalue_xy}, \code{dev_exp_yx},
\code{dev_exp_xy}, \code{n_used}, \code{asymmetry_index}, plus Pearson /
Spearman / Kendall correlations (\code{cor_pearson}, \code{cor_spearman},
\code{cor_kendall}), the maximum tie ratio across \code{x} and \code{y}
(\code{tie_ratio}), and per-direction shape descriptors
(\code{monotonicity_index_yx}, \code{convexity_index_yx},
\code{monotonicity_index_xy}, \code{convexity_index_xy},
\code{n_turning_yx}, \code{n_inflect_yx}, \code{n_turning_xy},
\code{n_inflect_xy}, \code{shape_yx}, \code{shape_xy}).
When \code{derivatives} is non-empty, each pair additionally
carries \code{deriv_yx} and \code{deriv_xy}, each a named list keyed by
order (\code{"1"}, \code{"2"}) whose entries are data frames with
columns \code{x}, \code{fit}, \code{se}, \code{lo}, \code{hi}, \code{ci_type} matching the
schema of \code{pred_yx} / \code{pred_xy}. The \code{ci_type} column records
whether the \code{lo} / \code{hi} columns are \code{"pointwise"} (default),
\code{"simultaneous"} (Ruppert--Wand--Carroll / Simpson 2018
critical-multiplier bands), or \code{"none"}. When
\code{derivative_ci = "simultaneous"}, each derivative frame also
carries a \code{"crit_multiplier"} attribute giving the MC-derived
critical multiplier used.
See \code{\link[=janusplot_shape_metrics]{janusplot_shape_metrics()}} for the definition of the
monotonicity and convexity indices.}
\item{\code{call}}{Match call.}
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Companion to \code{\link[=janusplot]{janusplot()}} returning the raw list of GAM fits plus
per-cell metrics (EDF, F-test p-value, deviance explained, asymmetry
index, pairwise correlations, shape descriptors) without constructing
the ggplot. Useful for custom rendering or downstream analysis.
}
\examples{
# Per-pair fits + metrics on a small mtcars slice
out <- janusplot_data(mtcars[, c("mpg", "hp", "wt")])
out$pairs[[1L]]$asymmetry_index
out$pairs[[1L]]$cor_spearman
out$pairs[[1L]]$shape_yx
}
\seealso{
\code{\link[=janusplot]{janusplot()}} for the ggplot front-end,
\code{\link[=janusplot_shape_metrics]{janusplot_shape_metrics()}} for the shape-metric primitives.

Other smooth-associations: 
\code{\link{janusplot}()}
}
\concept{smooth-associations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape-sensitivity.R
\name{janusplot_shape_sensitivity_summary}
\alias{janusplot_shape_sensitivity_summary}
\title{Summarise a shape-sensitivity sweep}
\usage{
janusplot_shape_sensitivity_summary(results, level = c("fine", "archetype"))
}
\arguments{
\item{results}{Data frame returned by \code{\link[=janusplot_shape_sensitivity]{janusplot_shape_sensitivity()}}.}

\item{level}{One of \code{"fine"} (default) or \code{"archetype"}.}
}
\value{
A data frame with columns \code{truth}, \code{n}, \code{sigma}, \code{accuracy}.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Aggregate the raw output of \code{\link[=janusplot_shape_sensitivity]{janusplot_shape_sensitivity()}} into a
per-cell mean-accuracy table at either the fine (24-category) or
archetype (7-family) level.
}
\examples{
data("shape_sensitivity_demo", package = "janusplot")
head(janusplot_shape_sensitivity_summary(shape_sensitivity_demo,
                                         level = "archetype"))
}
\seealso{
Other shape-sensitivity: 
\code{\link{janusplot_shape_sensitivity}()},
\code{\link{janusplot_shape_sensitivity_plot}()},
\code{\link{janusplot_shape_sensitivity_shapes}()}
}
\concept{shape-sensitivity}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape-metrics.R
\name{janusplot_shape_metrics}
\alias{janusplot_shape_metrics}
\title{Shape metrics for a fitted univariate smooth}
\usage{
janusplot_shape_metrics(
  fit,
  x_name = NULL,
  newdata = NULL,
  n_grid = 200L,
  cutoffs = janusplot_shape_cutoffs()
)
}
\arguments{
\item{fit}{Either a list returned by a janusplot pair-fit helper
(must contain \code{pred} and \code{raw}), or a fitted \code{\link[mgcv:gam]{mgcv::gam()}} with
a single \code{s(x)} term.}

\item{x_name}{Character. Column name of the predictor when \code{fit} is
a \code{\link[mgcv:gam]{mgcv::gam()}} object. Ignored for pair-fit lists.}

\item{newdata}{Optional data frame supplying the raw predictor values
used for density weighting when \code{fit} is a \code{\link[mgcv:gam]{mgcv::gam()}} object.
If \code{NULL}, the model frame is used.}

\item{n_grid}{Integer. Prediction grid length when \code{fit} is a
\code{\link[mgcv:gam]{mgcv::gam()}} object. Default \code{200L}.}

\item{cutoffs}{Named list of classification thresholds; see
\code{\link[=janusplot_shape_cutoffs]{janusplot_shape_cutoffs()}}. Default uses package defaults.}
}
\value{
A named list with components:
\describe{
\item{\code{monotonicity_index}}{\code{M} in \verb{[-1, 1]}. See Description.}
\item{\code{convexity_index}}{\code{C} in \verb{[-1, 1]}. See Description.}
\item{\code{n_turning_points}}{Integer count of lobe-mass-weighted
sign changes of \verb{f'}. Equals the number of interior extrema.}
\item{\code{n_inflections}}{Integer count of lobe-mass-weighted
sign changes of \verb{f''}.}
\item{\code{flat_range_ratio}}{\code{range(f) / sd(y)} — small values
indicate a degenerate flat smooth.}
\item{\code{shape_category}}{One of 24 labels from
\code{\link[=janusplot_shape_hierarchy]{janusplot_shape_hierarchy()}} dispatched on
\verb{(n_turning_points, n_inflections)} with
\verb{(monotonicity_index, convexity_index)} disambiguation for
the monotone case.}
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Compute the continuous monotonicity and convexity indices, inflection
and turning-point counts, and rule-based shape category for a fitted
univariate smooth. Works on either a per-pair fit object returned
from the janusplot internal machinery or a freshly fitted
\code{\link[mgcv:gam]{mgcv::gam()}} with a single \code{s()} term.

Both indices are bounded in \verb{[-1, 1]} and weighted by the empirical
density of the predictor:
\itemize{
\item \code{monotonicity_index} (paper symbol \code{M}). Let \code{f} be the fitted
smooth evaluated on a dense grid of \code{n_grid} equally-spaced
points across the predictor range, \verb{f'} its numerical first
derivative, and \code{w} the empirical density of the predictor on
the same grid with \code{sum(w) = 1}. Then
\verb{monotonicity_index = sum(w * f') / sum(w * |f'|) in [-1, 1]}.
\code{+1} is strictly increasing, \code{-1} strictly decreasing, \code{0}
non-monotone.
\item \code{convexity_index} (paper symbol \code{C}). With \verb{f''} the numerical
second derivative on the same grid,
\verb{convexity_index = sum(w * f'') / sum(w * |f''|) in [-1, 1]}.
\code{+1} is globally convex (bowl-up), \code{-1} globally concave
(bowl-down), \code{0} inflection-dominated (S-curve, sine, flat).
}

Both indices are scale-invariant (replacing \code{y -> a*y + b} leaves
them unchanged) and density-weighted so they describe the smooth
\emph{where the data actually live}, not extrapolated tails.
}
\examples{
# On a fitted gam
set.seed(2026L)
n  <- 200L
x  <- stats::runif(n, 0, 10)
y  <- log1p(x) + stats::rnorm(n, sd = 0.3)
d  <- data.frame(x = x, y = y)
fit <- mgcv::gam(y ~ s(x), data = d, method = "REML")
janusplot_shape_metrics(fit, x_name = "x", newdata = d)
}
\seealso{
\code{\link[=janusplot_shape_cutoffs]{janusplot_shape_cutoffs()}}, \code{\link[=janusplot]{janusplot()}}, \code{\link[=janusplot_data]{janusplot_data()}}.

Other shape-metrics: 
\code{\link{janusplot_shape_cutoffs}()},
\code{\link{janusplot_shape_hierarchy}()}
}
\concept{shape-metrics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/janusplot.R
\name{janusplot}
\alias{janusplot}
\title{Asymmetric smoothed-association matrix}
\usage{
janusplot(
  data,
  vars = NULL,
  adjust = NULL,
  method = NULL,
  k = -1L,
  bs = "tp",
  engine = c("bam", "gam"),
  discrete = FALSE,
  nthreads = 1L,
  order = c("original", "hclust", "alphabetical"),
  show_data = TRUE,
  show_ci = TRUE,
  display = c("fit", "d1", "d2"),
  derivative_ci = c("none", "pointwise", "simultaneous"),
  derivative_ci_nsim = 1000L,
  n_grid = NULL,
  colour_by = c("pearson", "spearman", "kendall", "edf", "deviance_gap", "none"),
  fill_by = NULL,
  palette = NULL,
  annotations = c("edf", "A"),
  shape_cutoffs = janusplot_shape_cutoffs(),
  show_shape_legend = TRUE,
  glyph_style = c("ascii", "unicode"),
  labels = c("border", "diagonal", "none"),
  diagonal = c("auto", "blank", "name", "density"),
  label_srt = 45,
  label_cex = 1,
  signif_glyph = TRUE,
  show_asymmetry = NULL,
  na_action = c("pairwise", "complete"),
  parallel = FALSE,
  with_data = FALSE,
  text_scale_diag = 1,
  text_scale_off_diag = 1,
  show_glossary = TRUE,
  glossary_scale = 1,
  k_check_thresholds = NULL,
  auto_refit_k = FALSE,
  k_max_iter = 2L,
  compact = c("auto", "always", "never"),
  compact_threshold = 12L,
  compact_levels = NULL,
  focus_by = NA_character_,
  focus_threshold = "q90",
  focus_dim_alpha = 0.25,
  axes = c("original", "standardised", "centred", "rank"),
  save_as = NULL,
  save_width = NULL,
  save_height = NULL,
  save_dpi = 300,
  ...
)
}
\arguments{
\item{data}{A data frame with numeric columns to include.}

\item{vars}{Character vector of column names to use. \code{NULL} (default)
uses all numeric columns in \code{data}. Non-numeric columns trigger an
error listing offenders.}

\item{adjust}{A one-sided formula RHS giving additional covariates
and/or random effects to include in every pairwise GAM. For example,
\code{adjust = ~ s(age) + s(site, bs = "re")} fits
\code{gam(y ~ s(x) + s(age) + s(site, bs = "re"))} for each pair.
Default \code{NULL} fits unadjusted pairwise smooths.}

\item{method}{Smoothing-parameter estimation method passed to
the chosen fitting backend. Default \code{NULL} resolves
per-engine: \code{"fREML"} for \code{engine = "bam"} (mgcv's
recommended at scale), \code{"REML"} for \code{engine = "gam"}
(the v0.1.0 behaviour). Pass any valid mgcv method string
to override.}

\item{k}{Integer, or named list mapping variable names to integers.
Basis dimension for \code{s()}. Default \code{-1L} (mgcv's automatic choice).}

\item{bs}{Basis type for \code{s()}. Default \code{"tp"} (thin plate).}

\item{engine}{One of \code{"bam"} (default, \strong{new in v0.1.1}) or
\code{"gam"}. Selects mgcv's fitting backend:
\itemize{
\item \code{"bam"} — \code{\link[mgcv:bam]{mgcv::bam()}}. Block-Lanczos solve + fREML
estimation + lower memory. ~3-10x speedup at janusplot's
scale (k = 15-25 vars, 600+ pairwise fits per call). The
\strong{default}, and the one non-byte-identical change in
v0.1.1: fREML differs from REML by ~1-3\% in EDF on
identical data, so the asymmetry index may shift by
similar amounts vs v0.1.0 output. Recoverable verbatim
via \code{engine = "gam"}.
\item \code{"gam"} — \code{\link[mgcv:gam]{mgcv::gam()}}. The v0.1.0 backend. Use for
backward-compat reproduction, very small n (< 200) where
bam's setup overhead exceeds its solve gain, or
methodologically sensitive contexts that require REML
rather than fREML.
}}

\item{discrete}{Logical. \code{bam}-only opt-in to mgcv's
covariate-discretisation optimisation. Further ~2-5x
speedup at the cost of small (sub-pixel at typical
janusplot resolution) prediction-shift. Default \code{FALSE}.
Ignored when \code{engine = "gam"}.}

\item{nthreads}{Integer. \code{bam}-only intra-fit threading.
Default \code{1L} to avoid oversubscription when combined with
\code{parallel = TRUE} (\code{future.apply} fans out pair fits
across cores; nthreads > 1 within each pair would
double-book CPUs). Raise above 1 only when \code{parallel = FALSE}. Ignored when \code{engine = "gam"}.}

\item{order}{One of \code{"original"} (default), \code{"hclust"} (reorder by
hierarchical clustering of Pearson correlations), or \code{"alphabetical"}.}

\item{show_data}{Logical. If \code{TRUE} (default), overlay raw data
points (low alpha) behind each spline. Only applies when
\code{display = "fit"}; derivative panels never overlay raw data.}

\item{show_ci}{Logical. If \code{TRUE} (default), overlay the 95\%
confidence envelope from \code{predict(gam, se.fit = TRUE)} on the
fit panel (i.e. when \code{display = "fit"}). CI rendering on
derivative panels is controlled separately by \code{derivative_ci}.}

\item{display}{One of \code{"fit"} (default), \code{"d1"}, or \code{"d2"}.
Selects which single quantity is rendered in every
off-diagonal cell of the matrix.
\itemize{
\item \code{"fit"} — the fitted smooth \eqn{\hat f(x)}; default,
behaviour identical to the pre-derivative release.
\item \code{"d1"} — the first derivative \eqn{\hat f'(x)} of the
fitted smooth. Zero crossings localise turning points of
\eqn{\hat f}.
\item \code{"d2"} — the second derivative \eqn{\hat f''(x)}. Zero
crossings localise inflection points of \eqn{\hat f}.
}

A single matrix shows a single quantity by design: stacked
multi-panel cells crowd the matrix at any realistic variable
count. To compare fit against derivative, render two or three
\code{janusplot()} calls side-by-side; each call keeps its own
\code{with_data = TRUE} summary table tagged with the \code{display}
column.

Orders \eqn{k \ge 3} are not exposed — higher-order derivatives
of penalised regression splines amplify noise and rarely carry
usable signal at realistic sample sizes. See
\code{vignette("janusplot")} for the theoretical justification and
applied use-cases.}

\item{derivative_ci}{One of \code{"none"} (default), \code{"pointwise"}, or
\code{"simultaneous"}. Controls whether — and how — a 95\%
confidence ribbon is drawn underneath the derivative curve when
\code{display \%in\% c("d1", "d2")}. Ignored when \code{display = "fit"}.
\itemize{
\item \code{"none"} — no ribbon. The curve and the zero reference line
are all you see. Default, because pointwise ribbons overshoot
nominal coverage as a joint region and can invite
over-reading of local features.
\item \code{"pointwise"} — 95\% pointwise ribbon from
\eqn{\sqrt{\mathrm{diag}(D V_p D^\top)}} (Wood 2017 §7.2.4).
Valid marginally; not a simultaneous statement.
\item \code{"simultaneous"} — 95\% simultaneous band via the Monte Carlo
construction of Ruppert, Wand & Carroll (2003) popularised for
GAMs by Simpson (2018, \emph{Frontiers Ecol. Evol.} 6:149): draw
\eqn{B} samples \eqn{\tilde{\boldsymbol\beta} \sim
    \mathcal{N}(\hat{\boldsymbol\beta}, V_p)}, compute
\eqn{\max_x |D_i(\tilde{\boldsymbol\beta} -
    \hat{\boldsymbol\beta})| / \mathrm{se}_i}, and use the
\eqn{(1-\alpha)} quantile as a critical multiplier on the
pointwise SE. Valid for feature localisation ("where is
\eqn{\hat f'(x)} significantly non-zero").
}}

\item{derivative_ci_nsim}{Integer. Number of Monte Carlo samples
used when \code{derivative_ci = "simultaneous"}. Default \code{1000L} —
a compromise between coverage accuracy (Simpson 2018 uses
10000) and CPU budget across every pair in a medium-sized
matrix. Ignored for any other \code{derivative_ci}.}

\item{n_grid}{Integer or \code{NULL}. Number of equally-spaced points
used to evaluate each fitted smooth (and its derivatives).
Default \code{NULL} resolves to \code{100} when \code{display = "fit"} and
\code{200} otherwise, because finite-difference second derivatives
visibly degrade below \eqn{\sim 150} points on moderate-\code{k}
smooths. Supplying \code{n_grid} directly overrides both defaults.
Larger grids shift the numerical shape-metric values
(\eqn{M}, \eqn{C}, turning / inflection counts) slightly
because they are computed on this same grid. Shapes and
asymmetry are the primary reading; \code{M}, \code{C} and the counts are
secondary diagnostics and the grid-induced drift is tolerable.}

\item{colour_by}{One of \code{"pearson"} (default), \code{"spearman"},
\code{"kendall"}, \code{"edf"}, \code{"deviance_gap"}, or \code{"none"}. Encodes the
per-cell fill colour by the chosen scalar. Correlation choices use
a diverging palette with limits \code{c(-1, 1)} and a shared \code{corr}
colour-bar title; \code{"edf"} and \code{"deviance_gap"} use a sequential
palette labelled by the metric.}

\item{fill_by}{Deprecated alias for \code{colour_by}. If supplied emits a
single soft deprecation warning and is forwarded to \code{colour_by}.}

\item{palette}{Character. Colour palette for the cell fill scale.
Defaults to \code{"RdBu"} when \code{colour_by} is a correlation and
\code{"viridis"} otherwise. Sequential choices: \code{"viridis"}, \code{"magma"},
\code{"inferno"}, \code{"plasma"}, \code{"cividis"}, \code{"mako"}, \code{"rocket"},
\code{"turbo"} (not CB-safe), \code{"YlOrRd"}, \code{"YlGnBu"}, \code{"Blues"},
\code{"Greens"}. Diverging choices: \code{"RdYlBu"}, \code{"RdBu"}, \code{"PuOr"},
\code{"Spectral"} (not CB-safe). Passing a sequential palette while
\code{colour_by} is a correlation silently upgrades to the default
diverging palette.}

\item{annotations}{Character vector, a subset of
\code{c("edf", "A", "shape", "code")}. Controls which corner
annotations appear on each off-diagonal cell:
\itemize{
\item \code{"code"} — 2-letter ASCII shape code, \strong{top-left} corner.
\item \code{"A"} and \code{"edf"} — asymmetry index and effective degrees of
freedom, stacked \strong{bottom-left}.
\item \code{"shape"} — shape glyph (Unicode or ASCII per \code{glyph_style}),
\strong{bottom-right} corner.
}

Default \code{c("edf", "A")}. \code{"code"} and \code{"shape"} occupy distinct
corners so both can be requested together. See
\code{\link[=janusplot_shape_hierarchy]{janusplot_shape_hierarchy()}} for the full code list.}

\item{shape_cutoffs}{Named list of classification thresholds used to
map the continuous shape indices into discrete \code{shape_category}
labels; see \code{\link[=janusplot_shape_cutoffs]{janusplot_shape_cutoffs()}}.}

\item{show_shape_legend}{Logical. If \code{TRUE} (default), attach a
standing shape-types legend plate below the matrix that
illustrates every category in the taxonomy as a canonical
thumbnail spline. Independent of \code{annotations}.}

\item{glyph_style}{One of \code{"ascii"} (default) or \code{"unicode"}.
Controls how cell shape glyphs render when \code{"shape"} is included
in \code{annotations}. Default is \code{"ascii"} for maximum portability
across typesetting pipelines; switch to \code{"unicode"} only when
the target font is known to cover the curve glyph set.}

\item{labels}{One of \code{"border"} (default), \code{"diagonal"}, or
\code{"none"}. Controls where variable names are rendered:
\itemize{
\item \code{"border"} — names along the top (rotated per \code{label_srt}) and
left margins of the matrix; diagonal cells are left blank.
Mirrors \code{corrplot}'s \code{tl.pos = "lt"} convention.
\item \code{"diagonal"} — names centred on the diagonal cells (the
pre-0.1 layout).
\item \code{"none"} — labels suppressed entirely; diagonal cells blank.
}}

\item{diagonal}{One of \code{"auto"} (default), \code{"blank"}, \code{"name"},
or \code{"density"}. Controls what is rendered in the diagonal
cells of the matrix.
\itemize{
\item \code{"auto"} — preserves the historical behaviour: variable name
when \code{labels = "diagonal"}, blank otherwise.
\item \code{"blank"} — empty bordered panel (uniform grid reading).
\item \code{"name"} — variable name centred in the cell, bold.
\item \code{"density"} — kernel density of the variable filled in
translucent grey, with a rug of raw values along the bottom
edge. Mirrors the \code{GGally::ggpairs} convention; surfaces
tail weight, bimodality, and support clipping that the
pairwise smooths alone cannot reveal. Variable names should
come from the border (\code{labels = "border"}, the default) when
this mode is on.
}}

\item{label_srt}{Numeric. Rotation (degrees) of top labels when
\code{labels = "border"}. Default \code{45}; set to \code{0} for horizontal or
\code{90} for vertical. Ignored when \code{labels != "border"}.}

\item{label_cex}{Positive numeric multiplier on the border-label
font size. Default \code{1}. Ignored when \code{labels = "none"}.}

\item{signif_glyph}{Logical. If \code{TRUE} (default), annotate cells
with \verb{·} / \code{*} / \verb{**} reflecting the smooth's F-test p-value.}

\item{show_asymmetry}{Deprecated. Use \code{annotations} instead
(\code{"A" \%in\% annotations}). When supplied, a soft deprecation
warning fires and the argument is merged into \code{annotations}.}

\item{na_action}{One of \code{"pairwise"} (default; per-cell complete
observations) or \code{"complete"} (listwise; all cells use the same
rows).}

\item{parallel}{Logical. If \code{TRUE}, use \code{future.apply::future_mapply()}
to fit pairs in parallel. Requires the \code{future.apply} package and a
user-configured \code{future::plan()}. Default \code{FALSE}.}

\item{with_data}{Logical. If \code{TRUE}, return a two-element list
\code{list(plot, data)} where \code{data} is a flat per-cell summary
(one row per off-diagonal cell) of everything the plot displays.
The \code{data} element is always a plain \code{data.frame} (base R — no
\code{data.table} dependency). Default \code{FALSE} — in which case only
the ggplot is returned.}

\item{text_scale_diag}{Positive numeric multiplier applied to the
diagonal variable-name labels. Default \code{1}. Diagonal labels
additionally auto-shrink for long variable names
(\code{nchar(var) > 10}) so they fit the cell regardless of this value.}

\item{text_scale_off_diag}{Positive numeric multiplier applied to
all off-diagonal annotations (\code{n} / \code{EDF} readouts, significance
glyphs, asymmetry-index labels). Default \code{1}. Use \verb{< 1} when
cells are small and the annotations crowd the fit line; use
\verb{> 1} for presentation plots.}

\item{show_glossary}{Logical. If \code{TRUE} (default), attach a
multi-line caption below the matrix describing the on-plot
abbreviations (\code{n}, \code{EDF}, \code{A}, fill encoding, significance
glyphs). Only keys actually displayed are listed.}

\item{glossary_scale}{Positive numeric multiplier on the glossary
caption font size. Default \code{1}.}

\item{k_check_thresholds}{Named list giving the three flag-criterion
thresholds used by \code{mgcv::k.check()}-style basis-dimension
diagnostics. Required entries: \code{edf_ratio} (Wood's
\eqn{\widehat{\mathrm{edf}}/k'} ratio above which the smooth is
too close to its basis cap), \code{k_index} (residual-difference
variance ratio below which the basis appears underspecified), and
\code{p} (the simulation p-value below which the basis-deficiency
signal is significant). Defaults — \code{edf_ratio = 0.9}, \code{k_index = 1.0},
\code{p = 0.05} — track \code{mgcv::gam.check()} and Wood (2017) §5.9.}

\item{auto_refit_k}{Logical. If \code{TRUE}, every cell whose Wood
trifecta flags an underfit is refit with a doubling-k loop until
either the flag clears, the per-cell unique-x cap is reached, or
\code{k_max_iter} iterations have passed. Default \code{FALSE} — the
diagnostic (\code{k_check_status}, \code{k_flag}, \code{k_prime}, \code{k_index},
\code{k_p}) is always computed and surfaced regardless of this flag,
but the refit is opt-in because it can multiply wall time on
pathological data.}

\item{k_max_iter}{Integer. Maximum number of doublings allowed
per cell when \code{auto_refit_k = TRUE}. Default \code{2L} (so a cell
that starts at the \code{mgcv} default \code{k = 10} will visit at most
\code{k = 20} and then \code{k = 40}, capped by the per-cell unique-x
limit). Ignored when \code{auto_refit_k = FALSE}.}

\item{compact}{One of \code{"auto"} (default), \code{"always"}, or \code{"never"}.
Controls scale-aware content suppression per cell:
\itemize{
\item \code{"auto"} — tier 0 at \code{n_var < compact_threshold} (the v0.1.0
behaviour); progressively suppresses scatter, then CI, then
annotations, then the spline itself as \code{n_var} crosses the
\code{compact_levels} ladder. The matrix remains readable as
\code{k} grows toward 25–30 by trading detail for legibility.
\item \code{"always"} — force at least tier 1 regardless of \code{n_var}.
Useful for very dense fixed-size renders.
\item \code{"never"} — force tier 0 regardless of \code{n_var}. Useful for
reproducing v0.1.0 figures on large matrices.
}}

\item{compact_threshold}{Integer. The \code{n_var} value at which
tier 1 (drop scatter) auto-activates under \code{compact = "auto"}.
Default \code{12L}, anchored on the 150 × 150 px-per-cell pixel
budget at typical 6"×6" 300 DPI R Journal figures.}

\item{compact_levels}{Optional named list with entries \code{t1}, \code{t2},
\code{t3} overriding the auto-tier ladder. Defaults derive from
\code{compact_threshold}: \code{t1 = compact_threshold},
\code{t2 = compact_threshold + 6}, \code{t3 = compact_threshold + 13}.
\code{NULL} (default) uses these derived defaults.}

\item{focus_by}{One of \code{NA} (default — no filter), \code{"asymmetry"},
\code{"edf"}, \code{"k_flag"}, or \code{"non_linearity"} (defined as \code{edf - 1}).
When set, cells whose chosen metric falls below \code{focus_threshold}
are rendered in \code{grey85} at alpha \code{focus_dim_alpha}; the matrix
shape is preserved so attention drains visually to high-metric
cells. This is a \strong{visual filter, not a statistical one} —
the underlying fits are unchanged and the \code{with_data} table
carries every cell.}

\item{focus_threshold}{Either a quantile-string like \code{"q90"}
(default) or a numeric cutoff. The quantile is taken over the
non-\code{NA} distribution of \code{focus_by} across all off-diagonal
cells in the matrix. Ignored when \code{focus_by = NA}.}

\item{focus_dim_alpha}{Numeric in \verb{[0, 1]}. Alpha applied to the
\code{grey85} wash on unfocused cells. Default \code{0.25}. Ignored when
\code{focus_by = NA}.}

\item{axes}{One of \code{"original"} (default), \code{"standardised"},
\code{"centred"}, or \code{"rank"}. \strong{Rendering-only knob} — the
underlying \code{mgcv::gam} fits are byte-identical across all four
modes (verifiable via \code{digest::digest()} on the fit list); the
transformation lives entirely inside the cell renderer and
propagates to (a) the raw scatter, (b) the spline prediction
grid, (c) the CI ribbon, and (d) the variable label on the
matrix border. Use:
\itemize{
\item \code{"original"} — raw units. Maximum interpretability per cell.
v0.1.0 behaviour.
\item \code{"standardised"} — \code{(x - mean(x)) / sd(x)} per variable.
Border label becomes e.g. \code{"mpg (z)"}. Pairs scaled into a
comparable visual range; useful at \code{k >= 15} when raw-unit
panels look disparate.
\item \code{"centred"} — \code{x - mean(x)} per variable. Border label
becomes e.g. \code{"mpg (centred)"}. Preserves units while
anchoring the origin.
\item \code{"rank"} — empirical-CDF-based rank, scaled to \verb{[0, n]} per
variable. Border label becomes \code{"rank(mpg)"}. Sanity-check
view: collapses outliers; if the smooth changes shape vs
\code{"original"} the relationship is monotone-but-not-linear.
}

At compact tier 3 (\code{n_var >= 25} under \code{compact = "auto"}),
the cells render only colour fill + shape-class glyph — no
curve, no scatter — so \code{axes} becomes a \strong{documented no-op}
(the border labels still pick up the mode suffix).}

\item{save_as}{Optional file path with extension. When set,
the final assembled plot is written to this path via
\code{\link[ggplot2:ggsave]{ggplot2::ggsave()}}; the device is inferred from the
extension. Supported extensions: \code{.png}, \code{.pdf}, \code{.svg},
\code{.jpg} / \code{.jpeg}, \code{.tif} / \code{.tiff}, \code{.eps}, \code{.ps}, \code{.bmp}.
Default \code{NULL} (no file written; \code{janusplot()} still
returns the ggplot). Width / height default to \code{pmax(6, 0.65 * k_n)} inches each — square aspect, scaling with
matrix dimension.}

\item{save_width}{Numeric. Override width (inches) for
\code{save_as}. Default \code{NULL} uses the auto-resolved square.}

\item{save_height}{Numeric. Override height (inches) for
\code{save_as}. Default \code{NULL} uses the auto-resolved square.}

\item{save_dpi}{Integer. Raster DPI for \code{save_as} when the
inferred device is bitmap (png/jpg/tiff/bmp). Default \code{300},
matching the R Journal target.}

\item{...}{Additional arguments passed to \code{\link[mgcv:gam]{mgcv::gam()}}.}
}
\value{
If \code{with_data = FALSE} (default), a \link[ggplot2:ggplot]{ggplot2::ggplot} object
(via \code{\link[patchwork:wrap_plots]{patchwork::wrap_plots()}}) carrying a top-of-matrix title
that names the displayed quantity (\code{"Direct fit"},
\code{"First derivative f'"}, or \code{"Second derivative f''"}). If
\code{with_data = TRUE}, a list with two elements: \code{plot} (the
ggplot) and \code{data} (a tidy table with columns \code{var_x}, \code{var_y},
\code{position}, \code{n_used}, \code{edf}, \code{pvalue}, \code{signif}, \code{dev_exp},
\code{asymmetry_index}, \code{cor_pearson}, \code{cor_spearman},
\code{cor_kendall}, \code{tie_ratio}, \code{monotonicity_index},
\code{convexity_index}, \code{n_turning_points}, \code{n_inflections},
\code{flat_range_ratio}, \code{shape_category}, \code{colour_value},
\code{display}, one row per off-diagonal cell). The \code{display}
column tags which quantity the call rendered, so separate
calls for fit / d1 / d2 yield comparable, stackable tables.
Derivative \emph{curves} themselves (grid of \eqn{x}, fitted
\eqn{\hat f^{(k)}}, SE) live on \code{janusplot_data()} — see there.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Render a pairwise, asymmetric matrix of smoothed associations between
numeric variables. Each cell [i, j] where \code{i != j} shows the fitted
spline from \code{\link[mgcv:gam]{mgcv::gam()}}:
\itemize{
\item Upper triangle (\code{i < j}): \verb{gam(x_j ~ s(x_i) + <adjust>)}.
\item Lower triangle (\code{i > j}): \verb{gam(x_i ~ s(x_j) + <adjust>)}.
\item Diagonal: blank panel when labels live on the border (default),
or a variable-name label when \code{labels = "diagonal"}.
}

The two triangles intentionally differ — the asymmetry reveals
heteroscedasticity, leverage, and directional non-linearity that a
single scalar correlation hides.
}
\examples{
# Minimal runnable example — 3 variables, 6 asymmetric pairwise GAM fits.
janusplot(mtcars[, c("mpg", "hp", "wt")])

\donttest{
# Heteroscedastic DGP: Pearson r is ~ 0.9 but the inverse fit is
# clearly non-linear, yielding asymmetry index > 0.5.
set.seed(2026L)
n  <- 200L
x1 <- stats::runif(n, 0, 10)
x2 <- x1 + stats::rnorm(n, sd = 0.2 * x1)
janusplot(data.frame(x1 = x1, x2 = x2, x3 = stats::rnorm(n)))

# A single matrix renders a single quantity. To compare the fit
# against its derivatives, render three calls and place them
# side-by-side; each call's title makes the quantity explicit.
set.seed(2026L)
xs <- stats::runif(300L, -3, 3)
df <- data.frame(
  x  = xs,
  y1 = sin(xs)  + stats::rnorm(300L, sd = 0.3),
  y2 = xs^2     + stats::rnorm(300L, sd = 0.6)
)
janusplot(df, display = "fit")
janusplot(df, display = "d1")
janusplot(df, display = "d2")

# Simultaneous CI bands on a derivative panel, per Simpson (2018).
janusplot(df, display = "d1", derivative_ci = "simultaneous")
}
}
\seealso{
\code{\link[=janusplot_data]{janusplot_data()}} for the raw per-cell fits + metrics.

Other smooth-associations: 
\code{\link{janusplot_data}()}
}
\concept{smooth-associations}
